/**
* @file Controller/RoboCupCtrl.h
*
* This file declares the class RoboCupCtrl.
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#pragma once

#include <QList>
#include <list>

#include "SimRobotCore2.h"
#include "Oracle.h"

class Robot;

/**
* The class implements the SimRobot controller for RoboCup.
*/
class RoboCupCtrl : public SimRobot::Module
{
public:
  static RoboCupCtrl* controller; /**< A pointer to the SimRobot controller. */
  static SimRobot::Application* application; /**< The interface to the SimRobot GUI */

  /**
  * Constructor.
  */
  RoboCupCtrl(SimRobot::Application& application);

  /**
  * Adds a scene graph object to the scene graph displayed in SimRobot
  * @param object The scene graph object to add
  * @param parent The parent scene graph object (e.g. a category or null)
  * @param flags Some flags for registering the scene graph object (see SimRobot::Flag)
  */
  void addView(SimRobot::Object* object, const SimRobot::Object* parent = 0, int flags = 0);

  /**
  * Adds a scene graph object to the scene graph displayed in SimRobot
  * @param object The scene graph object to add
  * @param categoryName The full name of the parent categroy
  * @param flags Some flags for registering the scene graph object (see SimRobot::Flag)
  */
  void addView(SimRobot::Object* object, const QString& categoryName, int flags = 0);

  /**
  * Adds a category to the scene graph displayed in SimRobot that can be used for grouping views
  * @param name The name of the category
  * @param parent The parent scene graph object (e.g. a category or null)
  * @return The category
  */
  SimRobot::Object* addCategory(const QString& name, const SimRobot::Object* parent = 0);

  /**
  * Adds a category to the scene graph displayed in SimRobot that can be used for grouping views
  * @param name The name of the category
  * @param parentName The name of the parent scene graph object (e.g. a category)
  * @return The category
  */
  SimRobot::Object* addCategory(const QString& name, const QString& parentName);

  /**
  * The function returns the full name of the robot currently constructed.
  * @return The name of the robot.
  */
  static const char* getRobotFullName() {return controller->robotName;}

  /**
  * The function returns the name of the robot associated to the current thread.
  * @return The name of the robot.
  */
  std::string getRobotName() const;

  /**
  * The function returns the model of the robot associated to the current thread.
  * @return The model of the robot.
  */
  std::string getModelName() const;

  /**
  * The function returns the simulation time.
  * @return The pseudo-time in milliseconds.
  */
  unsigned getTime() const;

protected:
  const char* robotName; /**< The name of the robot currently constructed. */
  std::list<Robot*> robots; /**< The list of all robots. */
  int simStepLength; /**< The length of one simulation step (in ms) */
  bool simTime; /**< Switches between simulation time mode and real time mode. */
  bool dragTime; /**< Drag simulation to avoid running faster then realtime. */
  int time; /**< The simulation time. */
  unsigned lastTime; /**< The last time execute was called. */
  std::string statusText; /**< The text to be printed in the status bar. */
  bool putBallBack; /**< Put a ball that leaves the field back. */
  Oracle::BallOut ballOut; /**< State of the current ball out of the field. */
  unsigned ballOutTime; /**< Time which is used for waiting in states when ball went out */

  /**
  * Destructor.
  */
  virtual ~RoboCupCtrl();

  /**
  * The function is called to initialize the module.
  */
  virtual bool compile();

  /**
  * The function is called in each simulation step.
  */
  virtual void update();

  /**
  * Has to be called by derived class to start processes.
  */
  void start();

  /**
  * Has to be called by derived class to stop processes.
  */
  void stop();

private:
  QList<SimRobot::Object*> views; /**< List of registered views */
};
