/**
* @file SimObjectWidget.cpp
* Implementation of class SimObjectWidget
* @author Colin Graf
*/

#include <QMouseEvent>
#include <QApplication>
#include <QSettings>
#include <QMenu>
#include <QSignalMapper>
#include <QClipboard>
#include <QFileDialog>

#include "SimObjectWidget.h"
#include "CoreModule.h"
#include "Platform/Assert.h"
#include "Simulation/Simulation.h"
#include "Simulation/SimObject.h"

SimObjectWidget::SimObjectWidget(SimObject& simObject) : QGLWidget(QGLFormat(QGL::SampleBuffers | QGL::AccumBuffer), 0, Simulation::simulation->renderer.getWidget()), 
  object(simObject), objectRenderer(simObject),
  wkey(false), akey(false), skey(false), dkey(false),
  dragAndDropMenu(0), cameraMenu(0), openingAngleMenu(0), appearanceMenu(0), physicsMenu(0), drawingsMenu(0),
  flagsSignalMapper(0), showSensorsAct(0)
{
  setFocusPolicy(Qt::StrongFocus);

  // load layout settings
  QSettings* settings = &CoreModule::application->getLayoutSettings();
  settings->beginGroup(object.getFullName());

  objectRenderer.setSurfaceShadeMode(SimRobotCore2::Renderer::ShadeMode(settings->value("SurfaceShadeMode", int(objectRenderer.getSurfaceShadeMode())).toInt()));
  objectRenderer.setPhysicsShadeMode(SimRobotCore2::Renderer::ShadeMode(settings->value("PhysicsShadeMode", int(objectRenderer.getPhysicsShadeMode())).toInt()));
  objectRenderer.setDrawingsShadeMode(SimRobotCore2::Renderer::ShadeMode(settings->value("DrawingsShadeMode", int(objectRenderer.getDrawingsShadeMode())).toInt()));
  objectRenderer.setCameraMode(SimRobotCore2::Renderer::CameraMode(settings->value("CameraMode", int(objectRenderer.getCameraMode())).toInt()));
  fovy = settings->value("FovY", objectRenderer.getFovY()).toInt();
  objectRenderer.setDragPlane(SimRobotCore2::Renderer::DragAndDropPlane(settings->value("DragPlane", int(objectRenderer.getDragPlane())).toInt()));
  objectRenderer.setDragMode(SimRobotCore2::Renderer::DragAndDropMode(settings->value("DragMode", int(objectRenderer.getDragMode())).toInt()));
  objectRenderer.setRenderFlags(settings->value("RenderFlags", objectRenderer.getRenderFlags()).toInt());

  float pos[3];
  float target[3];
  objectRenderer.getCamera(pos, target);
  pos[0] = settings->value("cameraPosX", pos[0]).toDouble();
  pos[1] = settings->value("cameraPosY", pos[1]).toDouble();
  pos[2] = settings->value("cameraPosZ", pos[2]).toDouble();
  target[0] = settings->value("cameraTargetX", target[0]).toDouble();
  target[1] = settings->value("cameraTargetY", target[1]).toDouble();
  target[2] = settings->value("cameraTargetZ", target[2]).toDouble();
  objectRenderer.setCamera(pos, target);

  settings->endGroup();
}

SimObjectWidget::~SimObjectWidget()
{
  // save layout settings
  QSettings* settings = &CoreModule::application->getLayoutSettings();
  settings->beginGroup(object.getFullName());

  settings->setValue("SurfaceShadeMode", int(objectRenderer.getSurfaceShadeMode()));
  settings->setValue("PhysicsShadeMode", int(objectRenderer.getPhysicsShadeMode()));
  settings->setValue("DrawingsShadeMode", int(objectRenderer.getDrawingsShadeMode()));
  settings->setValue("CameraMode", int(objectRenderer.getCameraMode()));
  settings->setValue("FovY", objectRenderer.getFovY());
  settings->setValue("DragPlane", int(objectRenderer.getDragPlane()));
  settings->setValue("DragMode", int(objectRenderer.getDragMode()));
  settings->setValue("RenderFlags", objectRenderer.getRenderFlags());

  float pos[3];
  float target[3];
  objectRenderer.getCamera(pos, target);

  settings->setValue("cameraPosX", pos[0]);
  settings->setValue("cameraPosY", pos[1]);
  settings->setValue("cameraPosZ", pos[2]);
  settings->setValue("cameraTargetX", target[0]);
  settings->setValue("cameraTargetY", target[1]);
  settings->setValue("cameraTargetZ", target[2]);

  settings->endGroup();
}

void SimObjectWidget::initializeGL()
{
  objectRenderer.init(isSharing());
}

void SimObjectWidget::paintGL()
{
  objectRenderer.draw();
}

void SimObjectWidget::resizeGL(int width, int height)
{
  objectRenderer.resize(fovy, width, height);
}

void SimObjectWidget::paintEvent(QPaintEvent* event)
{
  QGLWidget::paintEvent(event);
}

void SimObjectWidget::mouseMoveEvent(QMouseEvent* event)
{
  QGLWidget::mouseMoveEvent(event);

  if(objectRenderer.moveDrag(event->x(), event->y()))
  {
    event->accept();
    updateGL();
  }
}

void SimObjectWidget::mousePressEvent(QMouseEvent* event)
{
  QGLWidget::mousePressEvent(event);

  if(event->button() == Qt::LeftButton || event->button() == Qt::MidButton)
    if(objectRenderer.startDrag(event->x(), event->y(), QApplication::keyboardModifiers() & Qt::ShiftModifier ? SimObjectRenderer::DRAG_ROTATE : SimObjectRenderer::DRAG_NORMAL))
    {
      event->accept();
      updateGL();
    }
}

void SimObjectWidget::mouseReleaseEvent(QMouseEvent* event)
{
  QGLWidget::mouseReleaseEvent(event);

  if(objectRenderer.releaseDrag(event->x(), event->y()))
  {
    event->accept();
    updateGL();
  }
}

void SimObjectWidget::mouseDoubleClickEvent(QMouseEvent* event)
{
  QGLWidget::mouseDoubleClickEvent(event);

  if(event->button() == Qt::LeftButton)
  {
    SimRobotCore2::Object* selectedObject = objectRenderer.getDragSelection();
    if(selectedObject)
        CoreModule::application->selectObject(*selectedObject);
  }
}

void SimObjectWidget::keyPressEvent(QKeyEvent* event)
{
  if(event->modifiers() != 0)
  {
    QGLWidget::keyPressEvent(event);
    return;
  }

  switch(event->key())
  {
  case Qt::Key_PageUp:
  case Qt::Key_Plus:
    event->accept();
    objectRenderer.zoom(-100.);
    updateGL();
    break;

  case Qt::Key_PageDown:
  case Qt::Key_Minus:
    event->accept();
    objectRenderer.zoom(100.);
    updateGL();
    break;

  case Qt::Key_W:
  case Qt::Key_A:
  case Qt::Key_S:
  case Qt::Key_D:
    event->accept();
    switch(event->key())
    {
    case Qt::Key_W: wkey = true; break;
    case Qt::Key_A: akey = true; break;
    case Qt::Key_S: skey = true; break;
    case Qt::Key_D: dkey = true; break;
    default: break;
    }
    objectRenderer.setCameraMove(akey, dkey, wkey, skey);
    updateGL();
    break;

  default:
    QGLWidget::keyPressEvent(event);
    break;
  }
}

void SimObjectWidget::keyReleaseEvent(QKeyEvent* event)
{
  if(event->modifiers() != 0)
  {
    QGLWidget::keyPressEvent(event);
    return;
  }

  switch(event->key())
  {
  case Qt::Key_W:
  case Qt::Key_A:
  case Qt::Key_S:
  case Qt::Key_D:
    event->accept();
    if(!event->isAutoRepeat())
    {
      switch(event->key())
      {
      case Qt::Key_W: wkey = false; break;
      case Qt::Key_A: akey = false; break;
      case Qt::Key_S: skey = false; break;
      case Qt::Key_D: dkey = false; break;
      default: break;
      }
      objectRenderer.setCameraMove(akey, dkey, wkey, skey);
    }
    break;

  default:
    QGLWidget::keyReleaseEvent(event);
    break;
  }
}

void SimObjectWidget::wheelEvent(QWheelEvent* event)
{
  if(event->orientation() == Qt::Vertical)
  {
    objectRenderer.zoom(event->delta());
    updateGL();
    event->accept();
  }
  else
    QGLWidget::wheelEvent(event);
}

void SimObjectWidget::update()
{
  QGLWidget::update();
}

QMenu* SimObjectWidget::createEditMenu()
{
  QMenu* menu = new QMenu(tr("&Edit"));
  
  QAction* action = menu->addAction(QIcon(":/Icons/page_copy.png"), tr("&Copy"));
  action->setShortcut(QKeySequence(QKeySequence::Copy));
  action->setStatusTip(tr("Copy the rendered object to the clipboard"));
  connect(action, SIGNAL(triggered()), this, SLOT(copy()));
  
  return menu;
}

QMenu* SimObjectWidget::createUserMenu()
{
  QMenu* menu = new QMenu(tr("&Object"));
  QWidget* owner = 
#ifdef MACOSX  
  menu;
#else
  this;

  if(!dragAndDropMenu)
#endif
  {
    QMenu* subMenu = dragAndDropMenu = new QMenu(tr("&Drag and Drop"), owner);
    QAction* action = subMenu->menuAction();
    action->setIcon(QIcon(":/Icons/DragPlane.png"));
    action->setStatusTip(tr("Select the drag and drop dynamics mode and plane along which operations are performed"));
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setDragPlane(int)));
    action = subMenu->addAction(tr("X/Y Plane"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::XY_PLANE);
    action->setShortcut(QKeySequence(Qt::Key_Z));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("X/Z Plane"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::XZ_PLANE);
    action->setShortcut(QKeySequence(Qt::Key_Y));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("Y/Z Plane"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::YZ_PLANE);
    action->setShortcut(QKeySequence(Qt::Key_X));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getDragPlane()))->setChecked(true);
    subMenu->addSeparator();
    actionGroup = new QActionGroup(subMenu);
    signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setDragMode(int)));
    action = subMenu->addAction(tr("&Keep Dynamics"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::KEEP_DYNAMICS);
    action->setShortcut(QKeySequence(Qt::Key_8));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Reset Dynamics"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::RESET_DYNAMICS);
    action->setShortcut(QKeySequence(Qt::Key_9));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Apply Dynamics"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::APPLY_DYNAMICS);
    action->setShortcut(QKeySequence(Qt::Key_0));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getDragMode()))->setChecked(true);
  }
  menu->addMenu(dragAndDropMenu);

  menu->addSeparator();

#ifndef MACOSX  
  if(!cameraMenu)
#endif
  {
    QMenu* subMenu = cameraMenu = new QMenu(tr("&Camera"), owner);
    QAction* action = subMenu->menuAction();
    action->setIcon(QIcon(":/Icons/camera.png"));
    action->setStatusTip(tr("Select different camera modes for displaying the scene"));
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setCameraMode(int)));
    action = subMenu->addAction(tr("&Target Mode"));
    action->setCheckable(true);
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::TARGETCAM);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Free Mode"));
    action->setCheckable(true);
    actionGroup->addAction(action);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::FREECAM);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));

    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getCameraMode()))->setChecked(true);
    subMenu->addSeparator();
    action = subMenu->addAction(tr("&Reset"));
    action->setShortcut(QKeySequence(Qt::Key_R));
    connect(action, SIGNAL(triggered()), this, SLOT(resetCamera()));
    action = subMenu->addAction(tr("&Toggle"));
    action->setShortcut(QKeySequence(Qt::Key_T));
    connect(action, SIGNAL(triggered()), this, SLOT(toggleCameraMode()));

    /*
    action = subMenu->addAction(tr("&Fit"));
    action->setShortcut(QKeySequence(Qt::Key_F));
    connect(action, SIGNAL(triggered()), this, SLOT(fitCamera()));
    subMenu->addSeparator();
    */
  }
  menu->addMenu(cameraMenu);
  
#ifndef MACOSX  
  if(!openingAngleMenu)
#endif
  {
    QMenu* subMenu = openingAngleMenu = new QMenu(tr("&Vertical Opening Angle"), owner);
    QAction* action = subMenu->menuAction();
    action->setIcon(QIcon(":/Icons/opening_angle.png"));
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setFovY(int)));
    action = subMenu->addAction(tr("&20\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 20);
    action->setShortcut(QKeySequence(Qt::Key_1));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&40\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 40);
    action->setShortcut(QKeySequence(Qt::Key_2));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&60\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 60);
    action->setShortcut(QKeySequence(Qt::Key_3));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&80\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 80);
    action->setShortcut(QKeySequence(Qt::Key_4));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("100\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 100);
    action->setShortcut(QKeySequence(Qt::Key_5));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("120\xb0"));
    actionGroup->addAction(action);
    signalMapper->setMapping(action, 120);
    action->setShortcut(QKeySequence(Qt::Key_6));
    action->setCheckable(true);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getFovY()))->setChecked(true);
  }
  menu->addMenu(openingAngleMenu);
  
  menu->addSeparator();

#ifndef MACOSX  
  if(!appearanceMenu)
#endif
  {
    QMenu* subMenu = appearanceMenu = new QMenu(tr("&Appearances Rendering"), owner);
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setSurfaceShadeMode(int)));
    QAction* action = subMenu->menuAction();
    action->setIcon(QIcon(":/Icons/layers.png"));
    action->setStatusTip(tr("Select different shading techniques for displaying the scene"));
    action = subMenu->addAction(tr("&Off"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::OFF);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Wire Frame"));
    actionGroup->addAction(action);
    action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::WIREFRAME);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Flat Shading"));
    actionGroup->addAction(action);
    action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_F));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::FLAT);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Smooth Shading"));
    actionGroup->addAction(action);
    action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_M));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::SMOOTH);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getSurfaceShadeMode()))->setChecked(true);
  }
  menu->addMenu(appearanceMenu);

#ifndef MACOSX  
  if(!physicsMenu)
#endif
  {
    QMenu* subMenu = physicsMenu = new QMenu(tr("&Physics Rendering"), owner);
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setPhysicsShadeMode(int)));
    QAction* action = subMenu->menuAction();
    //action->setIcon(QIcon(":/Icons/layers.png"));
    action->setStatusTip(tr("Select different shading techniques for displaying the physical representation of objects"));
    action = subMenu->addAction(tr("&Off"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::OFF);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Wire Frame"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::WIREFRAME);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Flat Shading"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_F));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::FLAT);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Smooth Shading"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_M));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::SMOOTH);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getPhysicsShadeMode()))->setChecked(true);
  }
  menu->addMenu(physicsMenu);

#ifndef MACOSX  
  if(!drawingsMenu)
#endif
  {
    QMenu* subMenu = drawingsMenu = new QMenu(tr("&Drawings Rendering"), owner);
    QActionGroup* actionGroup = new QActionGroup(subMenu);
    QSignalMapper* signalMapper = new QSignalMapper(subMenu);
    connect(signalMapper, SIGNAL(mapped(int)), SLOT(setDrawingsShadeMode(int)));
    QAction* action = subMenu->menuAction();
    action->setIcon(QIcon(":/Icons/chart_line.png"));
    action->setStatusTip(tr("Select different shading techniques for displaying controller drawings"));
    action = subMenu->addAction(tr("&Off"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::OFF);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Wire Frame"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_W));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::WIREFRAME);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Flat Shading"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_F));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::FLAT);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    action = subMenu->addAction(tr("&Smooth Shading"));
    actionGroup->addAction(action);
    //action->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_M));
    action->setCheckable(true);
    signalMapper->setMapping(action, SimRobotCore2::Renderer::SMOOTH);
    connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
    qobject_cast<QAction*>(signalMapper->mapping(objectRenderer.getDrawingsShadeMode()))->setChecked(true);
  }
  menu->addMenu(drawingsMenu);

  menu->addSeparator();

  QSignalMapper* signalMapper = new QSignalMapper(this);
  connect(signalMapper, SIGNAL(mapped(int)), SLOT(toggleRenderFlag(int)));

  QAction* action = menu->addAction(tr("Enable &Lights"));
  action->setStatusTip(tr("Enable lightning"));
  action->setCheckable(true);
  action->setChecked(objectRenderer.getRenderFlags() & SimRobotCore2::Renderer::enableLights);
  signalMapper->setMapping(action, SimRobotCore2::Renderer::enableLights);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));

  action = menu->addAction(tr("Enable &Textures"));
  action->setStatusTip(tr("Enable textures"));
  action->setCheckable(true);
  action->setChecked(objectRenderer.getRenderFlags() & SimRobotCore2::Renderer::enableTextures);
  signalMapper->setMapping(action, SimRobotCore2::Renderer::enableTextures);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));

  action = menu->addAction(tr("Enable &Multisample"));
  action->setStatusTip(tr("Enable multisampling"));
  action->setCheckable(true);
  action->setChecked(objectRenderer.getRenderFlags() & SimRobotCore2::Renderer::enableMultisample);
  signalMapper->setMapping(action, SimRobotCore2::Renderer::enableMultisample);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));

  menu->addSeparator();

  action = menu->addAction(tr("Show &Coordinate System"));
  action->setStatusTip(tr("Show the coordinate system of the displayed object"));
  action->setCheckable(true);
  action->setChecked(objectRenderer.getRenderFlags() & SimRobotCore2::Renderer::showCoordinateSystem);
  signalMapper->setMapping(action, SimRobotCore2::Renderer::showCoordinateSystem);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));

  if(!flagsSignalMapper)
  {
    ASSERT(showSensorsAct == 0);

    flagsSignalMapper = new QSignalMapper(this);
    connect(flagsSignalMapper, SIGNAL(mapped(int)), SLOT(toggleRenderFlag(int)));

    showSensorsAct = new QAction(QIcon(":/Icons/transmit_go.png"), tr("Show &Sensors"), this);
    showSensorsAct->setStatusTip(tr("Show the values of the sensors in the scene view"));
    showSensorsAct->setCheckable(true);
    showSensorsAct->setChecked(objectRenderer.getRenderFlags() & SimRobotCore2::Renderer::showSensors);
    flagsSignalMapper->setMapping(showSensorsAct, SimRobotCore2::Renderer::showSensors);
    connect(showSensorsAct, SIGNAL(triggered()), flagsSignalMapper, SLOT(map()));
  }
  menu->addAction(showSensorsAct);

/*
  menu->addSeparator();

  subMenu = menu->addMenu(tr("Export as Image..."));
  signalMapper = new QSignalMapper(subMenu);
  connect(signalMapper, SIGNAL(mapped(int)), SLOT(exportAsImage(int)));
  action = subMenu->addAction(tr("3840x2160"));
  signalMapper->setMapping(action, (3840 << 16) | 2160);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
  action = subMenu->addAction(tr("1920x1080"));
  signalMapper->setMapping(action, (1920 << 16) | 1080);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
  action = subMenu->addAction(tr("1280x1024"));
  signalMapper->setMapping(action, (1280 << 16) | 1024);
  connect(action, SIGNAL(triggered()), signalMapper, SLOT(map()));
*/
  return menu;
}
void SimObjectWidget::copy()
{
  QApplication::clipboard()->clear();
  QApplication::clipboard()->setPixmap(QPixmap::fromImage(grabFrameBuffer()));
}

void SimObjectWidget::exportAsImage(int resolution)
{
  /*
  int width = resolution >> 16;
  int height = resolution & 0xffff;

  QSettings& settings = Core2Module::application->getSettings();
  QString fileName = QFileDialog::getSaveFileName(this, 
    tr("Export as Image"), settings.value("ExportDirectory", "").toString(), tr("Portable Network Graphic (*.png)"));
  if(fileName.isEmpty())
    return;
  settings.setValue("ExportDirectory", QFileInfo(fileName).dir().path());

  QImage image;
  {
    unsigned int winWidth, winHeight;
    objectRenderer.getSize(winWidth, winHeight);

    // render object using a temporary widget
    QGLFormat format;
    format.setDoubleBuffer(false);
    QGLWidget widget(format);
    widget.setMaximumSize(width, height);
    widget.resize(width, height);
    widget.makeCurrent();
    objectRenderer.setSize(width, height);
    objectRenderer.init(false);
    objectRenderer.draw();
    image = widget.grabFrameBuffer();

    objectRenderer.setSize(winWidth, winHeight);
  }

  image.save(fileName);
  */
}

void SimObjectWidget::setSurfaceShadeMode(int style)
{
  objectRenderer.setSurfaceShadeMode(SimRobotCore2::Renderer::ShadeMode(style));
  updateGL();
}

void SimObjectWidget::setPhysicsShadeMode(int style)
{
  objectRenderer.setPhysicsShadeMode(SimRobotCore2::Renderer::ShadeMode(style));
  updateGL();
}

void SimObjectWidget::setDrawingsShadeMode(int style)
{
  objectRenderer.setDrawingsShadeMode(SimRobotCore2::Renderer::ShadeMode(style));
  updateGL();
}

void SimObjectWidget::setCameraMode(int mode)
{
  objectRenderer.setCameraMode(SimRobotCore2::Renderer::CameraMode(mode));
  updateGL();
}

void SimObjectWidget::setFovY(int fovy)
{
  unsigned int width, height;
  objectRenderer.getSize(width, height);
  makeCurrent();
  objectRenderer.resize(fovy, width, height);
  updateGL();
}

void SimObjectWidget::setDragPlane(int plane)
{
  objectRenderer.setDragPlane(SimRobotCore2::Renderer::DragAndDropPlane(plane));
  updateGL();
}

void SimObjectWidget::setDragMode(int mode)
{
  objectRenderer.setDragMode(SimRobotCore2::Renderer::DragAndDropMode(mode));
  updateGL();
}

void SimObjectWidget::resetCamera()
{
  objectRenderer.resetCamera();
  updateGL();
}

void SimObjectWidget::toggleCameraMode()
{
  objectRenderer.toggleCameraMode();
  updateGL();
}

void SimObjectWidget::fitCamera()
{
  /*
  objectRenderer.fitCamera();
  updateGL();
  */
}

void SimObjectWidget::toggleRenderFlag(int flag)
{
  unsigned int flags = objectRenderer.getRenderFlags();
  if(flags & flag)
    flags &= ~flag;
  else
    flags |= flag;
  objectRenderer.setRenderFlags(flags);

  updateGL();
}
