#include <QFormLayout>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPalette>
#include <QPushButton>
#include <QResizeEvent>
#include <QScrollArea>
#include <QTextDocument>
#include <QScrollBar>
#include <QtCore>
#include <QApplication>
#include "Utils/bush/cmdlib/Context.h"
#include "Utils/bush/tools/StringTools.h"
#include "Utils/bush/ui/CommandLineEdit.h"
#include "Utils/bush/ui/Console.h"
#include "Utils/bush/ui/TeamSelector.h"

VisualContext::VisualContext(QWidget *parent)
  : QFrame(parent),
    entries(),
    widgets(),
    formLayout(new QFormLayout()),
    nl(true)
{
  if (parent && parent->inherits("VisualContext"))
    setFrameStyle(QFrame::Box);
  formLayout->setSpacing(1);
  formLayout->setContentsMargins(1, 1, 1, 1);
  setLayout(formLayout);
  setAutoFillBackground(true);
  QPalette p = palette();
  p.setColor(QPalette::Background, p.color(QPalette::AlternateBase));
  setPalette(p);
}

VisualContext::Entry::~Entry()
{
  // do not delete context. It takes care of itself.
  if (text) delete text;
}

static inline VisualContext::Entry::Type targetToType(ConsolePrintTarget target)
{
  return target == CPT_PRINT || target == CPT_PRINT_LINE ? VisualContext::Entry::TEXT_OUTPUT : VisualContext::Entry::TEXT_ERROR;
}

void VisualContext::updateWidget(size_t index, Entry *entry)
{
  QWidget *widget = widgets[index];
  if (widget->inherits("QLabel"))
  {
    QLabel *label = dynamic_cast<QLabel*>(widget);
    QString text;
    if (entry->type == Entry::TEXT_ERROR)
      text = "<font color='red'>" + Qt::convertFromPlainText(*entry->text) + "</font>";
    else
      text = Qt::convertFromPlainText(*entry->text);
    label->setText(text);
    label->setMaximumHeight(label->sizeHint().height());
    updateMinHeight();
  }
}

void VisualContext::addWidget(Entry *entry, const QString &commandLine)
{
  QWidget *widget;
  if (entry->type == Entry::CONTEXT)
  {
    VisualContextDecoration *vd = new VisualContextDecoration(commandLine, this, entry->context);
    connect(vd, SIGNAL(minHeightChanged()), this, SLOT(updateMinHeight()));
    widget = vd;
  }
  else
  {
    QString text;
    if (entry->type == Entry::TEXT_ERROR)
      text = "<font color='red'>" + Qt::convertFromPlainText(*entry->text) + "</font>";
    else
      text = Qt::convertFromPlainText(*entry->text);

    QLabel *label = new QLabel(text, this);
    label->setTextInteractionFlags(label->textInteractionFlags() | Qt::TextSelectableByMouse);
    label->setMargin(1);
    widget = label;
    widget->setMinimumHeight(widget->sizeHint().height());
  }
  QFormLayout *l = getLayout();
  l->addRow(widget);
  widget->setAutoFillBackground(true);
  widgets << widget;
  updateMinHeight();
}

void VisualContext::doPrint(ConsolePrintTarget target, const QString &msg)
{
  Entry::Type currentType = targetToType(target);
  Entry *lastEntry = entries.empty() ? 0 : entries.last();
  if (lastEntry && !nl && lastEntry->type == currentType)
  {
      lastEntry->text->append(msg);
      updateWidget(entries.size() - 1, lastEntry);
  }
  else
  {
    Entry *entry = new Entry(currentType, msg);
    entries << entry;
    addWidget(entry);
  }
  nl = target == CPT_ERROR_LINE || target == CPT_PRINT_LINE;
}

void VisualContext::commandExecuted(Context *context, const QString &cmdLine)
{
  VisualContext *sub = new VisualContext(this);

  connect(context, SIGNAL(sExecute(Context *, const QString &)),
          sub, SLOT(commandExecuted(Context *, const QString &)),
          Qt::BlockingQueuedConnection);
  connect(context, SIGNAL(sPrint(ConsolePrintTarget, const QString &)),
          sub, SLOT(doPrint(ConsolePrintTarget, const QString &)));
  connect(context, SIGNAL(sFinished(bool)), sub, SLOT(commandFinished(bool)));
  connect(context, SIGNAL(sCancelFinished()), sub, SLOT(commandCanceled()));
  connect(sub, SIGNAL(sCancel()), context, SLOT(cancel()), Qt::DirectConnection);

  Entry *entry = new Entry(sub);
  entries << entry;
  addWidget(entry, cmdLine);
}

void VisualContext::commandFinished(bool status)
{
  emit statusChanged(status);
}

void VisualContext::commandCanceled()
{
  emit sCanceled();
}

void VisualContext::cancel()
{
  emit sCancel();
}

void VisualContext::executeInContext(Console *console, TeamSelector *teamSelector, const QString &cmdLine)
{
  Context *context = new Context(teamSelector->getSelectedRobots(),
                                 teamSelector->getSelectedTeam());

  /* Important: Use BlockingQueuedConnection so that the newly created thread
   * cannot emits all its print signals before they can be handled. */
  connect(context, SIGNAL(sExecute(Context *, const QString &)),
          this, SLOT(commandExecuted(Context *, const QString &)),
          Qt::BlockingQueuedConnection);
  connect(context, SIGNAL(sCancelFinished()), this, SLOT(commandCanceled()));

  /* Don't know if this signal is emitted a bit too often but this assures that
   * the newest output of the current command is always visible. */
  connect(this, SIGNAL(minHeightChanged()), console, SLOT(scroll()), Qt::UniqueConnection);
  connect(this, SIGNAL(sCancel()), context, SLOT(cancel()), Qt::DirectConnection);
  context->execute(toString(cmdLine));

  /* Do not forget to disconnect the signal so that a new command can trigger
   * the scrolling. */
  disconnect(this, SIGNAL(minHeightChanged()));
  disconnect(context);

  bool doQuit = context->isShudown();
  // Since the context is a QObject it is better to let Qt do the cleanup
  context->deleteLater();

  if (doQuit)
    QApplication::quit();
}

void VisualContext::updateMinHeight()
{
  int height = 0;
  for (int i = 0; i < widgets.size(); ++i)
    height += widgets[i]->sizeHint().height() + formLayout->verticalSpacing();
  if (widgets.size() == 1)
    height += 5; // <- HACK (does not work really)
  setMinimumHeight(height);
  emit minHeightChanged();
}

Icons Icons::theIcons;

void Icons::init()
{
  ICON_GRAY = QIcon(":icons/gray.png");
  ICON_GREEN = QIcon(":icons/green.png");
  ICON_ORANGE = QIcon(":icons/orange.png");
  ICON_RED = QIcon(":icons/red.png");
}

VisualContextDecoration::VisualContextDecoration(const QString &commandLine, VisualContext *parent, VisualContext *context)
  : QFrame(parent),
    button(new QPushButton(Icons::getInstance().ICON_GRAY, "", this)),
    header(new QLabel(commandLine)),
    visualContext(context),
    parentContext(parent)
{
  setAutoFillBackground(true);
  QPalette p = palette();
  p.setColor(QPalette::Background, p.color(QPalette::AlternateBase));
  setPalette(p);
  QFormLayout *layout = new QFormLayout();
  layout->setSpacing(3);
  header->setFrameStyle(QFrame::Box);
  layout->addRow(button, header);
  button->setMaximumWidth(25);
  button->setFlat(true);
  button->setCheckable(true);
  button->setChecked(true);
  layout->addRow(visualContext);
  setLayout(layout);

  connect(visualContext, SIGNAL(minHeightChanged()), this, SLOT(updateMinHeight()));
  connect(visualContext, SIGNAL(statusChanged(bool)), this, SLOT(updateStatus(bool)));
  connect(visualContext, SIGNAL(sCanceled()), this, SLOT(canceled()));
}

void VisualContextDecoration::updateMinHeight()
{
  int height = header->minimumSizeHint().height()
             + visualContext->minimumSizeHint().height()
             + layout()->spacing() * 4;
  setMinimumHeight(height);
  emit minHeightChanged();
}

void VisualContextDecoration::updateStatus(bool status)
{
  if (button->icon().cacheKey() != Icons::getInstance().ICON_ORANGE.cacheKey())
  {
    if (status)
      button->setIcon(Icons::getInstance().ICON_GREEN);
    else
      button->setIcon(Icons::getInstance().ICON_RED);
  }
}

void VisualContextDecoration::canceled()
{
  button->setIcon(Icons::getInstance().ICON_ORANGE);
}

Console::Console(TeamSelector *teamSelector)
  : visualContext(new VisualContext(this)),
    teamSelector(teamSelector),
    scrollArea(new QScrollArea(this)),
    prompt(0),
    cmdLine(0),
    scrollEnabled(true)
{
  cmdLine = new CommandLineEdit(this);

  prompt = new QLabel("bush>", cmdLine);
  prompt->setAutoFillBackground(true);
  QPalette p = prompt->palette();
  p.setColor(QPalette::Background, p.color(QPalette::AlternateBase));
  prompt->setPalette(p);

  QGridLayout *layout = new QGridLayout();
  layout->setHorizontalSpacing(0);
  scrollArea->setWidget(visualContext);
  scrollArea->setWidgetResizable(true);
  scrollArea->setBackgroundRole(QPalette::AlternateBase);
  connect(scrollArea->verticalScrollBar(), SIGNAL(valueChanged(int)), this, SLOT(updateScrollEndabled()));
  layout->addWidget(scrollArea, 0, 0);
  layout->setRowStretch(0, 1);
  QFormLayout *fl = new QFormLayout();
  fl->setFieldGrowthPolicy(QFormLayout::AllNonFixedFieldsGrow);
  fl->addRow(prompt, cmdLine);
  layout->addLayout(fl, 1, 0);
  setLayout(layout);

  connect(cmdLine, SIGNAL(returnPressed()), this, SLOT(returnPressed()));
}

void Console::returnPressed()
{
  QString cmdString = cmdLine->text();
  cmdLine->setText("");
  if (cmdString.size() > 0)
  {
    QtConcurrent::run(visualContext, &VisualContext::executeInContext, this, teamSelector, cmdString);
    cmdLine->setFocus();
  }
}

void Console::scroll()
{
  if (scrollEnabled)
    scrollArea->ensureVisible(0, visualContext->size().height());
}

void Console::updateScrollEndabled()
{
  QScrollBar *scrollBar = scrollArea->verticalScrollBar();
  scrollEnabled = scrollBar->value() == scrollBar->maximum();
}

void Console::showEvent(QShowEvent *event)
{
  cmdLine->setFocus();
  QFrame::showEvent(event);
}

void Console::setCommand(const std::string &command)
{
  cmdLine->setText(fromString(command));
  cmdLine->setFocus();
}

void Console::cancel()
{
  visualContext->cancel();
}
