/**
 * The class represents an appearance. A appearance specifies the visual properties
 * of geometries.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a> 
 */
class Appearance extends Node {
    /** The material. */
    Material material;
    
    /** The image texture. */
    ImageTexture texture;
    
    /** The counter to build unique names. */
    private static int count = 0;

    /**
     * The constructor.
     * 
     * @param name The appearance identifier name.
     * @param parent The reference to the parent object.
     */
    Appearance(String name, Node parent) {
        super(name, parent);
        this.appearance.add(this);
        name = name != null ? name : "APP_" + count++;
    }

    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        generateName("APP" + count ++);
    }

    /**
     * The method generates a reference to a specific appearance.
     * 
     * @param code The code output stream.
     * @param name The referenced appearance identifier name.
     */
    static public void generateCodeAppearanceRef(TreeStream code, String name) {
        code.indent();
        code.println("<Appearance ref=\"" + name + "\" />");
        code.unindent();
    }

    /**
     * The method generates a reference to this appearance.
     * 
     * @param code The code output stream.
     */
    public void generateCodeAppearanceRef(TreeStream code) {
        code.indent();
        code.println("<Appearance ref=\"" + name + "\" />");
        code.unindent();
    }

    /**
     * The method generates the appearance definition code for the blue and the red robot.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        boolean isLogo = texture != null && texture.url.contains("blue"),
            isBlue = material != null && material.diffuseColor != null && material.diffuseColor.length == 3 && material.diffuseColor[0] == 0 && material.diffuseColor[1] == 0 && material.diffuseColor[2] == 1;
        String nameRed = name;
        code.indent();

        if(isBlue || isLogo) {
            for(Node p = parent; p != null; p = p.parent)
                if(p instanceof Robot)
                    name = p.name;
            nameRed = name + "-red";
            name += "-blue";
            if(isLogo) {
                nameRed += "-logo";
                name += "-logo";
            }
        }

        code.println("<AppearanceDefinition name=\"" + name + "\">");
        if(material != null)
            material.generateCode(code);
        if(texture != null)
            texture.generateCode(code);
        code.println("</AppearanceDefinition>\n");

        if(isBlue || isLogo) {
            code.println("<AppearanceDefinition name=\"" + nameRed + "\">");
            if(material != null) {
                if(isBlue) {
                    material.diffuseColor[0] = 1;
                    material.diffuseColor[2] = 0;
                }
                material.generateCode(code);
            }
            if(texture != null) {
                if(isLogo)
                    texture.url = texture.url.replace("blue", "red");
                texture.generateCode(code);
            }
            code.println("</AppearanceDefinition>\n");
        }
        code.unindent();
    }

    /**
     * The method generates the appearance definition code for the blue and the red robot.
     * 
     * @param code The code output stream.
     */
    public void generateRoSi2Code(TreeStream code) {
        boolean isLogo = texture != null && texture.url.contains("blue"),
            isBlue = material != null && material.diffuseColor != null && material.diffuseColor.length == 3 && material.diffuseColor[0] == 0 && material.diffuseColor[1] == 0 && material.diffuseColor[2] == 1;
        String nameRed = name;
        code.indent();

        if(isBlue || isLogo) {
            for(Node p = parent; p != null; p = p.parent)
                if(p instanceof Robot)
                    name = p.name;
            nameRed = name + "-red";
            name += "-blue";
            if(isLogo) {
                nameRed += "-logo";
                name += "-logo";
            }
        }

        code.println("<Surface name=\"" + name + "\"" +(material != null ? material.getRoSi2Colors() : "")+ "/>\n");
        //if(material != null)
           // material.generateRoSi2Code(code);
        //if(texture != null)
          //  texture.generateRoSi2Code(code);
        //code.println("</Surface>\n");

        if(isBlue || isLogo) {
            
            if(material != null) {
                if(isBlue) {
                    material.diffuseColor[0] = 1;
                    material.diffuseColor[2] = 0;
                }
                //material.generateRoSi2Code(code);
            }
            code.println("<Surface name=\"" + nameRed + "\"" +(material != null ? material.getRoSi2Colors() : "")+ "/>\n");
            /*
            if(texture != null) {
                if(isLogo)
                    texture.url = texture.url.replace("blue", "red");
                texture.generateRoSi2Code(code);
            }
            code.println("</Surface>\n");
            */
        }
        code.unindent();
    }
}
