import java.util.Vector;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;

/**
 * The class represents a set of 3D coordinates to use vertex based geometry.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class Coordinate extends Node
{
    /** The point vector with the sets of points. */
    Vector<double[]> point = new Vector<double[]>();
    
    /** The counter to generate unique names. */
    private static int count = 0;
    
    /** The types of coordinates. */
    enum Type {COORD, TEXCOORD};
    
    /** The coordinate type. */
    Type type;

    /**
     * The constructor.
     * 
     * @param name The coordinate identifier name.
     * @param parent The reference to the parent object.
     * @param type The coordinate type.
     */
    Coordinate(String name, Node parent, Type type) {
        super(name, parent);
        generateName();
        this.type = type;
        coords.addLast(this);
    }

    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        generateName("VERTEX" + count ++);
    }

    /**
     * The method generates a textured or a none textured vertexlist code.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        int vertexCounter = 0;

        code.indent();
        if(type == Type.TEXCOORD) {
            code.println("<VertexListWithTexCoords name=\"" + name + "\">");
            code.indent();
            for(double[] p : point)
                code.println("<Vertex name=\"" + vertexCounter++ + "\" x=\"" + p[0] + "\" y=\"" + p[1] + "\" z=\"" + p[2] + "\" s=\"" + p[3] + "\" t=\"" + p[4] + "\" />");
            code.unindent();
            code.println("</VertexListWithTexCoords>\n");
        } else {
            code.println("<VertexList name=\"" + name + "\">");
            code.indent();
            for(double[] p : point)
                code.println("<Vertex name=\"" + vertexCounter++ + "\" x=\"" + p[0] + "\" y=\"" + p[1] + "\" z=\"" + p[2] + "\" />");
            code.unindent();
            code.println("</VertexList>\n");
        }
        code.unindent();
    }
    
     /**
     * The method generates the initial code for a RoSi2 include file
     * 
     * @param code The code stream object
     */
     public void generateRoSi2Code(TreeStream code) {

        code.indent();
        
        DecimalFormat f = new DecimalFormat("#.#####", DecimalFormatSymbols.getInstance(Locale.US)); 
        
        code.println("<Vertices name=\"" + name + "\">");
        code.pushIndention();
        code.resetIndention();
        for(double[] p : point)
            code.println("" + f.format(p[0]) + " " + f.format(p[1]) + " " + f.format(p[2]));
        code.popIndention();
        code.println("</Vertices>\n");

        if(type == Type.TEXCOORD) {
          code.println("<TexCoords name=\"" + name + "\">");
          code.pushIndention();
          code.resetIndention();
          for(double[] p : point)
              code.println("" + f.format(p[3]) + " " + f.format(p[4]));
          code.popIndention();
          code.println("</TexCoords>\n");
        }
        code.unindent();
     }
}
