/**
 * The class represents a cylinder.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class Cylinder extends Geometry {
    /** The bottom of the cylinder will be displayed. */
    boolean bottom = true;
    
    /** The top of the cylinder will be displayed. */
    boolean top = true;
    
    /** The height of the cylinder. */
    double height = 2;
    
    /** The radius of the cylinder. */
    double radius = 1;
    
    /** The subdevisor */
    int subdivision = 12;
    
    /** The counter to build unique names */
    private static int count = 0;

    /**
     * The constructor.
     * 
     * @param name The cylinder identifier name.
     * @param parent The reference to the parent object.
     */
    Cylinder(String name, Node parent) {
        super(name, parent);
    }

    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        generateName("CYLINDER" + count ++);
    }

    /**
     * The method generates the bounding code with a optional transformation and physics.
     * 
     * @param code The code output stream.
     * @param transform The tranformation for the cylinder.
     * @param physics The physics for the cylinder.
     */
    public void generateCodeBounding(TreeStream code, Transform transform, Physics physics) {
        code.indent();
        if(transform == null && physics == null)
            code.println("<SimpleCylinder name=\"" + name + "\" radius=\"" + radius + "\" height=\"" + height + "\" />");
        else {
            code.println("<SimpleCylinder name=\"" + name + "\" radius=\"" + radius + "\" height=\"" + height + "\">");
            if(transform != null)
                transform.generateCodeTransform(code);
            if(physics != null)
                physics.generateCode(code);
            code.println("</SimpleCylinder>");
        }
        code.unindent();
    }

    /**
     * The method generates the bounding code begin.
     * 
     * @param code The code output stream.
     * @param simple Is it a simple physical object?
     */
    public void generateCodeBoundingStart(TreeStream code, boolean simple) {
        code.indent();
        code.println("<" + (simple ? "Simple" : "") + "Cylinder name=\"" + name + "\" radius=\"" + radius + "\" height=\"" + height + "\" capped=\"false\" canCollide=\"false\">");
    }

    /**
     * The method generates the bounding code closure.
     * 
     * @param code The code output stream.
     * @param simple Is it a simple physical object?
     */
    public void generateCodeBoundingEnd(TreeStream code, boolean simple) {
        code.println("</" + (simple ? "Simple" : "") + "Cylinder>");
        code.unindent();
    }

    /**
     * The method generates a cylinder with a optional transoformation.
     * 
     * @param code The code output stream.
     * @param transform The tranformation for the cylinder.
     */
    public void generateCode(TreeStream code, Transform transform) {
        generateCodeBoundingStart(code, false);

        if(transform != null)
            transform.generateCodeTransform(code);

        if(((Shape)parent).appearance != null)
            ((Shape)parent).appearance.generateCodeAppearanceRef(code);
        else
            Appearance.generateCodeAppearanceRef(code, "STRUCTURE_APP");

        Physics.generateCodePhysicsEmpty(code);

        generateCodeBoundingEnd(code, false);
    }
}
