import java.util.LinkedList;
import java.util.Iterator;

/**
 * The class represents an indexed face set.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class IndexedFaceSet extends Geometry {
    /** The the set coordinates. */
    Coordinate coord;
    
    /** The set of textured coodinates. */
    Coordinate texCoord;
    
    /** The list of coordinates by index. */
    LinkedList<LinkedList<Integer>> coordIndex = new LinkedList<LinkedList<Integer>>();
    
    /** The list of textured coodinates by index. */
    LinkedList<LinkedList<Integer>> texCoordIndex = new LinkedList<LinkedList<Integer>>();
    
    /** The crease angle of the face set. */
    double creaseAngle;

    /** The counter to build unique names. */
    private static int count = 0;
    
    /** The counter to enumerate the triangles. */
    static int triangleCounter = 0;

    /**
     * The constructor.
     * 
     * @param name The indexed face set identifier name.
     * @param parent The reference to the parent object.
     */
    IndexedFaceSet(String name, Node parent) {
        super(name, parent);
    }

    /**
     * The method convertes the texture coordinates to a useful format for RoSiML.
     */
    public void completeTexCoords() {
        for (Iterator<LinkedList<Integer>> i = coordIndex.iterator(), j = texCoordIndex.iterator(); 
                i.hasNext() && j.hasNext();) {
            LinkedList<Integer> li = i.next();
            LinkedList<Integer> lj = j.next();
            for (Iterator<Integer> k = li.iterator(), l = lj.iterator(); 
                    k.hasNext() && l.hasNext();) {
                int lk = k.next();
                int ll = l.next();
                if (texCoord.point.get(ll).length == 2) {
                    double[] c = coord.point.get(lk);
                    double[] t = texCoord.point.get(ll);
                    double[] ct = new double[5];
                    ct[0] = c[0]; ct[1] = c[1]; ct[2] = c[2];
                    ct[3] = t[0]; ct[4] = t[1];
                    texCoord.point.set(ll, ct);
                }
            }
        }
    }
    
    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        if(coord.name != null)
            name = coord.name + "_MACRO" + count ++;
        else
            generateName("MACRO" + count ++);
    }

    /**
     * The method generates the parent shape with the given transformation and this indexed face set in it.
     * 
     * @param code The code output stream.
     * @param transform The tranformation of the shape.
     */
    public void generateCode(TreeStream code, Transform transform) {
        ((Shape)parent).generateCodeStart(code);
        transform.generateCodeTransform(code);
        ((Shape)parent).generateCodeEnd(code);
    }

    /**
     * The method generates a macro with the equivalent graphical representation of the indexed face set for RoSiML.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        triangleCounter = 0;

        code.indent();
        code.println("<Macro name=\""+ name +"\">");
        code.indent();
        code.println("<ComplexShape  name=\""+name +"\">");
        code.indent();
        if(((Shape)parent).appearance != null)
            code.println("<Appearance ref=\"" + ((Shape)parent).appearance.name + "\" />");
        code.println("<GraphicalRepresentation vertexList=\"" + (texCoord == null ? coord.name : texCoord.name) + "\">");
        code.indent();
        for(LinkedList<Integer> cIdx : (texCoord == null ? coordIndex : texCoordIndex)) {
            if(cIdx.size() == 3)
                code.println("<Triangle v1=\"" + cIdx.get(0) + "\" v2=\"" + cIdx.get(1) + "\" v3=\"" + cIdx.get(2) + "\" />");
            else if(cIdx.size() == 4)
                code.println("<Quad v1=\"" + cIdx.get(0) + "\" v2=\"" + cIdx.get(1) + "\" v3=\"" + cIdx.get(2) + "\"  v4=\"" + cIdx.get(3) + "\" />");
            else
                code.println("<!-- Unknown: " + cIdx.size() + " -->");
            triangleCounter++;
        }

        code.println("<GraphicalAttributes>");
        code.indent();
        code.println("<SmoothSurface value=\"true\"/>");
        code.println("<CombinePrimitives value=\"true\"/>");
        code.unindent();
        code.println("</GraphicalAttributes>");
        code.unindent();
        code.println("</GraphicalRepresentation>");

        if(parent.parent instanceof ObjectNode && ((ObjectNode)parent.parent).boundingObject != null) {
            code.println("<PhysicalRepresentation>");
            ((ObjectNode)parent.parent).generateCodeBounding(code);
            code.println("</PhysicalRepresentation>");
        }
        else
            code.println("<PhysicalRepresentation />");

        code.unindent();
        code.println("</ComplexShape>");
        code.unindent();
        code.println("</Macro>\n");
        code.unindent();
    }
    
    /**
     * The method generates the initial code for a RoSi2 include file
     * 
     * @param code The code stream object
     */
    public void generateRoSi2Code(TreeStream code) {

        code.indent();
        code.println("<ComplexAppearance name=\""+ name +"\">");
        code.indent();
        if(((Shape)parent).appearance != null)
            code.println("<Surface ref=\"" + ((Shape)parent).appearance.name + "\"/>");
        else {
            code.println("<!-- Unknown: surface -->");
            code.println("<Surface diffuseColor=\"#000\"/>");
        }
            
        code.println("<Vertices ref=\"" + (texCoord == null ? coord.name : texCoord.name) + "\"/>");
        //code.indent();
        
        boolean hasTriangles = false;
        boolean hasQuads = false;
        LinkedList<LinkedList<Integer>> coordIndex = texCoord == null ? this.coordIndex : texCoordIndex;
        Coordinate coord = texCoord == null ? this.coord : texCoord;
        
        for(LinkedList<Integer> cIdx : coordIndex) {
            if(cIdx.size() == 3) {
                if(!hasTriangles) {
                    hasTriangles = true;
                    code.println("<Triangles>");
                    code.pushIndention();
                    code.resetIndention();
                }
                
                code.println("" + cIdx.get(0) + " " + cIdx.get(1) + " " + cIdx.get(2));
            }
        }
        if(hasTriangles) {
            code.popIndention();
            code.println("</Triangles>");
        }
        for(LinkedList<Integer> cIdx : coordIndex) {
            if(cIdx.size() == 4) {
                if(!hasQuads) {
                    hasQuads = true;
                    code.println("<Quads>");
                    code.pushIndention();
                    code.resetIndention();
                }
                
                code.println("" + cIdx.get(0) + " " + cIdx.get(1) + " " + cIdx.get(2)+ " " + cIdx.get(3));
            }
        }
        if(hasQuads) {
            code.popIndention();
            code.println("</Quads>\n");
        }
        for(LinkedList<Integer> cIdx : coordIndex) {
            if(cIdx.size() != 3 && cIdx.size() != 4)
                code.println("<!-- Unknown: " + cIdx.size() + " -->");
        }

        code.unindent();
        code.println("</ComplexAppearance>\n");
        code.unindent();
    }
}
