import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.HashMap;

/**
 * The class performs a lexical anlysis. It identifies terminals that are defined
 * in {@link SyntaxAnalysis SyntaxAnalysis}. Identifiers and numbers follow this syntax:
 * <pre>
 * identifier   ::= letter { letter | digit }
 *
 * number       ::= ['-'] digit { digit } ['.' digit { digit } ] [ ('e' | 'E') ['-'] digit { digit } ]
 *
 * letter       ::= 'A' .. 'Z' | 'a' .. 'z' | '_'
 *
 * digit        ::= '0' .. '9'
 * </pre>
 * comments after '#' will be ignored.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class LexicalAnalyzer {
    /** The amount of keywords with its dedicated symbols. */
    private HashMap<String, Symbol.Id> keywords;

    /** The input stream of the input source file. */
    private InputStreamReader reader;

    /** Decides if identified symbols shall be printed to console. */
    private boolean printSymbols;

    /** The actual position in the source code */
    private Position position;

    /** The last read character. */
    private int c;

    /** The last identified symbol. */
    Symbol symbol;

    /**
     * The method reads the next character from the source code.
     * It is stored in the attribute {@link #c c}.
     *
     * @throws IOException  An readerror occures.
     */
    private void nextChar() throws IOException {
        position.next((char) c);
        c = reader.read();
    }

    /**
     * The constructor.
     * 
     * @param fileName The source code filename.
     * @param printSymbols Decides if identified symbols shall be printed to console.
     * @throws FileNotFoundException The source code was not found.
     * @throws IOException A readerror occures.
     */
    LexicalAnalyzer(String fileName, boolean printSymbols)
            throws CompileException, FileNotFoundException, IOException {
        FileInputStream stream = new FileInputStream(fileName);
        reader = new InputStreamReader(stream);
        this.printSymbols = printSymbols;

        keywords = new HashMap<String, Symbol.Id>();
        keywords.put("PROTO", Symbol.Id.PROTO);
        keywords.put("TRUE", Symbol.Id.TRUE);
        keywords.put("FALSE", Symbol.Id.FALSE);
        keywords.put("Robot", Symbol.Id.ROBOT_TYPE);
        keywords.put("translation", Symbol.Id.TRANSLATION);
        keywords.put("rotation", Symbol.Id.ROTATION);
        keywords.put("name", Symbol.Id.NAME);
        keywords.put("children", Symbol.Id.CHILDREN);
        keywords.put("DEF", Symbol.Id.DEF);
        keywords.put("Shape", Symbol.Id.SHAPE_TYPE);
        keywords.put("appearance", Symbol.Id.APPEARANCE);
        keywords.put("Appearance", Symbol.Id.APPEARANCE_TYPE);
        keywords.put("material", Symbol.Id.MATERIAL);
        keywords.put("Material", Symbol.Id.MATERIAL_TYPE);
        keywords.put("ambientIntensity", Symbol.Id.AMBIENT_INTENSITY);
        keywords.put("diffuseColor", Symbol.Id.DIFFUSE_COLOR);
        keywords.put("shininess", Symbol.Id.SHININESS);
        keywords.put("specularColor", Symbol.Id.SPECULAR_COLOR);
        keywords.put("geometry", Symbol.Id.GEOMETRY);
        keywords.put("IndexedFaceSet", Symbol.Id.INDEXED_FACE_SET_TYPE);
        keywords.put("coord", Symbol.Id.COORD);
        keywords.put("Coordinate", Symbol.Id.COORD_TYPE);
        keywords.put("point", Symbol.Id.POINT);
        keywords.put("coordIndex", Symbol.Id.COORD_INDEX);
        keywords.put("creaseAngle", Symbol.Id.CREASE_ANGLE);
        keywords.put("USE", Symbol.Id.USE);
        keywords.put("texCoord", Symbol.Id.TEX_COORD);
        keywords.put("TextureCoordinate", Symbol.Id.TEX_COORD_TYPE);
        keywords.put("texCoordIndex", Symbol.Id.TEX_COORD_INDEX);
        keywords.put("LED", Symbol.Id.LED_TYPE);
        keywords.put("emissiveColor", Symbol.Id.EMISSIVE_COLOR);
        keywords.put("color", Symbol.Id.COLOR);
        keywords.put("gradual", Symbol.Id.GRADUAL);
        keywords.put("Accelerometer", Symbol.Id.ACCELEROMETER_TYPE);
        keywords.put("lookupTable", Symbol.Id.LOOKUP_TABLE);
        keywords.put("Gyro", Symbol.Id.GYRO_TYPE);
        keywords.put("zAxis", Symbol.Id.Z_AXIS);
        keywords.put("DistanceSensor", Symbol.Id.DISTANCE_SENSOR_TYPE);
        keywords.put("type", Symbol.Id.TYPE);
        keywords.put("numberOfRays", Symbol.Id.NUMBER_OF_RAYS);
        keywords.put("aperture", Symbol.Id.APERTURE);
        keywords.put("Servo", Symbol.Id.SERVO_TYPE);
        keywords.put("maxVelocity", Symbol.Id.MAX_VELOCITY);
        keywords.put("maxForce", Symbol.Id.MAX_FORCE);
        keywords.put("minPosition", Symbol.Id.MIN_POSITION);
        keywords.put("maxPosition", Symbol.Id.MAX_POSITION);
        keywords.put("Transform", Symbol.Id.TRANSFORM_TYPE);
        keywords.put("Cylinder", Symbol.Id.CYLINDER_TYPE);
        keywords.put("bottom", Symbol.Id.BOTTOM);
        keywords.put("height", Symbol.Id.HEIGHT);
        keywords.put("radius", Symbol.Id.RADIUS);
        keywords.put("top", Symbol.Id.TOP);
        keywords.put("subdivision", Symbol.Id.SUBDIVISION);
        keywords.put("Camera", Symbol.Id.CAMERA_TYPE);
        keywords.put("windowPosition", Symbol.Id.WINDOW_POSITION);
        keywords.put("pixelSize", Symbol.Id.PIXEL_SIZE);
        keywords.put("fieldOfView", Symbol.Id.FIELD_OF_VIEW);
        keywords.put("width", Symbol.Id.WIDTH);
        keywords.put("near", Symbol.Id.NEAR);
        keywords.put("texture", Symbol.Id.TEXTURE);
        keywords.put("ImageTexture", Symbol.Id.IMAGE_TEXTURE_TYPE);
        keywords.put("url", Symbol.Id.URL);
        keywords.put("boundingObject", Symbol.Id.BOUNDING_OBJECT);
        keywords.put("Sphere", Symbol.Id.SPHERE_TYPE);
        keywords.put("physics", Symbol.Id.PHYSICS);
        keywords.put("Physics", Symbol.Id.PHYSICS_TYPE);
        keywords.put("density", Symbol.Id.DENSITY);
        keywords.put("mass", Symbol.Id.MASS);
        keywords.put("centerOfMass", Symbol.Id.CENTER_OF_MASS);
        keywords.put("Box", Symbol.Id.BOX_TYPE);
        keywords.put("size", Symbol.Id.SIZE);
        keywords.put("TouchSensor", Symbol.Id.TOUCH_SENSOR_TYPE);

        position = new Position(1, 0);
        nextChar();
    }

    /**
     * The method reads the next symbol. It is stored in the attribute
     * {@link #symbol symbol}.
     * @throws CompileException The source code differs from the syntax.
     * @throws IOException A readerror occures.
     */
    void nextSymbol() throws CompileException, IOException {
        for(;;) {
            // Leerraum ignorieren
            while (c != -1 && Character.isWhitespace((char) c)) {
                nextChar();
            }
            if (c == '#') { // comment until line ends
                nextChar();
                while (c != -1 && c != '\n') {
                    nextChar();
                }
                nextChar();
            } else {
                break;
            }
        }

        switch (c) {
        case -1:
            symbol = new Symbol(Symbol.Id.EOF, position);
            break;
        case '[':
            symbol = new Symbol(Symbol.Id.LBRACKET, position);
            nextChar();
            break;
        case ']':
            symbol = new Symbol(Symbol.Id.RBRACKET, position);
            nextChar();
            break;
        case '{':
            symbol = new Symbol(Symbol.Id.LBRACE, position);
            nextChar();
            break;
        case '}':
            symbol = new Symbol(Symbol.Id.RBRACE, position);
            nextChar();
            break;
        case ',':
            symbol = new Symbol(Symbol.Id.COMMA, position);
            nextChar();
            break;
        case '"':
            symbol = new Symbol(Symbol.Id.STRING, position);
            symbol.ident = "";
            nextChar();
            while (c != -1 && c != '"') {
                symbol.ident += (char) c;
                nextChar();
            }
            if (c != '"') {
                throw new CompileException("Unexpected end of string", position);
            }
            nextChar();
            break;
        case '-':
        case '+':
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            double sign = 1;
            symbol = new Symbol(Symbol.Id.NUMBER, position);
            if (c == '-' || c == '+') {
                sign = c == '-' ? -1 : 1;
                nextChar();
                if (c < '0' || c > '9') {
                    throw new CompileException("Number needs a cipher after a prefix", position);
                }
            }
            symbol.number = c - '0';
            nextChar();
            while (c != -1 && Character.isDigit((char) c)) {
                symbol.number = symbol.number * 10 + c - '0';
                nextChar();
            }
            if (c == '.') {
                nextChar();
                double exp = 0.1;
                while (c != -1 && Character.isDigit((char) c)) {
                    symbol.number += exp * (c - '0');
                    exp *= 0.1;
                    nextChar();
                }
            }
            if (c == 'e' || c == 'E') {
                nextChar();
                double expSign = 1;
                if (c == '-' || c == '+') {
                    expSign = c == '-' ? -1 : 1;
                    nextChar();
                    if (c < '0' || c > '9') {
                        throw new CompileException("Exponent needs a number after a prefix", position);
                    }
                }
                double exp = 0;
                while (c != -1 && Character.isDigit((char) c)) {
                    exp = exp * 10 + c - '0';
                    nextChar();
                }
                symbol.number = symbol.number * Math.pow(10, exp * expSign);
            }
            symbol.number *= sign;
            break;
        default:
            if (Character.isLetter((char) c) || c == '_') {
                symbol = new Symbol(Symbol.Id.IDENT, position);
                String ident = "" + (char) c;
                nextChar();
                while (c != -1 && (Character.isLetterOrDigit((char) c) || c == '_')) {
                    ident = ident + (char) c;
                    nextChar();
                }
                Symbol.Id id = keywords.get(ident);
                if (id != null) {
                    symbol.id = id;
                } else {
                    symbol.ident = ident;
                }
            } else {
                throw new CompileException("Unexpected character: " + (char) c + " (code " + c + ")", position);
            }
        }
        if (printSymbols) {
            System.out.println(symbol.toString());
        }
    }

    public void close() throws IOException {
        reader.close();
    }
}
