
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;

/**
 *  The class represents the surface material.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class Material extends Node {
    /** The reflection intensity of the ambient light. */
    double ambientIntensity = -1;
    
    /** The diffuse color reflection values. */
    double[] diffuseColor;
    
    /** The shininess. */
    double shininess = -1;
    
    /** The specular color for the shininess. */
    double[] specularColor;
    
    /** The emissive values of the color. */
    double[] emissiveColor;
    
    /** The counter to build a unique name. */
    private static int count = 0;

    /**
     * The constructor.
     * 
     * @param name The material identifier name.
     * @param parent The reference to the parent object.
     */
    Material(String name, Node parent) {
        super(name, parent);
    }

    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        if(parent != null && parent.parent != null && parent.parent.name != null)
            name = parent.parent.name + "_MATERIAL" + count ++;
        else
            generateName("MATERIAL" + count ++);
    }

    /**
     * The method generates the material code for a shape.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        code.indent();
        if(diffuseColor != null && diffuseColor.length == 3)
            code.println("<Color r=\"" + diffuseColor[0] + "\" g=\"" + diffuseColor[1] + "\" b=\"" + diffuseColor[2] + "\" a=\"1.0\" />");
        if(specularColor != null && specularColor.length == 3)
            code.println("<SpecularColor r=\"" + specularColor[0] + "\" g=\"" + specularColor[1] + "\" b=\"" + specularColor[2] + "\" a=\"1.0\" />");
        if(emissiveColor != null && emissiveColor.length == 3)
            code.println("<EmissionColor r=\"" + emissiveColor[0] + "\" g=\"" + emissiveColor[1] + "\" b=\"" + emissiveColor[2] + "\" a=\"1.0\" />");
        if(shininess != -1)
            code.println("<Shininess value=\"" + shininess + "\" />");
        if(ambientIntensity != -1)
            code.println("<Reflectivity value=\"" + ambientIntensity + "\" />");
        code.unindent();
    }
    
    /**
    * The method returns the colors in the RoSi2 format to be used in a "Surface" tag.
    * @return The formated colors
    */
    String getRoSi2Colors()
    {
      DecimalFormat f = new DecimalFormat("#.##", DecimalFormatSymbols.getInstance(Locale.US)); 
      String colors = "";
      if(diffuseColor != null && diffuseColor.length == 3) {
          colors += " diffuseColor=\"rgb("+f.format(diffuseColor[0]*100.0)+"%, "+f.format(diffuseColor[1]*100.0)+"%, "+f.format(diffuseColor[2]*100.0)+"%)\"";
          if(ambientIntensity != -1)
            colors += " ambientColor=\"rgb("+f.format(diffuseColor[0]*100.0*ambientIntensity)+"%, "+f.format(diffuseColor[1]*100.0*ambientIntensity)+"%, "+f.format(diffuseColor[2]*100.0*ambientIntensity)+"%)\"";
      }
      else
          colors += " diffuseColor=\"#000\"";
      if(specularColor != null && specularColor.length == 3)
          colors += " specularColor=\"rgb("+f.format(specularColor[0]*100.0)+"%, "+f.format(specularColor[1]*100.0)+"%, "+f.format(specularColor[2]*100.0)+"%)\"";
      if(emissiveColor != null && emissiveColor.length == 3)
        colors += " emissionColor=\"rgb("+f.format(emissiveColor[0]*100.0)+"%, "+f.format(emissiveColor[1]*100.0)+"%, "+f.format(emissiveColor[2]*100.0)+"%)\"";
      if(shininess != -1)
          colors += " shininess=\"" + f.format(shininess) + "\"";
      return colors;
    }
}
