/**
 * The class extends the tranformable objects by physics.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class ObjectNode extends Transform {
    /** Contains the physical bounding object of the object. */
    Node boundingObject;
    
    /** Contians the physical atrributes of the object. */
    Physics physics;

    /**
     * The constructor.
     * 
     * @param name The object identifier name.
     * @param parent The reference to the parent object.
     */
    ObjectNode(String name, Node parent) {
        super(name, parent);
    }

    /**
     * The method generates the begin bounding code.
     * 
     * @param code The code output stream.
     * @param simple Is it a physical simple object?
     */
    public void generateCodeBoundingStart(TreeStream code, boolean simple) {
        if(boundingObject == null)
            return;

        if(boundingObject instanceof Transform) {
            Transform transform = (Transform) boundingObject;
            transform.generateCodeBoundingStart(code, simple);
            if(physics != null && physics.centerOfMass != null) {
                if(transform.translation != null) {
                    physics.centerOfMass[0] -= transform.translation[0];
                    physics.centerOfMass[1] -= transform.translation[1];
                    physics.centerOfMass[2] -= transform.translation[2];
                }
                if(transform.rotation != null) {
                    Matrix3D mx = new Matrix3D();
                    Matrix3D my = new Matrix3D();
                    Matrix3D mz = new Matrix3D();
                    mx.setXRotation(-transform.rotation[0]);
                    my.setYRotation(-transform.rotation[1]);
                    mz.setZRotation(-transform.rotation[2]);
                    Vector3D v = new Vector3D();
                    v.v[0] = physics.centerOfMass[0];
                    v.v[1] = physics.centerOfMass[1];
                    v.v[2] = physics.centerOfMass[2];
                    v = mx.mult(my.mult(mz.mult(v)));
                    physics.centerOfMass[0] = (float) v.v[0];
                    physics.centerOfMass[1] = (float) v.v[1];
                    physics.centerOfMass[2] = (float) v.v[2];
                }
            }
        } else if(boundingObject instanceof Shape) {
            ((Shape) boundingObject).geometry.generateCodeBoundingStart(code, simple);
        } else {
            ((Geometry) boundingObject).generateCodeBoundingStart(code, simple);
        }

        if(physics != null)
            physics.generateCode(code);
    }

    /**
     * The method generates the closure bounding code.
     * 
     * @param code The code output stream.
     * @param simple Is it a physical simple object?
     */
    public void generateCodeBoundingEnd(TreeStream code, boolean simple) {
        if(boundingObject == null)
            return;

        if(boundingObject instanceof Transform)
            ((Transform)boundingObject).generateCodeBoundingEnd(code, simple);
        else if(boundingObject instanceof Shape)
            ((Shape)boundingObject).geometry.generateCodeBoundingEnd(code, simple);
        else
            ((Geometry)boundingObject).generateCodeBoundingEnd(code, simple);
    }

    /**
     * The method generates the bounding code of the object.
     * 
     * @param code The code output stream.
     */
    public void generateCodeBounding(TreeStream code) {
        generateCodeBoundingStart(code, true);
        generateCodeBoundingEnd(code, true);
    }

    /**
     * The method generates the code of the super class.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        super.generateCode(code);
    }
}
