import java.util.LinkedList;

/**
 * The class represents a shape. The shape is used to create rendered objects.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class Shape extends Node {
    /** The geometry. */
    Geometry geometry;
    
    /** The appearance contains visual attributes to be applied to the geometry. */
    Appearance appearance;
    
    /** The counter to build unique names. */
    private static int count = 0;
    
    /** Contains a possible subshape. */
    Shape subShape = null;

    /**
     * The constructor.
     * 
     * @param name The objects identifier name.
     * @param parent The reference to the parent object.
     */
    Shape(String name, Node parent) {
        super(name, parent);
        shapes.add(this);
    }

    /**
     * The method generates a unique name.
     */
    public void generateName() {
        if(name != null)
            return;
        if(appearance != null && appearance.name != null)
            name = appearance.name;
        generateName("SHAPE" + count ++);
    }

    /**
     * The method generates the begin code of a use macro to use the macro build out of this shape.
     * This allows to generate other elements in a use macro.
     * 
     * @param code The code output stream.
     */
    public void generateCodeStart(TreeStream code) {
        if(!(geometry instanceof IndexedFaceSet))
            return;

        code.indent();
        code.println("<Use macroName=\"" + ((IndexedFaceSet)geometry).name + "\" instanceName=\"" + this.name + "\">");
    }

    /**
     * The method generates an empty use macro to use the macro build out of this shape.
     * 
     * @param code The code output stream.
     */
    public void generateCodeSingle(TreeStream code) {
        if(!(geometry instanceof IndexedFaceSet))
            return;

        code.indent();
        code.println("<Use macroName=\"" + ((IndexedFaceSet)geometry).name + "\" instanceName=\"" + this.name + "\" />");
        code.unindent();

        if(subShape != null)
            subShape.generateCodeSingle(code);
    }

    /**
     * The method generates the closure of a use macro.
     * 
     * @param code The code output stream.
     */
    public void generateCodeEnd(TreeStream code) {
        if(!(geometry instanceof IndexedFaceSet))
            return;

        code.println("</Use>");
        code.unindent();
    }

    /**
     * The method generates the code of a shape without children.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        generateCode(code, null);
    }

    /**
     * The method generates the full code of a shape.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code, LinkedList<Node> children) {
        if(!(geometry instanceof IndexedFaceSet)) {
            geometry.generateCode(code);
            return;
        }

        generateCodeStart(code);

        if(parent instanceof LED && ((LED)parent).translation != null)
            ((LED)parent).generateCodeTransform(code);

        if(subShape != null || children != null) {
            code.indent();
            code.println("<Elements>");
            if(subShape != null)
                subShape.generateCodeSingle(code);
        }

        if(children != null)
            for(Node child = children.pollFirst(); child != null; child = children.pollFirst())
                child.generateCode(code);

        if(subShape != null || children != null) {
            code.println("</Elements>");
            code.unindent();
        }

        generateCodeEnd(code);
    }
}
