import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.LinkedList;
import java.util.HashMap;
import java.lang.Math;

/**
 * The class realises the syntactic analysis.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class SyntaxAnalyzer extends LexicalAnalyzer {
   HashMap<String, Node> macros = new HashMap<String, Node>();

    /**
     * The method creates an "unexpected symbol" error.
     * @throws CompileException The according errormessage.
     */
    private void unexpectedSymbol() throws CompileException {
        throw new CompileException("Unexpectet symbol " + symbol.id.toString(), symbol);
    }

    /**
     * The method checks if the actual symbol is expected. Is it expected, the next 
     * symbol is read.
     * @param id The expected symbol.
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private void expectSymbol(Symbol.Id id) throws CompileException, IOException {
        if (id != symbol.id) {
            unexpectedSymbol();
        }
        nextSymbol();
    }

    /**
     * The method checks if the actual symbol is an identifier. An identifier will be returned.
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private String expectIdent() throws CompileException, IOException {
        if (symbol.id != Symbol.Id.IDENT) {
            unexpectedSymbol();
        }
        String s = symbol.ident;
        nextSymbol();
        return s;
    }

    /**
     * The method checks if the actual symbol is a number. A number will be returned.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private double expectNumber() throws CompileException, IOException {
        if (symbol.id != Symbol.Id.NUMBER) {
            unexpectedSymbol();
        }
        double f = symbol.number;
        nextSymbol();
        return f;
    }

    /**
     * The method checks if the actual symbol is a number. A number will be returned.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private String expectString() throws CompileException, IOException {
        if (symbol.id != Symbol.Id.STRING) {
            unexpectedSymbol();
        }
        String s = symbol.ident.replace('/', '-');
        nextSymbol();
        return s;
    }

    
    /**
     * The method reads a vector of the given length. The vector will be returned.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private double[] expectVector(int length) throws CompileException, IOException {
        double[] f = new double[length];
        for (int i = 0; i < length; ++i) {
            f[i] = expectNumber();
        }
        return f;
    }

    
    /**
     * The method checks if the actual symbol is a boolean. A boolean will be returned.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private boolean expectBoolean() throws CompileException, IOException {
        if (symbol.id == Symbol.Id.TRUE) {
            nextSymbol();
            return true;
        } else if (symbol.id == Symbol.Id.FALSE) {
            nextSymbol();
            return false;
        } else {
            unexpectedSymbol();
            return false;
        }
    }

    
    /**
     * The method read an identifier of the type node. A node will be returned.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Node node(Node p) throws CompileException, IOException {
        Node n = null;
        String macroName = null;
        if (symbol.id == Symbol.Id.DEF) {
            nextSymbol();
            macroName = expectIdent();
        }
        switch (symbol.id) {
        case ACCELEROMETER_TYPE:
            nextSymbol();
            n = accelerometer(macroName, p);
            break;
        case APPEARANCE_TYPE:
            nextSymbol();
            n = appearance(macroName, p);
            break;
        case BOX_TYPE:
            nextSymbol();
            n = box(macroName, p);
            break;
        case CAMERA_TYPE:
            nextSymbol();
            n = camera(macroName, p);
            break;
        case COORD_TYPE:
            nextSymbol();
            n = coordinate(macroName, p, 3);
            break;
        case CYLINDER_TYPE:
            nextSymbol();
            n = cylinder(macroName, p);
            break;
        case DISTANCE_SENSOR_TYPE:
            nextSymbol();
            n = distanceSensor(macroName, p);
            break;
        case GYRO_TYPE:
            nextSymbol();
            n = gyro(macroName, p);
            break;
        case IMAGE_TEXTURE_TYPE:
            nextSymbol();
            n = imageTexture(macroName, p);
            break;
        case INDEXED_FACE_SET_TYPE:
            nextSymbol();
            n = indexedFaceSet(macroName, p);
            break;
        case LED_TYPE:
            nextSymbol();
            n = led(macroName, p);
            break;
        case MATERIAL_TYPE:
            nextSymbol();
            n = material(macroName, p);
            break;
        case PHYSICS_TYPE:
            nextSymbol();
            n = physics(macroName, p);
            break;
        case ROBOT_TYPE:
            nextSymbol();
            n = robot(macroName, p);
            break;
        case SERVO_TYPE:
            nextSymbol();
            n = servo(macroName, p);
            break;
        case SHAPE_TYPE:
            nextSymbol();
            n = shape(macroName, p);
            break;
        case SPHERE_TYPE:
            nextSymbol();
            n = sphere(macroName, p);
            break;
        case TEX_COORD_TYPE:
            nextSymbol();
            n = coordinate(macroName, p, 2);
            break;
        case TOUCH_SENSOR_TYPE:
            nextSymbol();
            n = touchSensor(macroName, p);
            break;
        case TRANSFORM_TYPE:
            nextSymbol();
            n = transform(macroName, p);
            break;
        case USE:
            nextSymbol();
            n = macros.get(expectIdent());
            break;
        default:
            assert false;
            return null;
        }
        if (macroName != null) {
            macros.put(macroName, n);
        }
        n.generateName();
        return n;
    }

    
    /**
     * The method parses a robot definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Robot robot(String name, Node parent) throws CompileException, IOException {
        Robot r = new Robot(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                r.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                r.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                r.name = expectString();
                break;
            case CHILDREN:
                nextSymbol();
                children(r);
                break;
            case BOUNDING_OBJECT:
                nextSymbol();
                r.boundingObject = node(r);
                break;
            case PHYSICS:
                nextSymbol();
                r.physics = (Physics) node(r);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return r;
    }

    /**
     * The method parses a shape definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Shape shape(String name, Node parent) throws CompileException, IOException {
        Shape s = new Shape(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case GEOMETRY:
                nextSymbol();
                s.geometry = (Geometry) node(s);
                break;
            case APPEARANCE:
                nextSymbol();
                s.appearance = (Appearance) node(s);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return s;
    }

    /**
     * The method parses an appearance definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Appearance appearance(String name, Node parent) throws CompileException, IOException {
        Appearance a = new Appearance(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case MATERIAL:
                nextSymbol();
                a.material = (Material) node(a);
                break;
            case TEXTURE:
                nextSymbol();
                a.texture = (ImageTexture) node(a);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return a;
    }

    /**
     * The method parses a material definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Material material(String name, Node parent) throws CompileException, IOException {
        Material m = new Material(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case AMBIENT_INTENSITY:
                nextSymbol();
                m.ambientIntensity = expectNumber();
                break;
            case DIFFUSE_COLOR:
                nextSymbol();
                m.diffuseColor = expectVector(3);
                break;
            case SHININESS:
                nextSymbol();
                m.shininess = expectNumber();
                break;
            case SPECULAR_COLOR:
                nextSymbol();
                m.specularColor = expectVector(3);
                break;
            case EMISSIVE_COLOR:
                nextSymbol();
                m.emissiveColor = expectVector(3);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return m;
    }

    /**
     * The method parses an image texture definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private ImageTexture imageTexture(String name, Node parent) throws CompileException, IOException {
        ImageTexture t = new ImageTexture(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case URL:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                t.url = expectString();
                expectSymbol(Symbol.Id.RBRACKET);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return t;
    }
    
    /**
     * The method parses an indexed face set definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private IndexedFaceSet indexedFaceSet(String name, Node parent) throws CompileException, IOException {
        IndexedFaceSet f = new IndexedFaceSet(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case COORD:
                nextSymbol();
                f.coord = (Coordinate) node(f);
                break;
            case TEX_COORD:
                nextSymbol();
                f.texCoord = (Coordinate) node(f);
                break;
            case COORD_INDEX:
            case TEX_COORD_INDEX:
                boolean isTex = symbol.id == Symbol.Id.TEX_COORD_INDEX;
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                LinkedList<Integer> l = new LinkedList<Integer>();
                while (symbol.id != Symbol.Id.RBRACKET) {
                    double n = expectNumber();
                    if (n == -1) {
                        if (isTex) {
                            f.texCoordIndex.add(l);
                        } else {
                            f.coordIndex.add(l);
                        }
                        l = new LinkedList<Integer>();
                    } else {
                        l.add((int) n);
                    }
                    if (symbol.id == Symbol.Id.COMMA) {
                        nextSymbol();
                    } else if (symbol.id != Symbol.Id.RBRACKET) {
                        unexpectedSymbol();
                    }
                }
                if (isTex) {
                    f.completeTexCoords();
                }
                nextSymbol();
                break;
            case CREASE_ANGLE:
                nextSymbol();
                f.creaseAngle = expectNumber();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return f;
    }

    /**
     * The method parses a coordinate definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * @param dims The dimension.
     * 
     * @throws CompileException An unexpected symbol was read
     * @throws IOException A readerror occures
     */
    private Coordinate coordinate(String name, Node parent, int dims) throws CompileException, IOException {
        Coordinate c = new Coordinate(name, parent, dims == 3 ? Coordinate.Type.COORD : Coordinate.Type.TEXCOORD);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case POINT:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                while (symbol.id != Symbol.Id.EOF && symbol.id != Symbol.Id.RBRACKET) {
                    c.point.add(dims == 3 ? convertTranslation(expectVector(dims)) : expectVector(dims));
                }
                nextSymbol();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return c;
    }

    /**
     * The method parses a cylinder definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Cylinder cylinder(String name, Node parent) throws CompileException, IOException {
        Cylinder c = new Cylinder(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case BOTTOM:
                nextSymbol();
                c.bottom = expectBoolean();
                break;
            case TOP:
                nextSymbol();
                c.top = expectBoolean();
                break;
            case HEIGHT:
                nextSymbol();
                c.height = expectNumber();
                break;
            case RADIUS:
                nextSymbol();
                c.radius = expectNumber();
                break;
            case SUBDIVISION:
                nextSymbol();
                c.subdivision = (int) expectNumber();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return c;
    }

    /**
     * The method parses a sphere definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Sphere sphere(String name, Node parent) throws CompileException, IOException {
        Sphere s = new Sphere(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case RADIUS:
                nextSymbol();
                s.radius = expectNumber();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return s;
    }

    /**
     * The method parses a box definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Box box(String name, Node parent) throws CompileException, IOException {
        Box b = new Box(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case SIZE:
                nextSymbol();
                b.size = convertTranslation(expectVector(3));
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return b;
    }

    /**
     * The method parses a led definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private LED led(String name, Node parent) throws CompileException, IOException {
        LED l = new LED(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                l.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                l.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                l.name = expectString();
                break;
            case CHILDREN:
                nextSymbol();
                children(l);
                break;
            case COLOR:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                if (symbol.id != Symbol.Id.RBRACKET) {
                    l.color = expectVector(3);
                }
                expectSymbol(Symbol.Id.RBRACKET);
                break;
            case GRADUAL:
                nextSymbol();
                l.gradual = expectBoolean();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        l.createAppearance();
        return l;
    }

    /**
     * The method parses a accelerometer definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Accelerometer accelerometer(String name, Node parent) throws CompileException, IOException {
        Accelerometer a = new Accelerometer(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                a.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                a.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                a.name = expectString();
                break;
            case LOOKUP_TABLE:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                while (symbol.id != Symbol.Id.RBRACKET) {
                    a.lookupTable.add(expectVector(3));
                }
                nextSymbol();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return a;
    }

    /**
     * The method parses a gyro definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Gyro gyro(String name, Node parent) throws CompileException, IOException {
        Gyro g = new Gyro(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                g.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                g.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                g.name = expectString();
                break;
            case LOOKUP_TABLE:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                while (symbol.id != Symbol.Id.RBRACKET) {
                    g.lookupTable.add(expectVector(3));
                }
                nextSymbol();
                break;
            case Z_AXIS:
                nextSymbol();
                g.zAxis = expectBoolean();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return g;
    }

    /**
     * The method parses a distance sensor definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private DistanceSensor distanceSensor(String name, Node parent) throws CompileException, IOException {
        DistanceSensor d = new DistanceSensor(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                d.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                d.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                d.name = expectString();
                break;
            case LOOKUP_TABLE:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                while (symbol.id != Symbol.Id.RBRACKET) {
                    d.lookupTable.add(expectVector(3));
                }
                nextSymbol();
                break;
            case TYPE:
                nextSymbol();
                d.type = expectString();
                break;
            case NUMBER_OF_RAYS:
                nextSymbol();
                d.numberOfRays = (int) expectNumber();
                break;
            case APERTURE:
                nextSymbol();
                d.aperture = expectNumber();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return d;
    }

    /**
     * The method parses a touch sensor definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private TouchSensor touchSensor(String name, Node parent) throws CompileException, IOException {
        TouchSensor t = new TouchSensor(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                t.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                t.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                t.name = expectString();
                break;
            case LOOKUP_TABLE:
                nextSymbol();
                expectSymbol(Symbol.Id.LBRACKET);
                while (symbol.id != Symbol.Id.RBRACKET) {
                    t.lookupTable.add(expectVector(3));
                }
                nextSymbol();
                break;
            case TYPE:
                nextSymbol();
                t.type = expectString();
                break;
            case CHILDREN:
                nextSymbol();
                children(t);
                break;
            case BOUNDING_OBJECT:
                nextSymbol();
                t.boundingObject = node(t);
                break;
            case PHYSICS:
                nextSymbol();
                t.physics = (Physics) node(t);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return t;
    }

    /**
     * The method parses a servo definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Servo servo(String name, Node parent) throws CompileException, IOException {
        Servo s = new Servo(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                s.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                s.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                s.name = expectString();
                break;
            case CHILDREN:
                nextSymbol();
                children(s);
                break;
            case MAX_VELOCITY:
                nextSymbol();
                s.maxVelocity = expectNumber();
                break;
            case MAX_FORCE:
                nextSymbol();
                s.maxForce = expectNumber();
                break;
            case MIN_POSITION:
                nextSymbol();
                s.minPosition = expectNumber();
                break;
            case MAX_POSITION:
                nextSymbol();
                s.maxPosition = expectNumber();
                break;
            case BOUNDING_OBJECT:
                nextSymbol();
                s.boundingObject = node(s);
                break;
            case PHYSICS:
                nextSymbol();
                s.physics = (Physics) node(s);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return s;
    }

    /**
     * The method parses a tranformation definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Transform transform(String name, Node parent) throws CompileException, IOException {
        Transform t = new Transform(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                t.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                t.rotation = convertRotation(expectVector(4));
                break;
            case CHILDREN:
                nextSymbol();
                children(t);
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return t;
    }

    /**
     * The method parses a camera definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Camera camera(String name, Node parent) throws CompileException, IOException {
        Camera c = new Camera(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case TRANSLATION:
                nextSymbol();
                c.translation = convertTranslation(expectVector(3));
                break;
            case ROTATION:
                nextSymbol();
                c.rotation = convertRotation(expectVector(4));
                break;
            case NAME:
                nextSymbol();
                c.name = expectString();
                break;
            case WINDOW_POSITION:
                nextSymbol();
                c.windowPosition = expectVector(2);
                break;
            case PIXEL_SIZE:
                nextSymbol();
                c.pixelSize = (int) expectNumber();
                break;
            case FIELD_OF_VIEW:
                nextSymbol();
                c.fieldOfView = expectNumber();
                break;
            case WIDTH:
                nextSymbol();
                c.width = (int) expectNumber();
                break;
            case HEIGHT:
                nextSymbol();
                c.height = (int) expectNumber();
                break;
            case NEAR:
                nextSymbol();
                c.near = expectNumber();
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return c;
    }

    /**
     * The method parses the physics definition and returns it.
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private Physics physics(String name, Node parent) throws CompileException, IOException {
        Physics p = new Physics(name, parent);
        expectSymbol(Symbol.Id.LBRACE);
        while (symbol.id != Symbol.Id.RBRACE) {
            switch (symbol.id) {
            case DENSITY:
                nextSymbol();
                p.density = expectNumber();
                break;
            case MASS:
                nextSymbol();
                p.mass = expectNumber();
                break;
            case CENTER_OF_MASS:
                nextSymbol();
                p.centerOfMass = convertTranslation(expectVector(3));
                break;
            default:
                unexpectedSymbol();
            }
        }
        nextSymbol();
        return p;
    }

    /**
     * The method parses for childrens. 
     * 
     * @param name The parsed instance name.
     * @param parent The parent object.
     * 
     * @throws CompileException An unexpected symbol was read.
     * @throws IOException A readerror occures.
     */
    private void children(Transform t) throws CompileException, IOException {
        Node n;
        Shape last = null;
        Shape first = null;
        expectSymbol(Symbol.Id.LBRACKET);
        while (symbol.id != Symbol.Id.RBRACKET) {
            n = node(t);
            if(n instanceof Shape) {
                if(last != null) {
                    last.subShape = (Shape)n;
                    n.parent = last;
                } else {
                    t.children.add(n);
                    first = (Shape)n;
                }
                last = (Shape)n;
            } else
                t.children.add(n);
        }
        nextSymbol();
    }

    /**
     * The method converts the rotation from the Webots definition in radiants
     * to the SimRobot definition in degrees.
     * 
     * @param r The rotation vector in Webots format
     * @returns The rotation vector in SimRobot format
     */
    private double[] convertRotation(double[] r) {
        double[] rnew = convertTranslation(r),
            rotation = new double[3];

        if(r[3] == 0) {
            rotation[0] = rnew[0];
            rotation[1] = rnew[1];
            rotation[2] = rnew[2];
            return rotation;
        }

        double
            pif = 180 / Math.PI,
            cos = Math.cos(r[3]),
            sin = Math.sin(r[3]),
            x = rnew[0],
            y = rnew[1],
            z = rnew[2],
            l = Math.sqrt(x * x + y * y + z * z);
        x /= l;
        y /= l;
        z /= l;

        // Calculate angles for x,y,z rotation out of the Webots Axis-Angle format
        rotation[1] = Math.atan2(-((1 - cos) * x * z + y * sin), Math.sqrt(Math.pow((1 - cos) * x * x + cos, 2) + Math.pow((1 - cos) * x * y - z * sin, 2)));

        if(Math.abs(rotation[1]) == Math.PI / 2) {
            rotation[0] = 0;
            rotation[2] = Math.signum(rotation[1]) * Math.atan2((1 - cos) * x * y + z * sin, (1 - cos) * y * y + cos);
        } else {
            double cosr = Math.cos(rotation[1]);
            rotation[0] =  Math.atan2(((1 - cos) * x * y - z * sin) / cosr, ((1 - cos) * x * x + cos) / cosr);
            rotation[2] =  Math.atan2(((1 - cos) * y * z - x * sin) / cosr, ((1 - cos) * z * z + cos) / cosr);
        }

        rotation[0] *= pif;
        rotation[1] *= pif;
        rotation[2] *= pif;

        return rotation;
    }

    private double[] convertTranslation(double[] t) {
        double[] translation = new double[3];

        translation[0] = t[2];
        translation[1] = t[0];
        translation[2] = t[1];

        return translation;
    }

    /**
     * The constructor
     * @param fileName The name of the source code file.
     * @param printSymbols The identified symbols will be printed to console.
     * @throws CompileException The source code does not expect the syntax rules.
     * @throws FileNotFoundException The source code was not found.
     * @throws IOException A readerror occures..
     */
    SyntaxAnalyzer(String fileName, boolean printSymbols)
            throws CompileException, FileNotFoundException, IOException {
        super(fileName, printSymbols);
    }

    /**
     * The method parses the souce code and returns the root element of the syntaxtree.
     * @throws CompileException The source code does not expect the syntax rules
     * @throws IOException A readerror occures.
     */
    Robot parse() throws CompileException, IOException {
        nextSymbol();
        Robot r = (Robot) node(null);
        expectSymbol(Symbol.Id.EOF);
        close();
        return r;
    }
}
