import java.util.LinkedList;

/**
 * The class extends a node by six degrees of freedom.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */

class Transform extends Node {
    /** The translation */
    double[] translation;
    
    /** The rotation */
    double[] rotation;
    
    /** The children of this object */
    LinkedList<Node> children = new LinkedList<Node>();

    /**
     * The constructor.
     * 
     * @param name The objects identifier name.
     * @param parent The reference to the parent object.
     */
    Transform(String name, Node parent) {
        super(name, parent);
    }

    /**
     * The method generates the begin bounding code.
     * 
     * @param code The code output stream.
     * @param simple Is it a physical simple object?
     */
    public void generateCodeBoundingStart(TreeStream code, boolean simple) {
        ((Geometry)children.getFirst()).generateCodeBoundingStart(code, simple);
        generateCodeTransform(code);
    }

    /**
     * The method generates the closure of the bounding code.
     * 
     * @param code The code output stream.
     * @param simple Is it a physical simple object?
     */
    public void generateCodeBoundingEnd(TreeStream code, boolean simple) {
        ((Geometry)children.getFirst()).generateCodeBoundingEnd(code, simple);
    }

    /**
     * The method generates a translation output with the given parameters.
     * 
     * @param code The code output stream.
     * @param x The x-axis translation.
     * @param y The y-axis translation.
     * @param z The z-axis translation.
     */
    static public void generateCodeTransform(TreeStream code, double x, double y, double z) {
        code.indent();
        code.println("<Translation x=\"" + x + "\" y=\"" + y + "\" z=\"" + z + "\" />");
        code.unindent();
    }

    /**
     * The method generates the rotation and translation for an object.
     * 
     * @param code The code output stream.
     */
    public void generateCodeTransform(TreeStream code) {
        code.indent();
        if(translation != null)
            code.println("<Translation x=\"" + translation[0] + "\" y=\"" + translation[1] + "\" z=\"" + translation[2] + "\" />");
        if(rotation != null)
            code.println("<Rotation x=\"" + rotation[0] + "\" y=\"" + rotation[1] + "\" z=\"" + rotation[2] + "\" />");
        code.unindent();
    }

    /**
     * The method generates the transform code.
     * 
     * @param code The code output stream.
     */
    public void generateCode(TreeStream code) {
        if(children.size() == 0)
            return;

        if(this.getClass().getName() == "Transform") {
            if(children.peekFirst() instanceof Shape && ((Shape)children.peekFirst()).geometry != null)
                ((Shape)children.peekFirst()).geometry.generateCode(code, this);
            return;
        }

        code.indent();
        code.println("<Elements>");

        for(int i = 0; i < children.size(); i++) {
            Node child = children.get(i);
            if(child instanceof TouchSensor) {
                child.generateCode(code);
                children.remove(child);
                i--;
            }
        }

        if(this instanceof Robot)
            children.add(2, children.pollLast());

        if(children.peekFirst() instanceof Shape && ((Shape)children.peekFirst()).geometry instanceof IndexedFaceSet)
            ((Shape)children.pollFirst()).generateCode(code, children.size() == 0 ? null : children);
        else
            for(Node child = children.pollFirst(); child != null; child = children.pollFirst())
                child.generateCode(code);

        code.println("</Elements>");
        code.unindent();
    }
}
