import java.io.OutputStream;
import java.io.PrintStream;
import java.util.LinkedList;

/**
 * The class implements an outputstream which can structure text tree-like.
 * The class inherits from {@link java.io.PrintStream PrintStream} and allows you to use all
 * methods who are able to write to the console.<br />
 * In addition, there are methods to control the indentation.
 * 
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas R&ouml;fer</a>
 * @author <a href="mailto:b21@informatik.uni-bremen.de">Oliver B&ouml;sche</a>
 * @author <a href="mailto:raider@informatik.uni-bremen.de">Thiemo Wiedemeyer</a>
 */
class TreeStream extends PrintStream {
    /**Tthe step size of the indention. */
    private int indentionStep;

    /**
     * Buffers the last printed character. If that was a '\n', there will be an 
     * indention before the output of next character.
     */
    private int lastChar = 0;

    /** The actual indention depth. */
    private int indention = 0;
    
    /** A stack for storing and revering indention depths. */
    private LinkedList<Integer> indentionStack = new LinkedList<Integer>();

    /**
     * The constructor.
     * @param stream The outputsteam int that will be written.
     * @param indentionStep Die Schrittweite der Einr�ckung.
     */
    TreeStream(OutputStream stream, int indentionStep) {
        super(stream);
        this.indentionStep = indentionStep;
    }

    /**
     * The method increments the indention depth.
     */
    void indent() {
        indention += indentionStep;
    }

    /**
     * The method decrements the indention depth.
     */
    void unindent() {
        indention -= indentionStep;
        assert indention >= 0;
    }
    
    /**
    * Resets the current indention depth.
    */
    void resetIndention() {
        indention = 0;
    }
    
    /**
    * Pushes the current indention depth on a stack.
    */
    void pushIndention() {
        indentionStack.addLast(indention);
    }
    
    /**
    * Restores the indention depth that was pushed on the stack before.
    */
    void popIndention() {
        indention = indentionStack.removeLast();
    }

    /**
     * This method overrides the output method of the base class.
     * It makes shure that the indention will be printed.
     * 
     * @param buf The output buffer.
     * @param off The offset to describe the first character to print.
     * @param len The number of character that issued.
     */
    public void write(byte[] buf, int off, int len) {
        for (int i = 0; i < len; ++i) {
            write((int) buf[off + i]);
        }
    }

    /**
     * This method overrides the output method of the base class.
     * It makes shure that the indention will be printed.
     * 
     * @param str The output string.
     * @param off The offset to describe the first character to print.
     * @param len The number of character that shall issued.
     */
    public void write(String str, int off, int len) {
        write(str.getBytes(), off, len);
    }

    /**
     This method overrides the output method of the base class.
     * It makes shure that the indention will be printed.
     * 
     * @param b The issued character.
     */
    public void write(int b) {
        if (lastChar == '\n') {
            for (int i = 0; i < indention; ++i) {
                super.write(' ');
            }
        }
        lastChar = b;
        super.write(b);
    }
}
