/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../stack.h"
#include "../../executionContext.h"
#include "../../tools/asserts.h"
#include "../declarationCollection.h"
#include "../identifier.h"
#include "../contextAnalysisException.h"
#include "../statements/statementBlock.h"
#include "variableDeclaration.h"
#include "functionDeclaration.h"

using namespace ast;

FunctionDeclaration::ParameterDeclaration::~ParameterDeclaration()
{
    for(size_t i = 0; i < size(); i++)
        delete at(i);
    clear();
}

FunctionDeclaration::FunctionDeclaration(Identifier *ident,
                                         VariableDeclaration *returnVarDecl,
                                         ParameterDeclaration *paramDecls,
                                         StatementBlock *statements,
                                         const Location &location,
                                         bool isConst)
    : Declaration(ident),
      returnVarDecl(returnVarDecl),
      paramDecls(paramDecls),
      statements(statements),
      isConst(isConst)
{
    ASSERT(ident);
    ASSERT(paramDecls);
}

FunctionDeclaration::~FunctionDeclaration()
{
    if(returnVarDecl)
        delete returnVarDecl;

    delete paramDecls;

    if(statements)
        delete statements;
}

void FunctionDeclaration::contextAnalysis(DeclarationCollection &declarationCollection)
{
    declarationCollection.enter();

    declarationCollection.varIdInFrame = -int(paramDecls->size()) - 1;
    if(returnVarDecl)
        returnVarDecl->contextAnalysis(declarationCollection);

    if(statements)
        statements->contextAnalysisFunctionDeclarations(declarationCollection);

    declarationCollection.varIdInFrame = -int(paramDecls->size());

    for(size_t i = 0; i < paramDecls->size(); i++)
    {
        if(paramDecls->at(i)->initializeStatement)
            throw ContextAnalysisException(ident->location, "default parameters not supported!");
        paramDecls->at(i)->contextAnalysis(declarationCollection);
    }

    ASSERT(declarationCollection.varIdInFrame == 0);
    declarationCollection.functionDeclaration = this;
    if(statements)
    {
        statements->contextAnalysisStatements(declarationCollection);

        if(statements->isReturnStatement())
        {
            if(!returnVarDecl)
                throw ContextAnalysisException(ident->location, "return statement in void function");
        }
        else
            if(returnVarDecl)
                throw ContextAnalysisException(ident->location, "no return statement in non void function");
    }

    declarationCollection.leave();

    declarationCollection.add(ident->name, this);
}

void FunctionDeclaration::call(Stack2 &stack)
{
    PUSH_CALL_TREE(stack.executionContext->callTree, ident->name);

    ASSERT(statements);
    stack.pushFrame();

    statements->execute(stack);
    ASSERT(stack.interupt == Stack2::None || stack.interupt == Stack2::Return);
    stack.interupt = Stack2::None;

    for(size_t i = 0; i < paramDecls->size(); i++)
        stack.popPtr();
    stack.popFrame();

    POP_CALL_TREE(stack.executionContext->callTree);
}

