/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../tools/asserts.h"
#include "../declarationCollection.h"
#include "../identifier.h"
#include "classDeclaration.h"
#include "variableDeclaration.h"
#include "functionDeclaration.h"
#include "enumDeclaration.h"
#include "moduleDeclaration.h"
#include "valueDeclaration.h"

using namespace ast;

ModuleDeclaration::ModuleDeclaration(const char *name)
    : Declaration(new Identifier(name, Location::builtInLocation))
{
    ASSERT(ident);
}

ModuleDeclaration::~ModuleDeclaration()
{
    for(size_t i = 0; i < functionDeclarations.size(); i++)
        delete functionDeclarations.at(i);
    functionDeclarations.clear();

    for(size_t i = 0; i < valueDeclarations.size(); i++)
        delete valueDeclarations.at(i);
    valueDeclarations.clear();

    localDeclarationCollection.clear();
}

void ModuleDeclaration::contentContextAnalysis(DeclarationCollection &declarationCollection)
{
    for(size_t i = 0; i < valueDeclarations.size(); i++)
        valueDeclarations.at(i)->contextAnalysis(declarationCollection);

    for(size_t i = 0; i < functionDeclarations.size(); i++)
        functionDeclarations.at(i)->contextAnalysis(declarationCollection);
}

void ModuleDeclaration::contextAnalysis(DeclarationCollection &declarationCollection)
{
    declarationCollection.add(ident->name, this);
    declarationCollection.moduleDeclaration = this;

    declarationCollection.enter();

    this->contentContextAnalysis(declarationCollection);

    localDeclarationCollection = declarationCollection;
    if(ident->name != "_master_")
        localDeclarationCollection.stripToTopLevel();

    declarationCollection.moduleDeclaration = 0;

    declarationCollection.leave();
}

/*
 * EXTERN MODULE
 */

ExternModuleDeclaration::~ExternModuleDeclaration()
{
    for(size_t i = 0; i < classDeclarations.size(); i++)
        delete classDeclarations.at(i);
    classDeclarations.clear();
    for(size_t i = 0; i < enumDeclarations.size(); i++)
        delete enumDeclarations.at(i);
    enumDeclarations.clear();
}

void ExternModuleDeclaration::contentContextAnalysis(DeclarationCollection &declarationCollection)
{
    for(size_t i = 0; i < classDeclarations.size(); i++)
        classDeclarations.at(i)->contextAnalysis(declarationCollection);
    for(size_t i = 0; i < enumDeclarations.size(); i++)
        enumDeclarations.at(i)->contextAnalysis(declarationCollection);

    ModuleDeclaration::contentContextAnalysis(declarationCollection);
}

/*
 * MASTER MODULE
 */

MasterModuleDeclaration::MasterModuleDeclaration()
    : ExternModuleDeclaration("_master_"),
      mainCallStatement(0),
      mainModule(0),
      inputDeclaration(0),
      outputDeclaration(0)
{}

MasterModuleDeclaration::~MasterModuleDeclaration()
{
    //delete in reverse order!
    //this is necessary since the script modules might use
    //class declarations from the c++ modules added before
    //and need the class declarations when destroying the
    //stack of unfinished tasks
    for(int i = moduleDeclarations.size() - 1; i >= 0 ; i--)
        delete moduleDeclarations.at(i);
    moduleDeclarations.clear();

    if(mainCallStatement)
        delete mainCallStatement;
    if(mainModule)
        delete mainModule;
    if(inputDeclaration)
        delete inputDeclaration;
    if(outputDeclaration)
        delete outputDeclaration;
}

void MasterModuleDeclaration::contentContextAnalysis(DeclarationCollection &declarationCollection)
{
    ExternModuleDeclaration::contentContextAnalysis(declarationCollection);

    for(size_t i = 0; i < moduleDeclarations.size(); i++)
        moduleDeclarations.at(i)->contextAnalysis(localDeclarationCollection);

    if(mainModule)
        mainModule->contextAnalysis(localDeclarationCollection);
    if(mainCallStatement)
        mainCallStatement->contextAnalysis(localDeclarationCollection);
    if(inputDeclaration)
        inputDeclaration->contextAnalysis(localDeclarationCollection);
    if(outputDeclaration)
        outputDeclaration->contextAnalysis(localDeclarationCollection);
}

void MasterModuleDeclaration::execute(Stack2 &stack)
{
    mainCallStatement->execute(stack);
}

