/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <string.h>
#include "../../contextAnalysisException.h"
#include "../../statements/statementBlock.h"
#include "../variableDeclaration.h"
#include "../../identifier.h"
#include "typeToName.h"
#include "funcWrappers.h"

FuncWrapper::FuncWrapper(ast::Identifier *identifier, FunctorBase *functor)
    : FunctionDeclaration(identifier,
                          0, //returnType
                          new ast::FunctionDeclaration::ParameterDeclaration,
                          0,
                          Location::builtInLocation,
                          functor->isConst),
      functor(functor)
{}

void FuncWrapper::generateSyntaxTree()
{
    ASSERT(!returnVarDecl);
    const char *returnTypeIdName = functor->getTypeIdName(0);
    if(returnTypeIdName)
    {
        const TypeToName::TypeToNameMap::mapped_type *returnInfo = TypeToName::getTypeInfo(returnTypeIdName);
        ast::ResolvableIdentifier *typeIdent = new ast::ResolvableIdentifier(returnInfo->second, Location::builtInLocation);
        if(returnInfo->first != "_master_")
            typeIdent->pushPath(returnInfo->first);
        returnVarDecl = new ast::VariableDeclaration(
                            new ast::Identifier("_return_value_", Location::builtInLocation),
                            typeIdent,
                            0//initializeStatement
                            );
    }

    ASSERT(paramDecls->size() == 0);
    std::string ident("a");
    //in FullBScriptFunctino the last parameter is ExecutioContext, don't insert it into the syntax tree
    int lastParam = isFullBScriptFunction() ? functor->numOfParams : functor->numOfParams + 1;
    for(int i = 1; i < lastParam; i++)
    {
        const char* paramTypeIdName = functor->getTypeIdName(i);
        ASSERT(paramTypeIdName);

        const TypeToName::TypeToNameMap::mapped_type *paramInfo = TypeToName::getTypeInfo(paramTypeIdName);
        ast::ResolvableIdentifier *typeIdent = new ast::ResolvableIdentifier(paramInfo->second, Location::builtInLocation);
        if(paramInfo->first != "_master_")
            typeIdent->pushPath(paramInfo->first);
        ast::VariableDeclaration *decl;
        decl = new ast::VariableDeclaration(new ast::Identifier(ident, Location::builtInLocation),
                                            typeIdent,
                                            0); 
        paramDecls->push_back(decl);
        ident[0]++;
    }
}

void FuncWrapper::contextAnalysis(DeclarationCollection &declarationCollection)
{
    try
    {
        generateSyntaxTree();
    }
    catch(TypeToName::TypeNotRegisteredException &e)
    {
        throw ContextAnalysisException(getLocation(), e.what());
    }

    //if fullBScriptFunction -> last param must be ExecutionContext
    if(isFullBScriptFunction() && strcmp(typeid(ExecutionContext*).name(), functor->getTypeIdName(functor->numOfParams)))
        throw ContextAnalysisException(Location::builtInLocation, "last parameter of FullBScript function %s is not ExecutionContext", ident->name.c_str());
    FunctionDeclaration::contextAnalysis(declarationCollection);
}

TaskWrapperBase::TaskWrapperBase(ast::Identifier *identifier, FunctorBase *functor)
    : TaskDeclarationBase(identifier,
                          new ast::FunctionDeclaration::ParameterDeclaration,
                          0,
                          Location::builtInLocation),
      functor(functor)
{}

void TaskWrapperBase::generateSyntaxTree()
{
    ASSERT(!returnVarDecl);
    const char *returnTypeIdName = functor->getTypeIdName(0);
    if(returnTypeIdName)
    {
        const TypeToName::TypeToNameMap::mapped_type *returnInfo = TypeToName::getTypeInfo(returnTypeIdName);
        ast::ResolvableIdentifier *typeIdent = new ast::ResolvableIdentifier(returnInfo->second, Location::builtInLocation);
        if(returnInfo->first != "_master_")
            typeIdent->pushPath(returnInfo->first);
        returnVarDecl = new ast::VariableDeclaration(
                            new ast::Identifier("_return_value_", Location::builtInLocation),
                            typeIdent,
                            0//initializeStatement
                            );
    }

    ASSERT(paramDecls->size() == 0);
    std::string ident("a");
    //last parameter is ExecutioContext, don't insert it into the syntax tree
    for(int i = 1; i < functor->numOfParams; i++)
    {
        const char* paramTypeIdName = functor->getTypeIdName(i);
        ASSERT(paramTypeIdName);

        const TypeToName::TypeToNameMap::mapped_type *paramInfo = TypeToName::getTypeInfo(paramTypeIdName);
        ast::ResolvableIdentifier *typeIdent = new ast::ResolvableIdentifier(paramInfo->second, Location::builtInLocation);
        if(paramInfo->first != "_master_")
            typeIdent->pushPath(paramInfo->first);
        ast::VariableDeclaration *decl;
        decl = new ast::VariableDeclaration(new ast::Identifier(ident, Location::builtInLocation),
                                            typeIdent,
                                            0); 
        paramDecls->push_back(decl);
        ident[0]++;
    }
}

void TaskWrapperBase::contextAnalysis(DeclarationCollection &declarationCollection)
{
    try
    {
        generateSyntaxTree();
    }
    catch(TypeToName::TypeNotRegisteredException &e)
    {
        throw ContextAnalysisException(getLocation(), e.what());
    }

    if(strcmp(typeid(ExecutionContext*).name(), functor->getTypeIdName(functor->numOfParams)))
        throw ContextAnalysisException(Location::builtInLocation, "last parameter of task %s is not ExecutionContext", ident->name.c_str());
    if(returnVarDecl)
        throw ContextAnalysisException(Location::builtInLocation, "task %s is not a void function", ident->name.c_str());

    TaskDeclarationBase::contextAnalysis(declarationCollection);
}

