/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "../../../tools/asserts.h"
#include "../../../tools/functors.h"
#include "../../../stack.h"
#include "../../identifier.h"
#include "../moduleDeclaration.h"
#include "../functionDeclaration.h"
#include "../taskDeclaration.h"

class FuncWrapper : public ast::FunctionDeclaration
{
    public:
        FuncWrapper(ast::Identifier *identifier, FunctorBase *functor);

        virtual ~FuncWrapper()
        {
            delete functor;
        }

        virtual void generateSyntaxTree();
        virtual void contextAnalysis(DeclarationCollection &declarationCollection);
        virtual void call(Stack2 &stack) { functor->call(stack); }
        virtual bool isFullBScriptFunction() { return false; }
    protected:
        FunctorBase *functor;
};

class FullBScriptFuncWrapper : public FuncWrapper
{
    public:
        FullBScriptFuncWrapper(ast::Identifier *identifier, FunctorBase *f)
          : FuncWrapper(identifier, f)
        {}

        virtual void call(Stack2 &stack)
        {
            stack.pushPtr(&stack.executionContext);
            functor->call(stack);
        }

        virtual bool isFullBScriptFunction() { return true; }
};

class TaskWrapperBase : public ast::TaskDeclarationBase
{
    public:
        TaskWrapperBase(ast::Identifier *identifier, FunctorBase *functor);

        virtual ~TaskWrapperBase()
        {
            delete functor;
        }

        virtual void generateSyntaxTree();
        virtual void contextAnalysis(DeclarationCollection &declarationCollection);

        virtual void call(Stack2 &stack)
        {
            stack.pushPtr(&stack.executionContext);
            functor->call(stack);
        }

        virtual bool isNativeTask() { return true; }
        virtual const char* getContextTypeIdNamePtr() = 0;
        virtual void* getContextPointer() = 0;

    private:
        FunctorBase *functor;
};

template<typename F, typename C>
class TaskWrapper : public TaskWrapperBase
{
    public:
        C _taskContext;

        TaskWrapper(ast::Identifier *identifier, F f)
          : TaskWrapperBase(identifier, new FunctorVoid<F>(f))
        {}

        virtual const char* getContextTypeIdNamePtr() { return typeid(C).name(); }
        virtual void* getContextPointer() { return &_taskContext; }
};

template<typename F>
void registerFullBScriptFunctionVoid(ast::ModuleDeclaration *module,
                                     const char *name,
                                     F f,
                                     const char *cppName = "")
{
    FuncWrapper *wrapper = new FullBScriptFuncWrapper(new ast::Identifier(name, Location::builtInLocation, cppName), new FunctorVoid<F>(f));
    module->functionDeclarations.push_back(wrapper);
}

template<typename F>
void registerFullBScriptFunction(ast::ModuleDeclaration *module,
                                 const char *name,
                                 F f,
                                 const char *cppName = "")
{
    FuncWrapper *wrapper = new FullBScriptFuncWrapper(new ast::Identifier(name, Location::builtInLocation, cppName), new Functor<F>(f));
    module->functionDeclarations.push_back(wrapper);
}

template<typename F>
void registerFunctionVoid(ast::ModuleDeclaration *module,
                          const char *name,
                          F f,
                          const char *cppName = "")
{
    FuncWrapper *wrapper = new FuncWrapper(new ast::Identifier(name, Location::builtInLocation, cppName), new FunctorVoid<F>(f));
    module->functionDeclarations.push_back(wrapper);
}

template<typename F>
void registerFunction(ast::ModuleDeclaration *module,
                      const char *name,
                      F f,
                      const char *cppName = "")
{
    FuncWrapper *wrapper = new FuncWrapper(new ast::Identifier(name, Location::builtInLocation, cppName), new Functor<F>(f));
    module->functionDeclarations.push_back(wrapper);
}

template<typename C, typename F>
void registerTask(ast::ModuleDeclaration *module,
                  const char *name,
                  F f,
                  const char *cppName = "")
{
    TaskWrapperBase *wrapper = new TaskWrapper<F, C>(new ast::Identifier(name, Location::builtInLocation, cppName), f);
    module->functionDeclarations.push_back(wrapper);
}

