/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../stack.h"
#include "../declarationCollection.h"
#include "../identifier.h"
#include "../contextAnalysisException.h"
#include "../declarations/classDeclaration.h"
#include "../expressions/memberAccessExpression.h"
#include "../expressions/callExpression.h"
#include "binaryExpression.h"

using namespace ast;

const char *BinaryExpression::operationNames[] = {
    "_plus_",
    "_min_",
    "_mul_",
    "_div_",
    "_mod_",
    "_lt_",
    "_leq_",
    "_gt_",
    "_geq_",
    "_eq_",
    "_neq_",
    "_and_",
    "_or_",
};

BinaryExpression::BinaryExpression(Type type,
                                   Expression *leftExpr,
                                   Expression *rightExpr)
    : Expression(leftExpr->getLocation()),
      type(type),
      leftExpr(leftExpr),
      rightExpr(rightExpr),
      customOpCall(0)
{
    ASSERT(leftExpr);
    ASSERT(rightExpr);
}

BinaryExpression::~BinaryExpression()
{
    if(!customOpCall)
    {
        delete leftExpr;
        delete rightExpr;
    }
    else
        delete customOpCall;
}

void BinaryExpression::contextAnalysis(DeclarationCollection &declarationCollection)
{
    leftExpr->contextAnalysis(declarationCollection);
    rightExpr->contextAnalysis(declarationCollection);
    if(!leftExpr->getType() || !rightExpr->getType())
      throw ContextAnalysisException(getLocation(), "void expression in binary expression");

    ASSERT(leftExpr->getType()->getDeclarationType() == Declaration::Class);
    ASSERT(rightExpr->getType()->getDeclarationType() == Declaration::Class);
    ClassDeclaration *leftType = (ClassDeclaration*)leftExpr->getType(),
                     *rightType = (ClassDeclaration*)rightExpr->getType();

    if(leftType->ident->name == "int")
        dataType = int_;
    else if(leftType->ident->name == "float")
        dataType = float_;
    else if(leftType->ident->name == "bool")
        dataType = bool_;
    else
        dataType = other;

    switch(dataType)
    {
        case bool_:
            typeDecl = leftType;
            if(rightType != leftType)
                throw ContextAnalysisException(getLocation(), "expected bool expression");

            if(type < eq)
                throw ContextAnalysisException(getLocation(),
                                               "operator %s not defined for bool",
                                               operationNames[type]);
            break;
        case int_:
        case float_:
            //convert operant to float if necessary
            if(leftType->ident->name == "float" && rightType->ident->name == "int")
            {
                CallExpression::ParameterExpressions *params;
                params = new CallExpression::ParameterExpressions;
                params->push_back(rightExpr);
                rightExpr = new CallExpression(
                                new ResolvableIdentifier("toFloat", rightExpr->getLocation()),
                                params);
                rightExpr->contextAnalysis(declarationCollection);
                rightType = (ClassDeclaration*)rightExpr->getType();
            }
            else if(leftType->ident->name == "int" && rightType->ident->name == "float")
            {
                CallExpression::ParameterExpressions *params;
                params = new CallExpression::ParameterExpressions;
                params->push_back(leftExpr);
                leftExpr = new CallExpression(
                               new ResolvableIdentifier("toFloat", leftExpr->getLocation()),
                               params);
                leftExpr->contextAnalysis(declarationCollection);
                leftType = (ClassDeclaration*)leftExpr->getType();
                if(leftType->ident->name == "int")
                    dataType = int_;
                else if(leftType->ident->name == "float")
                    dataType = float_;
                else
                    ASSERT(false);
            }

            if(dataType == float_ && type == mod)
                throw ContextAnalysisException(getLocation(),
                                               "operator %s not defined for float",
                                               operationNames[type]);
            if(leftType != rightType)
                throw ContextAnalysisException(getLocation(), "not compatible types in binary expression");

            switch(type)
            {
                case plus:
                case minus:
                case mul:
                case div:
                case mod:
                    typeDecl = leftType;
                    break;
                case lt:
                case leq:
                case gt:
                case geq:
                case eq:
                case neq:
                    {
                        ResolvableIdentifier boolIdent("bool", Location::builtInLocation);
                        declarationCollection.resolve(&boolIdent);
                        typeDecl = (ClassDeclaration*)boolIdent.declaration;
                        break;
                    }
                default:
                    throw ContextAnalysisException(getLocation(),
                                                   "operator %s not defined for int/float",
                                                   operationNames[type]);
            }
            break;
        case other:
            {
                CallExpression::ParameterExpressions *params;
                params = new CallExpression::ParameterExpressions;
                params->push_back(rightExpr);
                customOpCall = new MemberAccessExpression(
                                   leftExpr,
                                   new CallExpression(
                                       new ResolvableIdentifier(operationNames[type], leftExpr->getLocation()),
                                       params
                                       )
                                   );
                try
                {
                    customOpCall->contextAnalysis(declarationCollection);
                }
                catch(ContextAnalysisException &e)
                {
                    throw ContextAnalysisException(getLocation(),
                                                   "can't resolve operator '%s' for class '%s'\n    because %s",
                                                   operationNames[type],
                                                   leftExpr->getType()->ident->name.c_str(),
                                                   e.what());
                }
                ASSERT(customOpCall->getType()->getDeclarationType() == Declaration::Class);
                typeDecl = (ClassDeclaration*)customOpCall->getType();
            }
            break;
        default:
            ASSERT(false);
    }
}

void BinaryExpression::execute(Stack2 &stack)
{
    if(!customOpCall)
    {
        leftExpr->execute(stack);
        rightExpr->execute(stack);

        void *rightPtr = stack.popPtr(),
             *leftPtr = stack.popPtr();
        typeDecl->createInstance(stack, true);
        void *rPtr = stack.popPtr();
        stack.pushPtr(rPtr);

        switch(dataType)
        {
            case int_:
                {
                    int l = *(int*)leftPtr,
                        r = *(int*)rightPtr;
                    switch(type)
                    {
                        case plus:
                            *(int*)rPtr = l + r;
                            break;
                        case minus:
                            *(int*)rPtr = l - r;
                            break;
                        case mul:
                            *(int*)rPtr = l * r;
                            break;
                        case div:
                            *(int*)rPtr = l / r;
                            break;
                        case mod:
                            *(int*)rPtr = l % r;
                            break;
                        case gt:
                            *(int*)rPtr = l > r;
                            break;
                        case geq:
                            *(int*)rPtr = l >= r;
                            break;
                        case lt:
                            *(int*)rPtr = l < r;
                            break;
                        case leq:
                            *(int*)rPtr = l <= r;
                            break;
                        case eq:
                            *(bool*)rPtr = l == r;
                            break;
                        case neq:
                            *(bool*)rPtr = l != r;
                            break;
                        default:
                            ASSERT(false);
                    }
                }
                break;
            case float_:
                {
                    float l = *(float*)leftPtr,
                          r = *(float*)rightPtr;
                    switch(type)
                    {
                        case plus:
                            *(float*)rPtr = l + r;
                            break;
                        case minus:
                            *(float*)rPtr = l - r;
                            break;
                        case mul:
                            *(float*)rPtr = l * r;
                            break;
                        case div:
                            *(float*)rPtr = l / r;
                            break;
                        case gt:
                            *(bool*)rPtr = l > r;
                            break;
                        case geq:
                            *(bool*)rPtr = l >= r;
                            break;
                        case lt:
                            *(bool*)rPtr = l < r;
                            break;
                        case leq:
                            *(bool*)rPtr = l <= r;
                            break;
                        case eq:
                            *(bool*)rPtr = l == r;
                            break;
                        case neq:
                            *(bool*)rPtr = l != r;
                            break;
                        default:
                            ASSERT(false);
                    }
                }
                break;
            case bool_:
                {
                    bool l = *(bool*)leftPtr,
                         r = *(bool*)rightPtr;
                    switch(type)
                    {
                        case eq:
                            *(bool*)rPtr = l == r;
                            break;
                        case neq:
                            *(bool*)rPtr = l != r;
                            break;
                        case and_:
                            *(bool*)rPtr = l && r;
                            break;
                        case or_:
                            *(bool*)rPtr = l || r;
                            break;
                        default:
                            ASSERT(false);
                    }
                }
                break;
            default:
                ASSERT(false);
        }
    }
    else
        customOpCall->execute(stack);
}

void BinaryExpression::executeCleanUp(Stack2 &stack)
{
    if(!customOpCall)
    {
        typeDecl->deleteInstance(stack);
        rightExpr->executeCleanUp(stack);
        leftExpr->executeCleanUp(stack);
    }
    else
        customOpCall->executeCleanUp(stack);

}

