/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../tools/asserts.h"
#include "../declarationCollection.h"
#include "../declarations/classDeclaration.h"
#include "../declarations/functionDeclaration.h"
#include "../declarations/variableDeclaration.h"
#include "../identifier.h"
#include "../contextAnalysisException.h"
#include "callExpression.h"

using namespace ast;

CallExpression::ParameterExpressions::~ParameterExpressions()
{
    for(size_t i = 0; i < size(); i++)
        delete at(i);
    clear();
}

CallExpression::CallExpression(ResolvableIdentifier *ident,
                               ParameterExpressions *paramExprs)
    : Expression(ident->location),
      ident(ident),
      paramExprs(paramExprs)
{
    ASSERT(ident);
    ASSERT(paramExprs);
}

CallExpression::~CallExpression()
{
    delete ident;

    delete paramExprs;
}

Declaration* CallExpression::getType()
{
    ASSERT(ident->declaration);

    VariableDeclaration *rDecl = ((FunctionDeclaration*)ident->declaration)->returnVarDecl;
    return rDecl ? rDecl->typeIdent->declaration : 0;
}

void CallExpression::contextAnalysis(DeclarationCollection &declarationCollection)
{
    contextAnalysis(declarationCollection, declarationCollection);
}

void CallExpression::contextAnalysis(DeclarationCollection &declarationCollection,
                                     DeclarationCollection &memberDeclarationCollection)
{
    //resolve function identifier
    if(!memberDeclarationCollection.resolve(ident))
        throw ContextAnalysisException(ident->location,
                                       "can't resolve function identifier %s",
                                       ident->fullName().c_str());
    else if(ident->declaration->getDeclarationType() != Declaration::Function
         && ident->declaration->getDeclarationType() != Declaration::Task)
        throw ContextAnalysisException(ident->location,
                                       "function identifier '%s' resolved to non function declaration",
                                       ident->fullName().c_str());

    FunctionDeclaration *decl = (FunctionDeclaration*)ident->declaration;

    //check parameters
    if(paramExprs->size() != decl->paramDecls->size())
        throw ContextAnalysisException(ident->location,
                                       "number of parameters to function %s mismatch",
                                       ident->fullName().c_str());
    for(size_t i = 0; i < paramExprs->size(); i++)
    {
        Expression *expr = paramExprs->at(i);
        expr->contextAnalysis(declarationCollection);
        if(!expr->isRValue())
            throw ContextAnalysisException(expr->getLocation(), "expression is not readable");
        if(!expr->getType())
            throw ContextAnalysisException(expr->getLocation(), "unexpected void expression");
        if(expr->getType() != decl->paramDecls->at(i)->typeIdent->declaration)
            throw ContextAnalysisException(expr->getLocation(),
                                           "parameter mismatch %s param #%d is %s instead of %s",
                                           ident->fullName().c_str(),
                                           i+1,
                                           expr->getType()->ident->name.c_str(),
                                           decl->paramDecls->at(i)->typeIdent->declaration->ident->name.c_str());
    }
}

void CallExpression::execute(Stack2 &stack)
{
    FunctionDeclaration *decl = (FunctionDeclaration*)ident->declaration;

    if(decl->returnVarDecl)
        decl->returnVarDecl->execute(stack);

    const size_t numOfParamExprs = paramExprs->size();
    for(size_t i = 0; i < numOfParamExprs; i++)
        paramExprs->at(i)->execute(stack);

    decl->call(stack);

    for(size_t i = 0; i < numOfParamExprs; i++)
        paramExprs->at(numOfParamExprs - 1 - i)->executeCleanUp(stack);
}

void CallExpression::executeCleanUp(Stack2 &stack)
{
    FunctionDeclaration *decl = (FunctionDeclaration*)ident->declaration;
    if(decl->returnVarDecl)
        ((ClassDeclaration*)decl->returnVarDecl->typeIdent->declaration)->deleteInstance(stack);
}

