/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../tools/asserts.h"
#include "../declarations/classDeclaration.h"
#include "../declarations/functionDeclaration.h"
#include "../declarations/moduleDeclaration.h"
#include "../declarations/enumDeclaration.h"
#include "../expressions/callExpression.h"
#include "../contextAnalysisException.h"
#include "../identifier.h"
#include "memberAccessExpression.h"

using namespace ast;

MemberAccessExpression::MemberAccessExpression(Expression *leftExpr,
                                               Expression *rightExpr)
    : Expression(leftExpr->getLocation()),
      leftExpr(leftExpr),
      rightExpr(rightExpr)
{
    ASSERT(leftExpr);
    ASSERT(rightExpr);
}

MemberAccessExpression::~MemberAccessExpression()
{
    delete leftExpr;
    delete rightExpr;
}

Declaration* MemberAccessExpression::getType()
{
    return rightExpr->getType();
}

void MemberAccessExpression::contextAnalysis(DeclarationCollection &declarationCollection)
{
    leftExpr->contextAnalysis(declarationCollection);
    if(!leftExpr->getType())
        throw ContextAnalysisException(getLocation(), "unexpected void expression");
    if(leftExpr->getType()->getDeclarationType() == Declaration::Class)
        rightExpr->contextAnalysis(declarationCollection,
                                   ((ClassDeclaration*)leftExpr->getType())->localDeclarationCollection);
    else
        throw ContextAnalysisException(getLocation(), "expression is not of class type, thus it can't be accessed via '.'");

    //check const correctness for member function calls
    if(rightExpr->getExpressionType() == Expression::FunctionCall)
        if(leftExpr->getType()->getDeclarationType() == Declaration::Class)
        {
            FunctionDeclaration *funcDecl;
            funcDecl = (FunctionDeclaration*)((CallExpression*)rightExpr)->ident->declaration;
            if(!funcDecl->isConst)
                if(leftExpr->isConst())
                    throw ContextAnalysisException(leftExpr->getLocation(),
                                                   "call to '%s' discards object qualifiers\n",
                                                   funcDecl->ident->name.c_str());
        }
}

void MemberAccessExpression::execute(Stack2 &stack)
{
    leftExpr->execute(stack);
    rightExpr->execute(stack);
}

void MemberAccessExpression::executeCleanUp(Stack2 &stack)
{
    rightExpr->executeCleanUp(stack);
    leftExpr->executeCleanUp(stack);
}

bool MemberAccessExpression::isLValue()
{
    return leftExpr->isLValue() && rightExpr->isLValue();
}

bool MemberAccessExpression::isRValue()
{
    return leftExpr->isRValue() && rightExpr->isRValue();
}

