/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../stack.h"
#include "../declarationCollection.h"
#include "../contextAnalysisException.h"
#include "../identifier.h"
#include "../declarations/classDeclaration.h"
#include "../expressions/memberAccessExpression.h"
#include "../expressions/callExpression.h"
#include "unaryExpression.h"

using namespace ast;

UnaryExpression::UnaryExpression(Type type, Expression *expr)
    : Expression(expr->getLocation()),
      type(type),
      expr(expr),
      customOpCall(0)
{
    ASSERT(expr);
}

UnaryExpression::~UnaryExpression()
{
    if(!customOpCall)
        delete expr;
    else
        delete customOpCall;
}

void UnaryExpression::contextAnalysis(DeclarationCollection &declarationCollection)
{
    expr->contextAnalysis(declarationCollection);
    if(!expr->getType())
      throw ContextAnalysisException(getLocation(), "void expression in unary expression");
    ASSERT(expr->getType()->getDeclarationType() == Declaration::Class);
    typeDecl = (ClassDeclaration*)expr->getType();


    if(type == not_)
    {
        if(typeDecl->ident->name == "bool")
            dataType = bool_;
        else
        {
            customOpCall = new MemberAccessExpression(
                               expr,
                               new CallExpression(
                                   new ResolvableIdentifier("_not_", expr->getLocation()),
                                   new CallExpression::ParameterExpressions
                                   )
                               );
            try
            {
                customOpCall->contextAnalysis(declarationCollection);
            }
            catch(ContextAnalysisException &e)
            {
                throw ContextAnalysisException(getLocation(),
                                               "can't resolve operator _not_ for class %s\n    because: %s\n",
                                               typeDecl->ident->name.c_str(),
                                               e.what());
            }
            typeDecl = (ClassDeclaration*)customOpCall->getType();
        }
    }
    else if(type == minus)
    {
        if(typeDecl->ident->name == "float")
            dataType = float_;
        else if(typeDecl->ident->name == "int")
            dataType = int_;
        else
        {
            customOpCall = new MemberAccessExpression(
                               expr,
                               new CallExpression(
                                   new ResolvableIdentifier("_unmin_", Location::builtInLocation),
                                   new CallExpression::ParameterExpressions
                                   )
                               );
            try
            {
                customOpCall->contextAnalysis(declarationCollection);
            }
            catch(ContextAnalysisException &e)
            {
                throw ContextAnalysisException(getLocation(),
                                               "can't resolve operator _unmin_ for class %s\n    because: %s\n",
                                               typeDecl->ident->name.c_str(),
                                               e.what());
            }
            typeDecl = (ClassDeclaration*)customOpCall->getType();
        }
    }
}

void UnaryExpression::execute(Stack2 &stack)
{
    if(!customOpCall)
    {
        expr->execute(stack);
        void *ptr = stack.popPtr();
        typeDecl->createInstance(stack, true);
        void *rPtr = stack.popPtr();
        stack.pushPtr(rPtr);
        if(type == not_)
            *(bool*)rPtr = !(*(bool*)ptr);
        else if(type == minus)
        {
            if(dataType == float_)
                *(float*)rPtr = -(*(float*)ptr);
            else if(dataType == int_)
                *(int*)rPtr = -(*(int*)ptr);
            else
                ASSERT(false);
        }
        else
            ASSERT(false);
    }
    else
        customOpCall->execute(stack);
}

void UnaryExpression::executeCleanUp(Stack2 &stack)
{
    if(!customOpCall)
    {
        typeDecl->deleteInstance(stack);
        expr->executeCleanUp(stack);
    }
    else
        customOpCall->executeCleanUp(stack);

}

