/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include "../../tools/asserts.h"
#include "../../stack.h"
#include "../../executionContext.h"
#include "../declarationCollection.h"
#include "../contextAnalysisException.h"
#include "../identifier.h"
#include "../declarations/variableDeclaration.h"
#include "../declarations/taskDeclaration.h"
#include "../expressions/expression.h"
#include "../statements/statementBlock.h"
#include "../statements/yieldStatement.h"
#include "forStatement.h"

using namespace ast;

ForStatement::ForStatement(VariableDeclaration *varDecl,
                           Expression *condition,
                           Statement *postStatement,
                           StatementBlock *statementBlock,
                           int maxLoopsPerFrame)
    : Statement(condition->getLocation()),
      varDecl(varDecl),
      postStatement(postStatement),
      condition(condition),
      statementBlock(statementBlock),
      maxLoopsPerFrame(maxLoopsPerFrame)
{
    ASSERT(statementBlock);
    ASSERT(condition);
}

ForStatement::~ForStatement()
{
    if(varDecl)
        delete varDecl;
    if(postStatement)
        delete postStatement;
    delete condition;
    delete statementBlock;
}

void ForStatement::contextAnalysis(DeclarationCollection &declarationCollection)
{
    declarationCollection.enter();
    declarationCollection.forStatementStack.push_back(this);

    if(varDecl)
        varDecl->contextAnalysis(declarationCollection);

    condition->contextAnalysis(declarationCollection);
    if(condition->getType()->ident->name != "bool")
        throw ContextAnalysisException(getLocation(), "non bool expression in for statement condition");

    statementBlock->contextAnalysis(declarationCollection);

    if(postStatement)
        postStatement->contextAnalysis(declarationCollection);

    if(varDecl)
        declarationCollection.varIdInFrame--;

    if(maxLoopsPerFrame <= 0)
        throw ContextAnalysisException(getLocation(), "non positive max loops per frame count");

    emergencyYieldEnabled = declarationCollection.functionDeclaration->getDeclarationType() == Declaration::Task;

    ASSERT(declarationCollection.forStatementStack.back() == this);
    declarationCollection.forStatementStack.pop_back();
    declarationCollection.leave();
}

void ForStatement::execute(Stack2 &stack)
{
    if(stack.owner && stack.owner->state == TaskDeclarationBase::restart)
        executionStuff.internalState = 0;

    bool b;
    YieldStatement yield(Location::builtInLocation);
    switch(executionStuff.internalState)
    {
        case 0:
        if(varDecl)
            varDecl->execute(stack);

        executionStuff.lastYield = 0;
        for(executionStuff.i = 0; ; executionStuff.i++)
        {
            condition->execute(stack);
            b = *(bool*)stack.popPtr();
            condition->executeCleanUp(stack);
            if(!b)
                break;

            if(emergencyYieldEnabled)
            {
                if(executionStuff.lastYield != stack.yieldCounter)
                {
                    executionStuff.i = 0;
                    executionStuff.lastYield = stack.yieldCounter;
                }

                if(maxLoopsPerFrame - executionStuff.i == 0)
                {
                    stack.executionContext->errorHandler("ERROR!!!! EMERGENCY YIELD STOPS THIS LOOP!\n  %s", condition->getLocation().toString().c_str());
                    yield.execute(stack);
                    executionStuff.internalState = 1;
                    return;
                    case 1:
                    continue;
                }
            }

            executionStuff.internalState = 2;
            case 2:
            statementBlock->execute(stack);
            switch(stack.interupt)
            {
                case Stack2::None:
                    break;
                case Stack2::Continue:
                    stack.interupt = Stack2::None;
                    break;
                case Stack2::Break:
                    stack.interupt = Stack2::None;
                case Stack2::Return:
                    goto breakOutter;
                    break;
                case Stack2::Yield:
                    return;
            }

            if(postStatement)
            {
                postStatement->execute(stack);
                if(stack.interupt == Stack2::Yield)
                {
                    executionStuff.internalState = 3;
                    return;
                }
                case 3:
                ;
            }
        }
    breakOutter:
        if(varDecl)
            varDecl->executeCleanUp(stack);
    }
    executionStuff.internalState = 0;
}

