/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../../tools/asserts.h"
#include "../../stack.h"
#include "../declarationCollection.h"
#include "../contextAnalysisException.h"
#include "../identifier.h"
#include "../declarations/classDeclaration.h"
#include "../declarations/taskDeclaration.h"
#include "../expressions/expression.h"
#include "statementBlock.h"
#include "ifStatement.h"

using namespace ast;

IfStatement::IfStatement(Expression *expression,
                         StatementBlock *ifStatements,
                         StatementBlock *elseStatements)
    : Statement(expression->getLocation()),
      expression(expression),
      ifStatements(ifStatements),
      elseStatements(elseStatements),
      internalState(0)
{
    ASSERT(expression);
    ASSERT(ifStatements);
}

IfStatement::~IfStatement()
{
    delete expression;
    delete ifStatements;
    if(elseStatements)
        delete elseStatements;
}

void IfStatement::contextAnalysis(DeclarationCollection &declarationCollection)
{
    expression->contextAnalysis(declarationCollection);
    if(!expression->isRValue())
        throw ContextAnalysisException(getLocation(), "expression is not readable");

    if(!expression->getType() || expression->getType()->ident->name != "bool")
        throw ContextAnalysisException(getLocation(), "expected bool expression");

    declarationCollection.enter();
    ifStatements->contextAnalysis(declarationCollection);
    declarationCollection.leave();

    if(elseStatements)
    {
        declarationCollection.enter();
        elseStatements->contextAnalysis(declarationCollection);
        declarationCollection.leave();
    }
}

void IfStatement::execute(Stack2 &stack)
{
    if(stack.owner && stack.owner->state == TaskDeclarationBase::restart)
        internalState = 0;

    bool b;
    switch(internalState)
    {
        case 0:
        expression->execute(stack);
        b = *(bool*)stack.popPtr();
        expression->executeCleanUp(stack);

        if(b)
        {
            internalState = 1;
            case 1:
            ifStatements->execute(stack);
            if(stack.interupt == Stack2::Yield)
                return;
        }
        else if(elseStatements)
        {
            internalState = 2;
            case 2:
            elseStatements->execute(stack);
            if(stack.interupt == Stack2::Yield)
                return;
        }
    }
    internalState = 0;
}

bool IfStatement::isReturnStatement()
{
    if(!elseStatements)
        return false;

    return ifStatements->isReturnStatement() && elseStatements->isReturnStatement();
}

