/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <stdarg.h>

#ifndef WIN32
#include <dlfcn.h>
#endif

#ifdef WIN32
#include "tools/win32snprintf.h"
#endif

#include "parser/parserInterface.h"
#include "ast/declarationCollection.h"
#include "ast/contextAnalysisException.h"
#include "ast/declarations/moduleDeclaration.h"
#include "ast/declarations/native/moduleCreatorBase.h"
#include "ast/declarations/native/funcWrappers.h"
#include "builtins/builtins.h"

#include "bScriptEngine.h"

using namespace ast;

void defaultErrorHandler(const char *format, ...)
{
    #define bufsize 512
    char buf[bufsize+1];

    va_list ap;
    va_start(ap, format);	
    vsnprintf(buf, bufsize, format, ap);
    va_end(ap);

    fprintf(stderr, "b-script ERROR: %s\n", buf);
}

BScriptEngine::BScriptEngine(const std::string &rootDir, ErrorHandler errorHandler)
    : executionContext(errorHandler),
      mainStack(new Stack2),
      masterModule(0),
      rootDir(rootDir)
{}

BScriptEngine::~BScriptEngine()
{
    if(masterModule)
        delete masterModule;

    for(size_t i = 0; i < externModuleHandles.size(); i++)
#ifndef WIN32
        VERIFY(!dlclose(externModuleHandles.at(i)));
#else
        VERIFY(FreeLibrary(externModuleHandles.at(i)));
#endif
    externModuleHandles.clear();

    delete mainStack;
}

bool BScriptEngine::init()
{
    if(masterModule)
        delete masterModule;

    for(size_t i = 0; i < externModuleHandles.size(); i++)
#ifndef WIN32
        VERIFY(!dlclose(externModuleHandles.at(i)));
#else
        VERIFY(FreeLibrary(externModuleHandles.at(i)));
#endif
    externModuleHandles.clear();

    executionContext.taskContextMap.clear();

    moduleMap.clear();
    masterModule = new MasterModuleDeclaration();

    bindBuiltins(masterModule);

    DeclarationCollection declarationCollection;
    try
    {
        masterModule->contextAnalysis(declarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        executionContext.errorHandler("builtins->contextAnalysis!\n%s", e.what());
#ifdef DEBUG
        declarationCollection.clear();
#endif
        return false;
    }

    return true;
}

bool BScriptEngine::loadModuleLibrary(const std::string &filename, bool global)
{
    LibraryHandle handle;

#ifndef WIN32
    int flags = RTLD_NOW;
    if(global)
        flags = flags | RTLD_GLOBAL;

    handle = dlopen(filename.c_str(), flags);
#else
    handle = LoadLibrary(filename.c_str());
#endif

    if(handle)
        externModuleHandles.push_back(handle);
    else
    {
        char *errorMsg = 0;
#ifndef WIN32
        errorMsg = dlerror();
#endif
        if(errorMsg)
            executionContext.errorHandler("%s", errorMsg);
        else
            executionContext.errorHandler("unknown error or running on windoof");
    }
    return handle != 0;
}

bool BScriptEngine::setInput(const std::string &moduleName,
                             const std::string &typeName,
                             void *ptr)
{
    ASSERT(masterModule);

    ResolvableIdentifier *typeIdent = new ResolvableIdentifier(typeName, Location::builtInLocation);
    if(moduleName != "_master_")
        typeIdent->pushPath(moduleName);

    masterModule->inputDeclaration = new GlobalVariableDeclaration(new Identifier("input", Location::builtInLocation),
                                                                   typeIdent,
                                                                   ptr,
                                                                   VariableDeclaration::readable);
    try
    {
        masterModule->inputDeclaration->contextAnalysis(masterModule->localDeclarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        delete masterModule->inputDeclaration;
        masterModule->inputDeclaration = 0;
        executionContext.errorHandler("while setting input\n%s", e.what());
        return false;
    }

    return true;
}

bool BScriptEngine::setOutput(const std::string &moduleName,
                              const std::string &typeName,
                              void *ptr)
{
    ASSERT(masterModule);

    ResolvableIdentifier *typeIdent = new ResolvableIdentifier(typeName, Location::builtInLocation);
    if(moduleName != "_master_")
        typeIdent->pushPath(moduleName);

    masterModule->outputDeclaration = new GlobalVariableDeclaration(new Identifier("output", Location::builtInLocation),
                                                                    typeIdent,
                                                                    ptr,
                                                                    VariableDeclaration::writable);
    try
    {
        masterModule->outputDeclaration->contextAnalysis(masterModule->localDeclarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        delete masterModule->outputDeclaration;
        masterModule->outputDeclaration = 0;
        executionContext.errorHandler("while setting output\n%s", e.what());
        return false;
    }

    return true;
}

bool BScriptEngine::addModule(ModuleDeclaration *module)
{
    try
    {
        module->contextAnalysis(masterModule->localDeclarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        executionContext.errorHandler("%s", e.what());
        delete module;
        return false;
    }

    masterModule->moduleDeclarations.push_back(module);

    for(size_t i = 0; i < module->functionDeclarations.size(); i++)
    {
        TaskWrapperBase *wrapper = dynamic_cast<TaskWrapperBase*>(module->functionDeclarations.at(i));
        if(wrapper)
        {
            ASSERT(executionContext.taskContextMap.find(wrapper->getContextTypeIdNamePtr()) == executionContext.taskContextMap.end());
            executionContext.taskContextMap[wrapper->getContextTypeIdNamePtr()] = wrapper;
        }
    }
    return true;
}

std::string BScriptEngine::getModuleName(const std::string &filename)
{
    return filename.substr(filename.rfind("/")+1, filename.rfind(".") - (filename.rfind("/") + 1));
}

bool BScriptEngine::requires(const std::string &filename)
{
    if(!loadModule(filename))
        return false;

    masterModule->requirements.push_back(filename);
    return true;
}

bool BScriptEngine::loadModule(const std::string &filename)
{
    std::string modulename = getModuleName(filename);

    //check for cyclic requirements and already loaded modules
    ModuleMap::iterator it = moduleMap.find(modulename);
    if(it != moduleMap.end())
    {
        if(!it->second)
        {
            executionContext.errorHandler("cyclic requirements");
            return false;
        }
        else
            return true;
    }

    moduleMap[modulename] = false;

    ModuleDeclaration *module = 0;
    //load from *.bs file or native module
    if(ModuleCreatorBase::registeredCreators.find(modulename) == ModuleCreatorBase::registeredCreators.end())
    {
        //no corresponding native module registered -> load from *.bs file
        module = new ModuleDeclaration(modulename.c_str());
        ParserInterface p(executionContext.errorHandler);
        //p.setLexerDebugOut("tokens.debug");

        if(!p.parse(module, filename))
        {
            delete module;
            return false;
        }
    }
    else
    {
        //load module native module
        module = ModuleCreatorBase::registeredCreators[modulename]->createModule();
    }
    ASSERT(module);

    //load requirements
    std::string dir = filename.substr(0, filename.rfind("/"));
    for(size_t i = 0; i < module->requirements.size(); i++)
        if(!loadModule(dir + "/" + module->requirements.at(i).c_str()))
        {
            delete module;
            return false;
        }

    if(!addModule(module))
        return false;

    moduleMap[modulename] = true;

    return true;
}

bool BScriptEngine::setMainCallCode(const std::string &mainCallCode)
{
    if(!masterModule)
        return false;

    if(masterModule->mainCallStatement)
    {
        delete masterModule->mainCallStatement;
        masterModule->mainCallStatement = 0;
    }

    if(masterModule->mainModule)
    {
        delete masterModule->mainModule;
        masterModule->mainModule = 0;
    }

    //load main module
    ModuleDeclaration *mainModule = new ModuleDeclaration("_main_");
    ParserInterface p(executionContext.errorHandler);
    if(!p.parse(mainModule, mainCallCode.c_str()))
    {
        delete mainModule;
        return false;
    }

    //load requirements
    for(size_t i = 0; i < mainModule->requirements.size(); i++)
        if(!loadModule(rootDir + "/" + mainModule->requirements.at(i).c_str()))
        {
            delete mainModule;
            return false;
        }

    DeclarationCollection declarationCollection = masterModule->localDeclarationCollection;
    try
    {
        mainModule->contextAnalysis(declarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        executionContext.errorHandler("%s", e.what());
#ifdef DEBUG
        declarationCollection.clear();
#endif
        delete mainModule;
        return false;
    }
    masterModule->mainModule = mainModule;

    //insert call to main into masterModule
    ResolvableIdentifier *mainIdent = new ResolvableIdentifier("main", Location::builtInLocation);
    mainIdent->pushPath("_main_");
    CallStatement *mainCall =
                new CallStatement(
                      new CallExpression(
                          mainIdent,
                          new CallExpression::ParameterExpressions));

    try
    {
        mainCall->contextAnalysis(declarationCollection);
    }
    catch(ContextAnalysisException &e)
    {
        executionContext.errorHandler("%s", e.what());
#ifdef DEBUG
        declarationCollection.clear();
#endif
        delete mainCall;
        return false;
    }

    masterModule->mainCallStatement = mainCall;

#ifdef DEBUG
    declarationCollection.clear();
#endif

    return true;
}

bool BScriptEngine::execute()
{
    if(!masterModule || !masterModule->mainCallStatement)
        return false;

    executionContext.callTree.clear();
    executionContext.frameID++;
    if(masterModule->inputDeclaration)
        executionContext.input = masterModule->inputDeclaration->externPtr;
    if(masterModule->outputDeclaration)
        executionContext.output = masterModule->outputDeclaration->externPtr;
    mainStack->executionContext = &executionContext;
    unsigned callTreeStackSize = executionContext.callTree.stackSize();
    try
    {
        masterModule->execute(*mainStack);
    }
    catch(StackException &e)
    {
        executionContext.errorHandler("stack error in mainCallCode. Message: %s", e.msg.c_str());
        mainStack->clear();
        executionContext.callTree.trace("!!!Stack error!!!");
        while(executionContext.callTree.stackSize() > callTreeStackSize)
            executionContext.callTree.popCall();
    }
    return true;
}

bool BScriptEngine::generateCode(CodeGeneratorInterface &cg)
{
    if(!masterModule)
        return false;
    masterModule->generateCode(cg);
    return true;
}

