/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include "../ast/ast.h"
#include "cppCodeGenerator.h"

#ifdef WIN32
#include "../tools/win32snprintf.h"
#endif

std::string getVariableCppTypeIdent(const ast::VariableDeclaration &varDecl)
{
    std::string typeString;

    if(varDecl.typeIdent->declaration->ident->cppName.size() > 0)
        typeString = varDecl.typeIdent->declaration->ident->cppName;
    else
    {
        ASSERT(varDecl.typeIdent->path.size() <= 1);
        if(varDecl.typeIdent->path.size() == 1)
        {
            typeString = "bs_";
            typeString += varDecl.typeIdent->path.at(0) + "::";
        }
        typeString += varDecl.typeIdent->name;
    }

    return typeString;
}

void CppCodeGenerator::generateCode(ast::StatementBlock &statementBlock)
{
    printer->println("{");
    printer->indent();

    ast::StatementBlock::StatementList::iterator i;
    for(i = statementBlock.functionDeclarations.begin(); i != statementBlock.functionDeclarations.end(); i++)
    {
        ASSERT((*i)->getStatementType() == ast::Statement::DeclarationStatement);
        ast::Declaration *funcDecl = (ast::Declaration*)*i;
        ASSERT(!funcDecl->ident->cppName.size());
        funcDecl->ident->cppName = "_bsl_";
        char buf[32];
        snprintf(buf, 32, "%d", localFunctionID++);
        funcDecl->ident->cppName += buf;
        funcDecl->ident->cppName += "_";
        funcDecl->ident->cppName += funcDecl->ident->name;
        localFunctions.push_back(*i);
    }
    for(i = statementBlock.variableDeclarations.begin(); i != statementBlock.variableDeclarations.end(); i++)
    {
        isStatement = true;
        (*i)->generateCode(*this);
    }
    for(i = statementBlock.statements.begin(); i != statementBlock.statements.end(); i++)
    {
        isStatement = true;
        (*i)->generateCode(*this);
    }

    printer->dindent();
    printer->println("}");
}

void CppCodeGenerator::generateCode(ast::CallStatement &callStmt)
{
    callStmt.funcCall->generateCode(*this);
    printer->print(";");
    printer->newline();
}

void CppCodeGenerator::generateCode(ast::IfStatement &ifStmt)
{
    printer->print("if(");
    ifStmt.expression->generateCode(*this);
    printer->print(")");
    printer->newline();
    ifStmt.ifStatements->generateCode(*this);
    if(ifStmt.elseStatements)
    {
        printer->println("else");
        ifStmt.elseStatements->generateCode(*this);
    }
}

void CppCodeGenerator::generateCode(ast::ForStatement &forStmt)
{
    //push loop id
    continueLabelIDStack.push_back(continueLabelID++);

    //varDecl
    if(forStmt.varDecl)
        forStmt.varDecl->generateCode(*this);

    //init emergency yield
    if(isTask)
    {
        headerPrinter->println("unsigned _bs_for%d_lastYield;", continueLabelIDStack.back());
        headerPrinter->println("unsigned _bs_for%d_itersSinceLastYield;", continueLabelIDStack.back());
        printer->println("_taskContext->_bs_for%d_lastYield = 0;", continueLabelIDStack.back());
        printer->println("_taskContext->_bs_for%d_itersSinceLastYield = 0;", continueLabelIDStack.back());
    }

    //for( varDecl = init; condition;)
    printer->print("for(");
    isStatement = false;
    if(forStmt.varDecl)
        forStmt.varDecl->initializeStatement->generateCode(*this);
    printer->print("; ");
    forStmt.condition->generateCode(*this);
    printer->print("; ");
    printer->print(")");

    //body
    printer->newline();
    printer->println("{");
    printer->indent();

    //emergency yield
    if(isTask)
        printer->println("EMERGENCY_YIELD(%d, %d, \"%s\");", continueLabelIDStack.back(), forStmt.maxLoopsPerFrame, forStmt.getLocation().toString().c_str());

    //statements
    forStmt.statementBlock->generateCode(*this);

    //increase itersSinceLastYield
    if(isTask)
        printer->println("_taskContext->_bs_for%d_itersSinceLastYield++;", continueLabelIDStack.back());

    //continueLabel / postStatement
    ASSERT(continueLabelIDStack.size() > 0);
    printer->println("continueLabel%d: ;", continueLabelIDStack.back());
    if(forStmt.postStatement)
    {
        isStatement = true;
        forStmt.postStatement->generateCode(*this);
    }

    printer->dindent();
    printer->println("}");

    //pop loop id
    continueLabelIDStack.pop_back();
}

void CppCodeGenerator::generateCode(ast::AssignmentStatement &assigStmt)
{
    //this is a little hack to make int2enum assignments possible
    if(assigStmt.target->getType()->ident->name == "int")
        printer->print("(int&)");
    assigStmt.target->generateCode(*this);

    printer->print(" = ");
    assigStmt.expr->generateCode(*this);
    if(isStatement)
    {
        printer->print(";");
        printer->newline();
    }
}

void CppCodeGenerator::generateCode(ast::YieldStatement &yieldStmt)
{
    printer->print("YIELD(");
    if(yieldStmt.stateExpr)
      yieldStmt.stateExpr->generateCode(*this);
    else
      printer->print("1");
    printer->print(")");
    if(isStatement)
    {
        printer->print(";");
        printer->newline();
    }
}

void CppCodeGenerator::generateCode(ast::ReturnStatement &returnStmt)
{
    printer->println("POP_CALL_TREE(_context->callTree);");
    printer->print("return ");
    returnStmt.expr->generateCode(*this);
    printer->print(";");
    printer->newline();
}

void CppCodeGenerator::generateCode(ast::BreakStatement &breakStmt)
{
    printer->println("break;");
}

void CppCodeGenerator::generateCode(ast::ContinueStatement &breakStmt)
{
    printer->println("goto continueLabel%d;", continueLabelIDStack.back());
}

void CppCodeGenerator::generateCode(ast::LiteralExpression<bool> &litBoolExpr)
{
    printer->print("%s", litBoolExpr.value ? "true" : "false");
}

void CppCodeGenerator::generateCode(ast::LiteralExpression<int> &litIntExpr)
{
    printer->print("%d", litIntExpr.value);
}

void CppCodeGenerator::generateCode(ast::LiteralExpression<float> &litFloatExpr)
{
    printer->print("%f", litFloatExpr.value);
}

void CppCodeGenerator::generateCode(ast::LiteralExpression<std::string> &litStringExpr)
{
    printer->print("\"%s\"", litStringExpr.value.c_str());
}

void CppCodeGenerator::generateCode(ast::CallExpression &callExpr)
{
    ASSERT(callExpr.ident->path.size() <= 1);
    std::string identName;
    if(callExpr.ident->declaration->ident->cppName.size() > 0)
        identName = callExpr.ident->declaration->ident->cppName;
    else
    {
        if(callExpr.ident->path.size())
        {
            identName += "bs_";
            identName += callExpr.ident->path.at(0) + "::";
        }
        identName += callExpr.ident->name;
    }
    printer->print("%s(", identName.c_str());

    bool isFullBScriptFunction = ((ast::FunctionDeclaration*)callExpr.ident->declaration)->isFullBScriptFunction();

    for(size_t i = 0; i < callExpr.paramExprs->size(); i++)
    {
        callExpr.paramExprs->at(i)->generateCode(*this);
        if(i != callExpr.paramExprs->size()-1 || isFullBScriptFunction)
            printer->print(", ");
    }
    if(isFullBScriptFunction)
        printer->print("_context");
    printer->print(")");
}

void CppCodeGenerator::generateCode(ast::IdentifierAccessExpression &identifierAccess)
{
    if(identifierAccess.ident->declaration->getDeclarationType() == ast::Declaration::Value)
    {
        ast::ValueDeclarationBase *decl = ((ast::ValueDeclarationBase*)identifierAccess.ident->declaration);
        ASSERT(decl->ident->cppName.size());
        printer->print(decl->ident->cppName.c_str());
    }
    else
    {
        ASSERT(identifierAccess.ident->declaration->getDeclarationType() == ast::Declaration::Variable);
        ast::VariableDeclaration *varDecl = ((ast::VariableDeclaration*)identifierAccess.ident->declaration);
        if(varDecl->getVariableType() == ast::VariableDeclaration::Global)
            printer->print("(*(%s*)_context->%s)",
                           getVariableCppTypeIdent(*varDecl).c_str(),
                           varDecl->ident->name.c_str());
        else
        {
            if(varDecl->ident->cppName.size() > 0)
                printer->print("%s", varDecl->ident->cppName.c_str());
            else
                printer->print("%s", identifierAccess.ident->fullName().c_str());
        }
    }
}

void CppCodeGenerator::generateCode(ast::MemberAccessExpression &memAccess)
{
    memAccess.leftExpr->generateCode(*this);
    printer->print(".");
    memAccess.rightExpr->generateCode(*this);
}

void CppCodeGenerator::generateCode(ast::UnaryExpression &unaryExpr)
{
    printer->print("%s", unaryExpr.type == ast::UnaryExpression::not_ ? "!" : "-");
    if(!unaryExpr.customOpCall)
        unaryExpr.expr->generateCode(*this);
    else
        unaryExpr.customOpCall->generateCode(*this);
}

void CppCodeGenerator::generateCode(ast::BinaryExpression &binaryExpr)
{
    if(!binaryExpr.customOpCall)
    {
        printer->print("(");
        binaryExpr.leftExpr->generateCode(*this);
        switch(binaryExpr.type)
        {
            case ast::BinaryExpression::plus:
                printer->print(" + ");
                break;
            case ast::BinaryExpression::minus:
                printer->print(" - ");
                break;
            case ast::BinaryExpression::mul:
                printer->print(" * ");
                break;
            case ast::BinaryExpression::div:
                printer->print(" / ");
                break;
            case ast::BinaryExpression::mod:
                printer->print(" %% ");
                break;
            case ast::BinaryExpression::lt:
                printer->print(" < ");
                break;
            case ast::BinaryExpression::leq:
                printer->print(" <= ");
                break;
            case ast::BinaryExpression::gt:
                printer->print(" > ");
                break;
            case ast::BinaryExpression::geq:
                printer->print(" >= ");
                break;
            case ast::BinaryExpression::eq:
                printer->print(" == ");
                break;
            case ast::BinaryExpression::neq:
                printer->print(" != ");
                break;
            case ast::BinaryExpression::and_:
                printer->print(" && ");
                break;
            case ast::BinaryExpression::or_:
                printer->print(" || ");
                break;
            default: ASSERT(false);
        }
        binaryExpr.rightExpr->generateCode(*this);
        printer->print(")");
    }
    else
    {
        binaryExpr.customOpCall->generateCode(*this);
    }
}

void CppCodeGenerator::generateCode(ast::TaskStateExpression &taskStateExpr)
{
    ASSERT(taskStateExpr.taskIdent->path.size() <= 1);
    
    if(taskStateExpr.taskIdent->path.size() == 1)
        printer->print("TASK_STATE2(%s, %s)", taskStateExpr.taskIdent->path.at(0).c_str(), taskStateExpr.taskIdent->name.c_str());
    else
        printer->print("TASK_STATE(%s)", taskStateExpr.taskIdent->fullName().c_str());
}

void CppCodeGenerator::generateCode(ast::VariableDeclaration &varDecl)
{
    TreePrinter *printerBackUp = printer;
    printer = isTask ? headerPrinter : this->printer;
    if(isTask)
    {
        varDecl.ident->cppName = "(_taskContext->";
        char buf[32];
        snprintf(buf, 32, "%d", identifierID);
        varDecl.ident->cppName += varDecl.ident->name + buf + ")";
    }
    else
    {
        char buf[32];
        snprintf(buf, 32, "%d", identifierID);
        varDecl.ident->cppName += varDecl.ident->name + buf;
    }

    std::string typeString = getVariableCppTypeIdent(varDecl);
    /*
    if(varDecl.initializeStatement)
    {
        printer->print("%s ", typeString.c_str());
        varDecl.initializeStatement->generateCode(*this);
    }
    else
    */
    {
        printer->print("%s %s%d", typeString.c_str(), varDecl.ident->name.c_str(), identifierID);
        if(isStatement)
        {
            printer->print(";");
            printer->newline();
        }
    }
    identifierID++;
    printer = printerBackUp;
    if(isTask)
        printer->println("%s = %s();", varDecl.ident->cppName.c_str(), typeString.c_str());
}

void CppCodeGenerator::generateCode(ast::GlobalVariableDeclaration &varDecl)
{
}

void CppCodeGenerator::generateCode(ast::FunctionDeclaration &funcDecl)
{
    ASSERT(!localFunctions.size() || isLocalFunctionDeclaration);
    isTask = false;
    if(!isLocalFunctionDeclaration)
        identifierID = 0;
    continueLabelID = 0;
    ASSERT(continueLabelIDStack.size() == 0);

    TreePrinter *printerBackUp = printer;
    printer = isLocalFunctionDeclaration ? localPrinter : this->printer;
    std::string signature = funcDecl.returnVarDecl ? getVariableCppTypeIdent(*funcDecl.returnVarDecl) : "void";
    signature += " ";
    if(funcDecl.ident->cppName.size() == 0)
        signature += funcDecl.ident->name + "(";
    else
        signature += funcDecl.ident->cppName + "(";

    isStatement = false;
    for(size_t i = 0; i < funcDecl.paramDecls->size(); i++)
    {
        ast::VariableDeclaration *varDecl = funcDecl.paramDecls->at(i);
        signature += "const ";
        signature += getVariableCppTypeIdent(*varDecl) + " &" + varDecl->ident->name + ", ";
    }
    signature += "ExecutionContext *_context";
    signature += ")";

    printer->println("inline %s", signature.c_str());
    headerPrinter->println("%s;", signature.c_str());

    printer->println("{");
    printer->indent();
    printer->println("PUSH_CALL_TREE(_context->callTree, \"%s\");", funcDecl.ident->name.c_str());

    if(funcDecl.statements)
        funcDecl.statements->generateCode(*this);

    printer->println("POP_CALL_TREE(_context->callTree);");
    printer->dindent();
    printer->println("}");

    printer = printerBackUp;

    std::vector<ast::Statement*> myLocalFunctions = localFunctions;
    localFunctions.clear();
    for(std::vector<ast::Statement*>::iterator i = myLocalFunctions.begin(); i != myLocalFunctions.end(); i++)
    {
        isLocalFunctionDeclaration = true;
        (*i)->generateCode(*this);
    }
}

void CppCodeGenerator::generateCode(ast::TaskDeclaration &taskDecl)
{
    ASSERT(!localFunctions.size());
    std::string signature = "void ";
    signature += taskDecl.ident->name + "(";

    isStatement = false;
    isTask = false;
    identifierID = 0;
    continueLabelID = 0;
    ASSERT(continueLabelIDStack.size() == 0);

    for(size_t i = 0; i < taskDecl.paramDecls->size(); i++)
    {
        ast::VariableDeclaration *varDecl = taskDecl.paramDecls->at(i);
        signature += "const ";
        signature += getVariableCppTypeIdent(*varDecl) + " &" + varDecl->ident->name + ", ";
    }
    signature += "ExecutionContext *_context";
    signature += ")";

    printer->println("inline %s", signature.c_str());
    headerPrinter->println("%s;", signature.c_str());

    headerPrinter->println("struct _bstc_%s", taskDecl.ident->name.c_str());
    headerPrinter->println("{");
    headerPrinter->indent();
    headerPrinter->println("_bstc_%s() : _bs_state(0), _bs_yieldCounter(0) {}", taskDecl.ident->name.c_str());
    headerPrinter->println("unsigned _bs_state;");
    headerPrinter->println("unsigned _bs_yieldCounter;");

    printer->println("{");
    printer->indent();

    printer->println("COROUTINE_BEGIN(%s);", taskDecl.ident->name.c_str());
    if(taskDecl.statements)
    {
        isTask = true;
        taskDecl.statements->generateCode(*this);
    }
    printer->println("COROUTINE_END();");
    printer->dindent();
    printer->println("}");

    headerPrinter->dindent();
    headerPrinter->println("};");

    std::vector<ast::Statement*> myLocalFunctions = localFunctions;
    localFunctions.clear();
    for(std::vector<ast::Statement*>::iterator i = myLocalFunctions.begin(); i != myLocalFunctions.end(); i++)
    {
        isLocalFunctionDeclaration = true;
        (*i)->generateCode(*this);
    }
}

void CppCodeGenerator::generateCode(ast::ClassDeclaration &classDecl)
{
    ASSERT(false);
}

std::string getModuleName(const std::string &filename)
{
    std::string moduleName = filename;
    size_t posOfDot = moduleName.rfind(".");
    if(posOfDot != moduleName.npos)
        moduleName.replace(posOfDot, moduleName.size() - posOfDot, "");

    size_t lastSlash = moduleName.find_last_of('/');
    if(lastSlash != moduleName.npos)
        moduleName = moduleName.substr(lastSlash + 1, moduleName.size() - lastSlash - 1);

    return moduleName;
}

void CppCodeGenerator::generateCode(ast::ModuleDeclaration &modDecl)
{
    std::string cppName = "bs_";
    cppName += modDecl.ident->name;

    printer = new TreePrinter((path + "/" + cppName + ".cpp").c_str());
    headerPrinter = new TreePrinter((path + "/" + cppName + ".h").c_str());

    localPrinter = new TreePrinter((path + "/" + cppName + ".locals").c_str());
    printer->println("#include \"builtins.h\"");
    printer->println("#include \"globalIncludes.h\"");

    headerPrinter->println("#pragma once");
    for(size_t i = 0; i < modDecl.requirements.size(); i++)
        headerPrinter->println("#include \"bs_%s.h\"", getModuleName(modDecl.requirements.at(i)).c_str());

    headerPrinter->println("namespace %s", cppName.c_str());
    headerPrinter->println("{");

    printer->println("#include \"%s.h\"", cppName.c_str());
    printer->println("namespace %s", cppName.c_str());
    printer->println("{");
    printer->println("#include \"%s.locals\"", cppName.c_str());

    localFunctionID = 0;
    for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
    {
        isLocalFunctionDeclaration = false;
        modDecl.functionDeclarations.at(i)->generateCode(*this);
    }
    printer->println("}");

    printer->println("BS_MODULE(%s,", modDecl.ident->name.c_str());
    printer->indent();

    for(size_t i = 0; i < modDecl.requirements.size(); i++)
        printer->println("BS_REQUIRES(%s);", getModuleName(modDecl.requirements.at(i)).c_str());

    for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
    {
        ast::FunctionDeclaration *funcDecl = modDecl.functionDeclarations.at(i);
        std::string registerCall = "registerFullBScriptFunction";
        if(!funcDecl->returnVarDecl)
            registerCall += "Void";

        if(funcDecl->getDeclarationType() == ast::Declaration::Task)
        {
          registerCall = "registerTask<";
          registerCall += cppName + "::_bstc_" + funcDecl->ident->name + ">";
        }
        printer->println("%s(module, \"%s\", &%s::%s);",
                         registerCall.c_str(),
                         funcDecl->ident->name.c_str(),
                         cppName.c_str(),
                         funcDecl->ident->name.c_str());
    }

    printer->println(");");
    printer->dindent();

    delete printer;
    headerPrinter->println("}");
    delete headerPrinter;
    delete localPrinter;
}

void CppCodeGenerator::generateCode(ast::ExternModuleDeclaration &modDecl)
{
    //ignore extern modules....
    //they should be available in the main binary or some module library
}

void CppCodeGenerator::generateCode(ast::MasterModuleDeclaration &modDecl)
{
    {
        TreePrinter globalIncludesPrinter((path + "/globalIncludes.h").c_str());
        for(size_t i = 0; i < modDecl.requirements.size(); i++)
        {
            std::string moduleName = modDecl.requirements.at(i);
            size_t posOfDot = moduleName.rfind(".");
            if(posOfDot != moduleName.npos)
                moduleName.replace(posOfDot, moduleName.size() - posOfDot, "");
            moduleName += ".h";

            size_t lastSlash = moduleName.find_last_of('/');
            if(lastSlash != moduleName.npos)
                moduleName = moduleName.substr(lastSlash + 1, moduleName.size() - lastSlash - 1);
            globalIncludesPrinter.println("#include \"bs_%s\"", moduleName.c_str());
        }
    }

    /*
    printer = new TreePrinter((path + "/mainModule.cpp").c_str());
    headerPrinter = new TreePrinter((path + "/mainModule.h").c_str());
    localPrinter = new TreePrinter((path + "/mainModule.locals.cpp").c_str());
    printer->println("#include \"bs__main_.h\"");
    printer->println("#include \"mainModule.h\"");
    printer->println("#include \"mainModule.locals.cpp\"");
    printer->println("int main(int argc, char **argv)");
    printer->println("{");
    printer->indent();
    modDecl.mainCallStatement->generateCode(*this);
    printer->dindent();
    printer->println("}");

    delete printer;
    delete headerPrinter;
    delete localPrinter;

    modDecl.mainModule->generateCode(*this);
    */

    for(size_t i = 0; i < modDecl.moduleDeclarations.size(); i++)
        modDecl.moduleDeclarations.at(i)->generateCode(*this);
}

void CppCodeGenerator::generateCode(ast::EnumDeclaration &enumDecl)
{
    ASSERT(false);
}

