/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "../tools/treePrinter.h"

#include "codeGeneratorInterface.h"
#include "../ast/ast.h"

class DebugTreeGenerator : public CodeGeneratorInterface
{
    public:
        DebugTreeGenerator(const std::string &filename = std::string("-"))
            : printer(filename)
        {}

        void generateCode(ast::StatementBlock &statementBlock)
        {
            printer.println("(StatementBlock)");
            printer.indent();

            ast::StatementBlock::StatementList::iterator i;
            for(i = statementBlock.functionDeclarations.begin(); i != statementBlock.functionDeclarations.end(); i++)
                (*i)->generateCode(*this);
            for(i = statementBlock.variableDeclarations.begin(); i != statementBlock.variableDeclarations.end(); i++)
                (*i)->generateCode(*this);
            for(i = statementBlock.statements.begin(); i != statementBlock.statements.end(); i++)
                (*i)->generateCode(*this);

            printer.dindent();
        }

        void generateCode(ast::CallStatement &callStmt)
        {
            printer.println("(CallStatement)");
            printer.indent();
            callStmt.funcCall->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::IfStatement &ifStmt)
        {
            printer.println("(IfStatement)");
            printer.indent();
            ifStmt.expression->generateCode(*this);
            ifStmt.ifStatements->generateCode(*this);
            if(ifStmt.elseStatements)
                ifStmt.elseStatements->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::ForStatement &forStmt)
        {
            printer.println("(ForStatement)");
            printer.indent();
            if(forStmt.varDecl)
                forStmt.varDecl->generateCode(*this);
            forStmt.condition->generateCode(*this);
            if(forStmt.postStatement)
                forStmt.postStatement->generateCode(*this);
            printer.indent();
            forStmt.statementBlock->generateCode(*this);
            printer.dindent();
            printer.dindent();
        }

        void generateCode(ast::AssignmentStatement &assigStmt)
        {
            printer.println("(AssignmentStatement)");
            printer.indent();
            assigStmt.target->generateCode(*this);
            assigStmt.expr->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::YieldStatement &yieldStmt)
        {
            printer.println("(YieldStatement)");
        }

        void generateCode(ast::ReturnStatement &returnStmt)
        {
            printer.println("(ReturnStatement)");
            printer.indent();
                returnStmt.target->generateCode(*this);
                returnStmt.expr->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::BreakStatement &breakStmt)
        {
            printer.println("(BreakStatement)");
        }

        void generateCode(ast::ContinueStatement &breakStmt)
        {
            printer.println("(ContinueStatement)");
        }

        void generateCode(ast::LiteralExpression<bool> &litBoolExpr)
        {
            printer.println("(LiteralExpression<bool>, %s, %s)",
                            litBoolExpr.value ? "true" : "false",
                            litBoolExpr.typeIdent->toString().c_str());
        }

        void generateCode(ast::LiteralExpression<int> &litIntExpr)
        {
            printer.println("(LiteralExpression<int>, %d, %s)",
                            litIntExpr.value,
                            litIntExpr.typeIdent->toString().c_str());
        }

        void generateCode(ast::LiteralExpression<float> &litFloatExpr)
        {
            printer.println("(LiteralExpression<float>, %f, %s)",
                            litFloatExpr.value,
                            litFloatExpr.typeIdent->toString().c_str());
        }

        void generateCode(ast::LiteralExpression<std::string> &litStringExpr)
        {
            printer.println("(LiteralExpression<string>, %s, %s)",
                            litStringExpr.value.c_str(),
                            litStringExpr.typeIdent->toString().c_str());
        }

        void generateCode(ast::CallExpression &callExpr)
        {
            printer.println("(CallExpression, %s)", callExpr.ident->toString().c_str());
            printer.indent();
            printer.println("(ParameterList)");
            printer.indent();
            for(size_t i = 0; i < callExpr.paramExprs->size(); i++)
                callExpr.paramExprs->at(i)->generateCode(*this);
            printer.dindent();
            printer.dindent();
        }

        void generateCode(ast::IdentifierAccessExpression &identifierAccess)
        {
            printer.println("(IdentifierAccessExpression, %s)", identifierAccess.ident->toString().c_str());
        }

        void generateCode(ast::MemberAccessExpression &memAccess)
        {
            printer.println("(MemberAccessExpression)");
            printer.indent();
            memAccess.leftExpr->generateCode(*this);
            memAccess.rightExpr->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::UnaryExpression &unaryExpr)
        {
            printer.println("(UnaryExpression, %s)", unaryExpr.type == ast::UnaryExpression::not_ ? "not" : "minus");
            printer.indent();
            if(!unaryExpr.customOpCall)
                unaryExpr.expr->generateCode(*this);
            else
                unaryExpr.customOpCall->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::BinaryExpression &binaryExpr)
        {
            printer.println("(BinaryExpression, %s)", ast::BinaryExpression::operationNames[binaryExpr.type]);
            printer.indent();
            if(!binaryExpr.customOpCall)
            {
                binaryExpr.leftExpr->generateCode(*this);
                binaryExpr.rightExpr->generateCode(*this);
            }
            else
                binaryExpr.customOpCall->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::TaskStateExpression &taskStateExpr)
        {
            printer.println("(TaskStateExpression, %s)", taskStateExpr.taskIdent->fullName().c_str());
        }

        void generateCode(ast::VariableDeclaration &varDecl)
        {
            printer.println("(VariableDeclaration, %s, %s)",
                            varDecl.ident->name.c_str(),
                            varDecl.typeIdent->toString().c_str());
            printer.indent();
            if(varDecl.initializeStatement)
                varDecl.initializeStatement->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::GlobalVariableDeclaration &varDecl)
        {
            printer.println("(GlobalVariableDeclaration, %s, %s)",
                            varDecl.ident->name.c_str(),
                            varDecl.typeIdent->toString().c_str());
        }

        void generateCode(ast::FunctionDeclaration &funcDecl)
        {
            printer.println("(FunctionDeclaration, %s -> %s)", funcDecl.ident->name.c_str(), funcDecl.returnVarDecl ? funcDecl.returnVarDecl->typeIdent->fullName().c_str() : "void");
            printer.indent();

            printer.println("(ParameterDeclaration)");
            printer.indent();
            for(size_t i = 0; i < funcDecl.paramDecls->size(); i++)
                funcDecl.paramDecls->at(i)->generateCode(*this);
            printer.dindent();

            if(funcDecl.statements)
                funcDecl.statements->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::TaskDeclaration &taskDecl)
        {
            printer.println("(TaskDeclaration, %s)", taskDecl.ident->name.c_str());
            printer.indent();
            if(taskDecl.statements)
                taskDecl.statements->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::ClassDeclaration &classDecl)
        {
            printer.println("(ClassDeclaration, %s)", classDecl.ident->name.c_str());
            printer.indent();
            for(size_t i = 0; i < classDecl.memberFunctions.size(); i++)
                classDecl.memberFunctions.at(i)->generateCode(*this);
            for(size_t i = 0; i < classDecl.memberVariables.size(); i++)
                classDecl.memberVariables.at(i)->generateCode(*this);
            printer.dindent();
        }

        void generateCode(ast::ModuleDeclaration &modDecl)
        {
            printer.println("(ModuleDeclaration, %s)", modDecl.ident->name.c_str());

            printer.indent();
            for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
                modDecl.functionDeclarations.at(i)->generateCode(*this);
            printer.dindent();

        }

        void generateCode(ast::ExternModuleDeclaration &modDecl)
        {
            printer.println("(ExternModuleDeclaration, %s)", modDecl.ident->name.c_str());

            printer.indent();
            for(size_t i = 0; i < modDecl.enumDeclarations.size(); i++)
                modDecl.enumDeclarations.at(i)->generateCode(*this);
            for(size_t i = 0; i < modDecl.classDeclarations.size(); i++)
                modDecl.classDeclarations.at(i)->generateCode(*this);
            for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
                modDecl.functionDeclarations.at(i)->generateCode(*this);
            printer.dindent();

        }

        void generateCode(ast::MasterModuleDeclaration &modDecl)
        {
            printer.println("(MasterModuleDeclaration, %s)", modDecl.ident->name.c_str());

            printer.indent();
            for(size_t i = 0; i < modDecl.classDeclarations.size(); i++)
                modDecl.classDeclarations.at(i)->generateCode(*this);
            for(size_t i = 0; i < modDecl.enumDeclarations.size(); i++)
                modDecl.enumDeclarations.at(i)->generateCode(*this);
            for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
                modDecl.functionDeclarations.at(i)->generateCode(*this);
            for(size_t i = 0; i < modDecl.moduleDeclarations.size(); i++)
                modDecl.moduleDeclarations.at(i)->generateCode(*this);
            if(modDecl.mainModule)
                modDecl.mainModule->generateCode(*this);
            if(modDecl.mainCallStatement)
                modDecl.mainCallStatement->generateCode(*this);
            if(modDecl.inputDeclaration)
                modDecl.inputDeclaration->generateCode(*this);
            if(modDecl.outputDeclaration)
                modDecl.outputDeclaration->generateCode(*this);

            printer.dindent();
        }

        void generateCode(ast::EnumDeclaration &enumDecl)
        {
            printer.println("(EnumDeclaration, %s)", enumDecl.ident->name.c_str());
            printer.indent();

            ast::ValueDeclaration<int> *val;
            for(size_t i = 0; i < enumDecl.values.size(); i++)
            {
                val = enumDecl.values.at(i);
                printer.println("(%s, %d)", val->ident->name.c_str(), val->value);
            }

            printer.dindent();
        }

    private:
        TreePrinter printer;
};

