/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <vector>

#include "../tools/treePrinter.h"
#include "../tools/tr1Include.h"

#include "codeGeneratorInterface.h"
#include "../ast/ast.h"

class DotGraphGenerator : public CodeGeneratorInterface
{
    public:
        DotGraphGenerator(const std::string &filename = std::string("-"))
            : printer(filename)
        {}

        void generateCode(ast::StatementBlock &statementBlock)
        {
            ast::StatementBlock::StatementList::iterator i;
            for(i = statementBlock.functionDeclarations.begin(); i != statementBlock.functionDeclarations.end(); i++)
                (*i)->generateCode(*this);
            for(i = statementBlock.statements.begin(); i != statementBlock.statements.end(); i++)
                (*i)->generateCode(*this);
        }

        void generateCode(ast::CallStatement &callStmt)
        {
            callStmt.funcCall->generateCode(*this);
        }

        void generateCode(ast::IfStatement &ifStmt)
        {
            ifStmt.expression->generateCode(*this);
            ifStmt.ifStatements->generateCode(*this);
            if(ifStmt.elseStatements)
                ifStmt.elseStatements->generateCode(*this);
        }

        void generateCode(ast::ForStatement &forStmt)
        {
            forStmt.condition->generateCode(*this);
            forStmt.statementBlock->generateCode(*this);
            if(forStmt.postStatement)
                forStmt.postStatement->generateCode(*this);
        }

        void generateCode(ast::AssignmentStatement &assigStmt)
        {
            assigStmt.target->generateCode(*this);
            assigStmt.expr->generateCode(*this);
        }

        void generateCode(ast::YieldStatement &yieldStmt)
        {}

        void generateCode(ast::ReturnStatement &returnStmt)
        {
            returnStmt.expr->generateCode(*this);
        }

        void generateCode(ast::BreakStatement &breakStmt)
        {}

        void generateCode(ast::ContinueStatement &breakStmt)
        {}

        void generateCode(ast::LiteralExpression<bool> &litBoolExpr)
        {}

        void generateCode(ast::LiteralExpression<int> &litIntExpr)
        {}

        void generateCode(ast::LiteralExpression<float> &litFloatExpr)
        {}

        void generateCode(ast::LiteralExpression<std::string> &litStringExpr)
        {}

        void generateCode(ast::CallExpression &callExpr)
        {
            if(functionNameStack.size())
            {
                std::string calleeName;
                std::string edgeStr = "\"";
                edgeStr += functionNameStack.back();
                edgeStr += "\" -> \"";
                if(callExpr.ident->path.size())
                    edgeStr += callExpr.ident->fullName();
                else
                    edgeStr += actualModuleName + "::" + callExpr.ident->fullName();
                edgeStr += "\"";
                edges.insert(edgeStr);
            }
        }

        void generateCode(ast::IdentifierAccessExpression &identifierAccess)
        {}

        void generateCode(ast::MemberAccessExpression &memAccess)
        {}

        void generateCode(ast::UnaryExpression &unaryExpr)
        {
            unaryExpr.expr->generateCode(*this);
        }

        void generateCode(ast::BinaryExpression &binaryExpr)
        {
            binaryExpr.leftExpr->generateCode(*this);
            binaryExpr.rightExpr->generateCode(*this);
        }

        void generateCode(ast::TaskStateExpression &taskStateExpr)
        {}

        void generateCode(ast::VariableDeclaration &varDecl)
        {}

        void generateCode(ast::GlobalVariableDeclaration &varDecl)
        {}

        void generateCode(ast::FunctionDeclaration &funcDecl)
        {
            std::string fullName = actualModuleName + "::" + funcDecl.ident->name;
            functionNameStack.push_back(fullName);
            printer.println("\"%s\" [shape=egg];", fullName.c_str());
            printer.println("\"%s\" [style=filled];", fullName.c_str());
            printer.println("\"%s\" [color=lightblue];", fullName.c_str());
            printer.println("\"%s\" [label=%s];", fullName.c_str(), funcDecl.ident->name.c_str());
            if(funcDecl.statements)
                funcDecl.statements->generateCode(*this);
            functionNameStack.pop_back();
        }

        void generateCode(ast::TaskDeclaration &taskDecl)
        {
            std::string fullName = actualModuleName + "::" + taskDecl.ident->name;
            functionNameStack.push_back(fullName);
            printer.println("\"%s\" [shape=box];", fullName.c_str());
            printer.println("\"%s\" [style=filled];", fullName.c_str());
            printer.println("\"%s\" [color=green];", fullName.c_str());
            printer.println("\"%s\" [label=%s];", fullName.c_str(), taskDecl.ident->name.c_str());
            taskDecl.statements->generateCode(*this);
            functionNameStack.pop_back();
        }

        void generateCode(ast::ClassDeclaration &classDecl)
        {}

        void generateCode(ast::ModuleDeclaration &modDecl)
        {
            actualModuleName = modDecl.ident->name;
            printer.println("subgraph cluster_%s {\n label = \"%s\"; style=filled; color=grey;", actualModuleName.c_str(), actualModuleName.c_str());
            for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
                modDecl.functionDeclarations.at(i)->generateCode(*this);
            printer.println("}");
        }

        void generateCode(ast::ExternModuleDeclaration &modDecl)
        {
            actualModuleName = modDecl.ident->name;
            for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
                modDecl.functionDeclarations.at(i)->generateCode(*this);
        }

        void generateCode(ast::MasterModuleDeclaration &modDecl)
        {
            printer.println("digraph callGraph {");
            for(size_t i = 0; i < modDecl.moduleDeclarations.size(); i++)
                modDecl.moduleDeclarations.at(i)->generateCode(*this);

            printer.println("subgraph cluster_master_{\nlabel = \"_master_\"");
            //if(modDecl.mainModule)
            //    modDecl.mainModule->generateCode(*this);

            actualModuleName = modDecl.ident->name;

            //for(size_t i = 0; i < modDecl.functionDeclarations.size(); i++)
            //    modDecl.functionDeclarations.at(i)->generateCode(*this);
            //if(modDecl.mainCallStatement)
            //    modDecl.mainCallStatement->generateCode(*this);
            printer.println("}");

            std::tr1::unordered_set<std::string>::iterator i;
            for(i = edges.begin(); i != edges.end(); i++)
                printer.println(i->c_str());

            printer.println("}");
        }

        void generateCode(ast::EnumDeclaration &enumDecl)
        {}

    private:
        std::vector<std::string> functionNameStack;
        std::tr1::unordered_set<std::string> edges;
        std::string actualModuleName;
        TreePrinter printer;
};

