/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/


%{
    #include "../parserInterface.h"

    using namespace ast;

    void yy::parser::error(const yy::parser::location_type& l,
                           const std::string& m)
    {
        interface.error(l, m);
    }

%}

%error-verbose
%defines
%skeleton "lalr1.cc"
%locations

%parse-param { ParserInterface& interface }
%parse-param { ast::ModuleDeclaration *module }
%parse-param { const std::string &filename }
%lex-param   { ParserInterface& interface }

%code requires {
    #include "../../ast/ast.h"
}

%union {
    int                                              ival;
    float                                            fval;
    bool                                             bval;
    std::string                                     *sval;

    ast::FunctionDeclaration                        *funcDecl;
    ast::TaskDeclaration                            *taskDecl;
    ast::VariableDeclaration                        *varDecl;
    ast::FunctionDeclaration::ParameterDeclaration  *paramDecls;

    ast::StatementBlock                             *block;
    ast::StatementBlock::StatementList              *statementList;
    ast::Statement                                  *statement;
    ast::IfStatement                                *ifStatement;
    ast::ForStatement                               *forStatement;
    ast::AssignmentStatement                        *assigStatement;

    ast::Expression                                 *expression;
    ast::CallExpression                             *functionCall;
    ast::IdentifierAccessExpression                 *identAccess;
    ast::MemberAccessExpression                     *memberAccess;

    ast::Identifier                                 *identifier;
    ast::ResolvableIdentifier                       *resolvableIdentifier;
    ast::CallExpression::ParameterExpressions       *parameterExprs;
}

%token                      END      0                    "end of file"
%token <ival>               NUMBER                        "number"
%token <fval>               FLOAT                         "float"
%token <sval>               IDENTIFIER                    "identifier"
%token <sval>               STRING                        "string"
%token <bval>               BOOL                          "boolean"
%token                      NEW_LINE                      "new line"
%token                      INDENT                        "indentation"
%token                      DINDENT                       "dindentation"
%token                      REQUIRES                      "'requires'"
%token                      FUNC                          "'func'"
%token                      TASK                          "'task'"
%token                      YIELD                         "'yield'"
%token                      TASK_STATE                    "'taskState'"
%token                      IF                            "'if'"
%token                      ELSE                          "'else'"
%token                      FOR                           "'for'"
%token                      RETURN                        "'return'"
%token                      BREAK                         "'break'"
%token                      CONTINUE                      "'continue'"
%token                      PASS                          "'pass'"
%token                      LPARENT                       "'('"
%token                      RPARENT                       "')'"
%token                      COMMA                         "','"
%token                      DOT                           "'.'"
%token                      ASSIGN                        "'='"
%token                      NOT                           "'!'"
%token                      BECOMES_PLUS                  "'+='"
%token                      BECOMES_MINUS                 "'-='"
%token                      BECOMES_MUL                   "'*='"
%token                      BECOMES_DIV                   "'/='"
%token                      BECOMES_MOD                   "'%='"
%token                      MINUS                         "'-'"
%token                      SEMICOLON                     "';'"
%token                      COLON                         "':'"

%token                      INDENT_ERROR                  "indentation error"
%token                      UNTERMINATED_STRING_ERROR     "unterminated string"
%token                      NOT_EMPTY_EMPTY_LINE_ERROR    "not empty line containing only whitespaces"
%token                      UNKNOWN_TOKEN_ERROR           "unknown token"

%type <funcDecl>                func_decl
%type <statementList>           func_decl_list
%type <taskDecl>                task_decl
%type <varDecl>                 var_decl
%type <varDecl>                 full_var_decl
%type <paramDecls>              opt_param_decls
%type <paramDecls>              param_decls
%type <block>                   statements
%type <statement>               statement
%type <statement>               simple_statement
%type <ifStatement>             if_statement
%type <forStatement>            for_statement
%type <expression>              expression
%type <identifier>              identifier
%type <resolvableIdentifier>    resolvable_identifier
%type <functionCall>            function_call
%type <identAccess>             ident_access
%type <parameterExprs>          parameter_exprs
%type <parameterExprs>          opt_parameter_exprs
%type <expression>              var_or_call
%type <ival>                    becomes_operator
%type <ival>                    opt_emergency_qualifier

%left OR
%left AND
%nonassoc EQ NEQ
%nonassoc LT GT LEQ GEQ
%left PLUS MINUS
%left MUL DIV MOD
%nonassoc NOT NEG
%left DOT

%destructor { delete $$; } IDENTIFIER "string" func_decl func_decl_list task_decl full_var_decl var_decl opt_param_decls param_decls statements statement simple_statement if_statement for_statement expression identifier resolvable_identifier function_call ident_access var_or_call opt_parameter_exprs

%%

module:                   requirements 
                        | NEW_LINE
                        | module func_decl { module->functionDeclarations.push_back($2); }
                        | module task_decl { module->functionDeclarations.push_back($2); }
                        ;

requirements:             /**/
                        | NEW_LINE
                        | requirements "'requires'" STRING NEW_LINE { module->requirements.push_back(*$3);
                                                                      delete $3; }
                        ;

statements:               func_decl_list  { $$ = new StatementBlock();
                                            $$->functionDeclarations = *$1;
                                            $1->clear();
                                            delete $1; } 
                        | statements statement { $$ = $1;
                                                 $$->statements.push_back($2); }
                        | statements PASS NEW_LINE { $$ = $1; }
                        ;

func_decl_list:           /**/ { $$ = new StatementBlock::StatementList; }
                        | func_decl_list func_decl { $$ = $1;
                                                     $$->push_back($2); }
                        ;

statement:                if_statement { $$ = $1; }
                        | for_statement { $$ = $1; }
                        | simple_statement NEW_LINE { $$ = $1; }
                        | full_var_decl NEW_LINE { $$ = $1; }
                        | RETURN expression NEW_LINE { $$ = new ReturnStatement(Location(yylloc, filename), $2); }
                        | BREAK NEW_LINE { $$ = new BreakStatement(Location(yylloc, filename)); }
                        | CONTINUE NEW_LINE { $$ = new ContinueStatement(Location(yylloc, filename)); }
                        ;

simple_statement:         expression { $$ = new CallStatement($1); }
                        | expression "'='" expression { $$ = new AssignmentStatement($1, $3); }
                        | expression becomes_operator expression
                          { $$ = new AssignmentStatement($1, new BinaryExpression((BinaryExpression::Type)$2, $1, $3), true); }
                        | YIELD {$$ = new YieldStatement(Location(yylloc, filename)); }
                        | YIELD expression { $$ = new YieldStatement(Location(yylloc, filename), $2); }
                        ;

becomes_operator:         "'+='" { $$ = BinaryExpression::plus; }
                        | "'-='" { $$ = BinaryExpression::minus; }
                        | "'/='" { $$ = BinaryExpression::div; }
                        | "'*='" { $$ = BinaryExpression::mul; }
                        | "'%='" { $$ = BinaryExpression::mod; }
                        ;

if_statement:             IF "'('" expression "')'" NEW_LINE
                              INDENT statements DINDENT { $$ = new IfStatement($3, $7); }
                        | IF "'('" expression "')'" NEW_LINE
                              INDENT statements DINDENT
                          ELSE NEW_LINE
                              INDENT statements DINDENT { $$ = new IfStatement($3, $7, $12); }
                        | IF "'('" expression "')'" NEW_LINE
                              INDENT statements DINDENT
                          ELSE if_statement
                          {
                              StatementBlock *elseifStmt = new StatementBlock();
                              elseifStmt->statements.push_back($10);
                              $$ = new IfStatement($3, $7, elseifStmt);
                          }
                        ;

for_statement:             FOR "'('" full_var_decl "';'" expression "';'" simple_statement "')'" opt_emergency_qualifier NEW_LINE
                              INDENT statements DINDENT { $$ = new ForStatement($3, $5, $7, $12, $9); }
                        |  FOR "'('" "';'" expression "';'" simple_statement "')'" opt_emergency_qualifier NEW_LINE
                              INDENT statements DINDENT { $$ = new ForStatement(0, $4, $6, $11, $8); }
                        |  FOR "'('" full_var_decl "';'" expression "';'" "')'" opt_emergency_qualifier NEW_LINE
                              INDENT statements DINDENT { $$ = new ForStatement($3, $5, 0, $11, $8); }
                        |  FOR "'('" "';'" expression "';'" "')'" opt_emergency_qualifier NEW_LINE
                              INDENT statements DINDENT { $$ = new ForStatement(0, $4, 0, $10, $7); }
                        ;

opt_emergency_qualifier:   /**/ { $$ = 1; }
                        | "':'" NUMBER { $$ = $2; }
                        ;

func_decl:                FUNC identifier "'('" opt_param_decls "')'" NEW_LINE
                              INDENT statements DINDENT 
                          { $$ = new FunctionDeclaration($2, NULL, $4, $8, $2->location); }
                        | FUNC identifier "'('" opt_param_decls "')'" "':'" resolvable_identifier NEW_LINE 
                              INDENT statements DINDENT 
                          {
                              VariableDeclaration *retVar = new VariableDeclaration(new Identifier("_return_", $7->location), $7, 0);
                              $$ = new FunctionDeclaration($2, retVar, $4, $10, $2->location);
                          }
                        ;

task_decl:                TASK identifier "'('" opt_param_decls "')'" NEW_LINE
                              INDENT statements DINDENT
                          { $$ = new TaskDeclaration($2, $4, $8, $2->location); }
                        ;

full_var_decl:            var_decl { $$ = $1; }
                        | var_decl "'='" expression
                          {
                              $$ = $1;
                              $$->initializeStatement = new AssignmentStatement(
                                                                 new IdentifierAccessExpression(
                                                                     new ResolvableIdentifier($$->ident->name, $$->ident->location)),
                                                                     $3);
                          }
                        ;

var_decl:                resolvable_identifier identifier
                          { $$ = new VariableDeclaration($2, $1, 0); }
                        ;

opt_param_decls:          /**/ { $$ = new FunctionDeclaration::ParameterDeclaration(); }
                        | param_decls { $$ = $1; }
                        ;

param_decls:              var_decl { $$ = new FunctionDeclaration::ParameterDeclaration();
                                     $1->qualifiers = VariableDeclaration::readable;
                                     $$->push_back($1); }
                        | param_decls "','" var_decl { $$ = $1;
                                                       $3->qualifiers = VariableDeclaration::readable;
                                                       $$->push_back($3); }
                        ;

expression:               BOOL { $$ = new LiteralExpression<bool>(new ResolvableIdentifier("bool", Location(yylloc, filename)), $1); }
                        | NUMBER { $$ = new LiteralExpression<int>(new ResolvableIdentifier("int", Location(yylloc, filename)), $1); }
                        | FLOAT { $$ = new LiteralExpression<float>(new ResolvableIdentifier("float", Location(yylloc, filename)), $1); }
                        | STRING { $$ = new LiteralExpression<std::string>(new ResolvableIdentifier("string", Location(yylloc, filename)), *$1); delete $1; }
                        | var_or_call { $$ = $1; }
                        | expression "'.'" var_or_call { $$ = new MemberAccessExpression($1, $3); }
                        | "'!'" expression { $$ = new UnaryExpression(UnaryExpression::not_, $2); }
                        | "'-'" expression %prec NEG { $$ = new UnaryExpression(UnaryExpression::minus, $2); }
                        | "'('" expression "')'" { $$ = $2; }
                        | expression PLUS expression { $$ = new BinaryExpression(BinaryExpression::plus, $1, $3); }
                        | expression MINUS expression { $$ = new BinaryExpression(BinaryExpression::minus, $1, $3); }
                        | expression MUL expression { $$ = new BinaryExpression(BinaryExpression::mul, $1, $3); }
                        | expression DIV expression { $$ = new BinaryExpression(BinaryExpression::div, $1, $3); }
                        | expression MOD expression { $$ = new BinaryExpression(BinaryExpression::mod, $1, $3); }
                        | expression LT expression { $$ = new BinaryExpression(BinaryExpression::lt, $1, $3); }
                        | expression LEQ expression { $$ = new BinaryExpression(BinaryExpression::leq, $1, $3); }
                        | expression GT expression { $$ = new BinaryExpression(BinaryExpression::gt, $1, $3); }
                        | expression GEQ expression { $$ = new BinaryExpression(BinaryExpression::geq, $1, $3); }
                        | expression EQ expression { $$ = new BinaryExpression(BinaryExpression::eq, $1, $3); }
                        | expression NEQ expression { $$ = new BinaryExpression(BinaryExpression::neq, $1, $3); }
                        | expression AND expression { $$ = new BinaryExpression(BinaryExpression::and_, $1, $3); }
                        | expression OR expression { $$ = new BinaryExpression(BinaryExpression::or_, $1, $3); }
                        | TASK_STATE "'('" resolvable_identifier "')'" { $$ = new TaskStateExpression($3); }
                        ;

var_or_call:              ident_access { $$ = $1; }
                        | function_call { $$ = $1; }

ident_access:             resolvable_identifier { $$ = new IdentifierAccessExpression($1); }
                        ;

function_call:            resolvable_identifier "'('" opt_parameter_exprs "')'" { $$ = new CallExpression($1, $3); }
                        ;

opt_parameter_exprs:      /**/ { $$ = new CallExpression::ParameterExpressions(); }
                        | parameter_exprs { $$ = $1; }
                        ;

parameter_exprs:          expression { $$ = new CallExpression::ParameterExpressions();
                                       $$->push_back($1); }
                        | parameter_exprs "','" expression { $$ = $1;
                                                             $$->push_back($3); }
                        ;

resolvable_identifier:    IDENTIFIER { $$ = new ResolvableIdentifier(*$1, Location(yylloc, filename));
                                       delete $1; }
                        | IDENTIFIER "':'" "':'" resolvable_identifier { $$ = $4;
                                                                         $4->pushPath(*$1);
                                                                         delete $1; }
                        ;

identifier:               IDENTIFIER { $$ = new Identifier(*$1, Location(yylloc, filename));
                                       delete $1; }
                        ;

%%

