/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <stdio.h>
#include <string>

#include "ast/declarations/classDeclaration.h"
#include "tools/asserts.h"
#include "tools/functors.h"

class StackException
{
    public:
      std::string msg;

      StackException(const std::string &msg)
        : msg(msg)
      {}
};

template<typename V, int N>
class StackImpl
{
    private:
        V array[N];
        unsigned topOfStack;

    public:

        StackImpl()
            : topOfStack(0)
        {}

        inline bool push(const V &el)
        {
            if(topOfStack >= N - 1)
                throw StackException("push");

            array[topOfStack++] = el;
            return true;
        }

        inline V& pop()
        {
            if(topOfStack == 0)
            {
                ASSERT(false);
                throw StackException("pop");
            }

            return array[--topOfStack];
        }

        inline V& at(unsigned i)
        {
            if(i > topOfStack - 1 || topOfStack == 0)
            {
                ASSERT(false);
                throw StackException("at");
            }

            return array[i];
        }

        inline V& back()
        {
            if(topOfStack == 0)
            {
                ASSERT(false);
                throw StackException("back");
            }

            return array[topOfStack-1];
        }

        inline unsigned size()
        {
            return topOfStack;
        }

        inline void clear()
        {
            topOfStack = 0;
        }
};

namespace ast
{
class TaskDeclarationBase;
}
struct ExecutionContext;

class Stack2 : public FunctorStackInterface
{
#define MEMORY_SIZE 1024
#define POINTER_STACK_SIZE 256
#define FRAME_STACK_SIZE 128
#define CLASS_HANDLER_STACK_SIZE 256
    public:
        unsigned yieldCounter;
        ast::TaskDeclarationBase *owner;
        ExecutionContext *executionContext;
        enum Interuptor
        {
            None,
            Return,
            Break,
            Continue,
            Yield,
        };
        Interuptor interupt;

        Stack2()
            : yieldCounter(0),
              owner(0),
              interupt(None),
              memory(0),
              topOfStack(0)
        {
            memory = new char[MEMORY_SIZE];
            topOfStack = memory;
            frameStack.push(0);
        }

        ~Stack2()
        {
            frameStack.pop();
            ASSERT(classHandlerStack.size() == 0);
            ASSERT(frameStack.size() == 0);
            ASSERT(ptrStack.size() == 0);
            ASSERT(memory == topOfStack);
            delete[] memory;
        }

        void clear()
        {
            while(classHandlerStack.size())
                classHandlerStack.back()->deleteInstance(*this);
            ptrStack.clear();
            ASSERT(memory == topOfStack);

            frameStack.clear();
            frameStack.push(0);
#ifdef DEBUG
            frameStackDebug.clear();
#endif

            yieldCounter = 0;
        }

        template<typename T>
        inline void* push(bool callCtor = false)
        {
            if(topOfStack + sizeof(T) >= memory + MEMORY_SIZE)
            {
                //executionContext->errorHandler("B-Script: StackOverflow!!!\n");
                throw StackException("memory overflow");
            }
            void *ptr = callCtor ? ((void*)new(topOfStack) T) : topOfStack;

            topOfStack += sizeof(T);

            return ptr;
        }

        inline void pushPtr(void *ptr)
        {
            ptrStack.push(ptr);
        }

        inline void pushClassHandler(ast::ClassDeclaration *c)
        {
            classHandlerStack.push(c);
        }

        inline void setAbsPtr(size_t id, void *ptr)
        {
            ASSERT(ptrStack.size() > id);
            ptrStack.at(id) = ptr;
        }

        inline void pushFrame()
        {
            frameStack.push(ptrStack.size());
#ifdef DEBUG
            frameStackDebug.push(std::pair<size_t, int>(ptrStack.size(), topOfStack - memory));
#endif
        }

        template<typename T>
        inline void pop()
        {
            topOfStack -= sizeof(T);
            T *t = (T*)topOfStack;
            t->~T();
        }

        inline void* popPtr()
        {
            return ptrStack.pop();
        }

        inline ast::ClassDeclaration* popClassHandler()
        {
            return classHandlerStack.pop();;
        }

        inline void popFrame()
        {
#ifdef DEBUG
            const std::pair<size_t, int> &d = frameStackDebug.pop();
            //printf("size %d expected size %d\n", ptrStack.size(), d.first);
            //ASSERT(d.first == ptrStack.size());
            ASSERT(d.second == topOfStack - memory);
#endif
            frameStack.pop();
        }

        inline void* getFramePtr(int id, int frameOffset = 0)
        {
            //printf("stack frame %d , offset %d,  id %d\n", frameStack.size(), frameOffset, id);
            //printf("ptrStackSize %d, framepointer %d\n", ptrStack.size(), frameStack.back());
            size_t framePointer = frameStack.at(frameStack.size() - 1 + frameOffset);
            return ptrStack.at(framePointer + id);
        }

        inline void* getAbsPtr(int id)
        {
            return ptrStack.at(id);
        }

        inline size_t ptrStackSize()
        {
            return ptrStack.size();
        }

    private:
        char *memory;
        char *topOfStack;

        StackImpl<void *, POINTER_STACK_SIZE> ptrStack;
        StackImpl<size_t, FRAME_STACK_SIZE> frameStack;
        StackImpl<ast::ClassDeclaration*, CLASS_HANDLER_STACK_SIZE> classHandlerStack;
#ifdef DEBUG
        StackImpl<std::pair<size_t, int>, FRAME_STACK_SIZE> frameStackDebug;
#endif
};

