/*
b-script
Copyright (C) 2011 Thijs Jeffry de Haas

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

/*
 * This file implements a class which makes it handy to print out a tree to
 * stdout or a file. The -Wformat check of the g++ won't apply to it's println
 * function, a wrong format string for a set of parameters might cause a
 * segfault!
 * It might be nicer to somehow overwrite std::cout, but it wasn't obvisous to
 * me how to do it, so I'm gonna stick to this for now.
 */

#include <string>
#include <stdarg.h>
#include <stdio.h>
#include <exception>

#ifdef WIN32
#include "win32snprintf.h"
#endif

#include "asserts.h"

class TreePrinter
{
    public:
        class TreePrinterException : public std::exception
        {
            public:
                TreePrinterException(const std::string &msg)
                  : msg(msg)
                {}

                ~TreePrinterException() throw() {}

                virtual const char* what() const throw()
                {
                    return msg.c_str();
                }
            private:
                std::string msg;
        };

        TreePrinter(const std::string &filename = std::string("-"))
            : fp(stdout),
              indentation(""),
              lastNewLine(true)
        {
            if(filename.compare("-"))
            {
#ifndef WIN32
                fp = fopen(filename.c_str(), "w");
#else
                fopen_s(&fp, filename.c_str(), "w");
#endif
                if(!fp)
                    throw TreePrinterException(filename + ": can't open file");
            }
        }

        ~TreePrinter()
        {
            if(fp != stdout)
            {
                fclose(fp);
                fp = 0;
            }
        }

        void println(const char *format, ...)
        {
            if(!lastNewLine)
                fprintf(fp, "\n");

            #define BUFSIZE 1024
            char buf[BUFSIZE+1];

            va_list ap;
            va_start(ap, format);	
            vsnprintf(buf, BUFSIZE, format, ap);
            va_end(ap);

            fprintf(fp, "%s%s\n", indentation.c_str(), buf);
            lastNewLine = true;
        }

        void print(const char *format, ...)
        {
            if(lastNewLine)
            {
                fprintf(fp, "%s", indentation.c_str());
                lastNewLine = false;
            }

            #define BUFSIZE 1024
            char buf[BUFSIZE+1];

            va_list ap;
            va_start(ap, format);	
            vsnprintf(buf, BUFSIZE, format, ap);
            va_end(ap);

            fprintf(fp, "%s", buf);
        }

        void newline()
        {
            fprintf(fp, "\n");
            lastNewLine = true;
        }

        void indent()
        {
            indentation.insert(indentation.size(), (size_t)INDENT_STEP, ' ');
        }

        void dindent()
        {
            if(indentation.size() >= INDENT_STEP)
                indentation.erase(indentation.size()-INDENT_STEP, INDENT_STEP);
            else
                ASSERT(false);
        }

    private:
        enum
        {
            INDENT_STEP=4,
        };

        FILE *fp;
        std::string indentation;
        bool lastNewLine;
};

